/**
 * \file BitmapStructs.h
 * <hr>
 * \internal This file contains the constants and structures needed for working with
 *           BMP (Windows(TM) Bitmap) file format.
 * <hr>
 */
// ############################################################################################
//  BitmapStructs.h
//  Definition of constants and structures used by Windows(TM) Bitmap
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  The constants and structures in this file should be available from many documentations
//  found in the web, for examples :
//    http://downloads.xara.com/opensource/doxygen/html/annotated.html
//    http://developers.videolan.org/vlc/vlc/doc/doxygen/html/annotated.html
//    http://dl.njfiw.gov.cn/softwares/%B1%E0%B3%CC%BF%AA%B7%A2/VisualASM/tasm/Inc/Windows.inc
//    http://msdn.microsoft.com/library/default.asp?url=/library/en-us/gdi/bitmaps_2w1f.asp
//  and also from the MS SDK header file "wingdi.h" (of course ;)
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___BITMAP_STRUCTS_H_
#define _2D_MANIPULATION___BITMAP_STRUCTS_H_

// VortexGE headers
#include "../Base/VBasics.h"


namespace vge {

  // Constants used by Windows Bitmap
  const uint32_t __BI_RGB       = uint32_t(0);
  const uint32_t __BI_RLE8      = uint32_t(1);
  const uint32_t __BI_RLE4      = uint32_t(2);
  const uint32_t __BI_BITFIELDS = uint32_t(3);
  const uint32_t __BI_JPEG      = uint32_t(4);
  const uint32_t __BI_PNG       = uint32_t(5);


  //!
  //! \internal
  //! \brief Structures used by Windows(TM) Bitmap file
  //!
  //! \api stable
  struct _attr_packed _TRGBTRIPLE {
    uint8_t un8B, un8G, un8R;
  };

  //!
  //! \internal
  //! \brief Structures used by Windows(TM) Bitmap file
  //!
  //! \api stable
  struct _attr_packed _TRGBQUAD {
    uint8_t un8B, un8G, un8R, un8Reserved;
  };

  //!
  //! \internal
  //! \brief Structures used by Windows(TM) Bitmap file
  //!
  //! \api stable
  struct _attr_packed _TBITMAPFILEHEADER {
    u_char   aucBM[2];      // Always contains string "BM"
    uint32_t un32Size;      // File size in bytes
    uint16_t un16Reserved1; // Reserved, must be zero
    uint16_t un16Reserved2; // Reserved, must be zero
    uint32_t un32OffBits;   // Offset (in bytes) from the start of this structure to the bitmap bits
  };

  //!
  //! \internal
  //! \brief Structures used by Windows(TM) Bitmap file
  //!
  //! \api stable
  struct _attr_packed _TBITMAPCOREHEADER {
    uint32_t un32Size;     // Size of this structure in bytes
    uint16_t un16Width;    // Width of the bitmap
    uint16_t un16Height;   // Height of the bitmap
    uint16_t un16Planes;   // Number of planes for the target devices, must be one
    uint16_t un16BitCount; // The bits-per-pixel, must be 1, 4, 8, or 24
  };

  //!
  //! \internal
  //! \brief Structures used by Windows(TM) Bitmap file
  //!
  //! \api stable
  struct _attr_packed _TBITMAPINFOHEADER {
    uint32_t un32Size;          // Size of this structure in bytes
    int32_t  n32Width;          // Width of the bitmap
    int32_t  n32Height;         // Height of the bitmap
    uint16_t un16Planes;        // Number of planes for the target devices, must be one
    uint16_t un16BitCount;      // The bits-per-pixel, must be 0, 1, 4, 8, 16, 24, or 32
    uint32_t un32Compression;   // Must be __BI_RGB, __BI_RLE8, __BI_RLE4, __BI_BITFIELDS, __BI_JPEG, or __BI_PNG
    uint32_t un32SizeImage;     // Size of the image, can be zero for __BI_RGB
    int32_t  un32XPelsPerMeter; // Horizontal resolution in pixels-per-meter
    int32_t  un32YPelsPerMeter; // Vertical resolution in pixels-per-meter
    uint32_t un32ClrUsed;       // Number of color indexes in the color table used by the bitmap
    uint32_t un32ClrImportant;  // Number of color indexes table required for displaying the bitmap
  };

} // namespace vge
#endif
