/** \file BasicFractals.h */
// ############################################################################################
//  BasicFractals.h
//  Definition & implementation of CBasicFractal, CDefiniteAreaFractal,
//  and CIndefiniteAreaFractal classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___FRACTAL___BASIC_FRACTALS_H_
#define _2D_MANIPULATION___FRACTAL___BASIC_FRACTALS_H_

// VortexGE headers
#include "../ImageBuilder.h"


namespace vge {

  //!
  //! \internal
  //! \brief 2D Transformation data structure
  //!
  //! Used for storing 2D transformation data of fractal's parts.
  //!
  //! \api experiment
  struct _TSTransform {
    int32_t n32Sx; //!< X scalling amount (in %)
    int32_t n32Sy; //!< Y scalling amount (in %)
    int32_t n32Rx; //!< X rotation amount
    int32_t n32Ry; //!< Y rotation amount
    int32_t n32Tx; //!< X translation amount
    int32_t n32Ty; //!< Y translation amount
    bool    bTBD;  //!< A flag to indicate whether this part will be drawn or just as parent
  };

  //!
  //! \internal
  //! \brief Affine transformation data structure
  //!
  //! Used for storing affine transformation data of fractal's parts.
  //!
  //! \api experiment
  struct _TAffine {
    float fA, fB, fC, fD; // The A matrix values
    float fE, fF;         // The T matrix values
  };


  //!
  //! \brief The base of all fractal classes
  //!
  //! \api experiment
  class VGE_EXPORT CBasicFractal : VCLASS_UNCOPYABLE {
    public:
      virtual ~CBasicFractal() throw();

      // Direct access to the fractal's random number generator
      //!
      //! Direct access to the fractal's random number generator
      inline CRandomCoreUInt32 &RandomNumberGenerator()
      { return(_rnGen); }
      //!
      //! Direct access to the fractal's random number generator
      inline const CRandomCoreUInt32 &RandomNumberGenerator() const
      { return(_rnGen); }

    protected:
      CBasicFractal();

      // Transformation functions
      //!
      //! Used for generating the corresponding affine transformation matrices
      //! from the given transformation data.\n
      //! \n
      //! \note The structure _TAffine actually contains 2 matrices.
      void GenerateAffineStruct(_TAffine &tADst, const _TSTransform &tSSrc);
      //!
      //! Perform Affine matrices multiplication according to the equation:\n
      //! <span style="margin-left:2em;"><b>|Dst| = |Dst| x |Src|</b></span>\n
      //! \note The structure _TAffine actually contains 2 matrices.
      void MultiplyAffineStruct(_TAffine &tADst, const _TAffine &tASrc);
      //!
      //! Perform the transformation:\n
      //! <span style="margin-left:2em;"><b>(SrcX, SrcY) -> (DstX, DstY)</b></span>\n
      //! using the given affine transformation.
      void AffineTransform(int32_t &n32DstX, int32_t &n32DstY, const _TAffine &tA, int32_t n32SrcX, int32_t n32SrcY);

      // Color functions
      //!
      //! Used for setting the active color.
      inline void SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
      { _un8R = un8R; _un8G = un8G; _un8B = un8B; }
      //!
      //! Used for getting the active color.
      inline void GetColor(uint8_t &un8R, uint8_t &un8G, uint8_t &un8B) const
      { un8R = _un8R; un8G = _un8G; un8B = _un8B; }

      // Basic drawing functions (all coordinates are in cartesian space)
      //!
      //! Used for drawing pixel using the active color in the given (X, Y) location.\n
      //! \n
      //! \note All coordinates are in cartesian space.
      inline void PutPixel(int32_t n32X, int32_t n32Y)
      { PutPixel(n32X, n32Y, _un8R, _un8G, _un8B); }
      //!
      //! Used for drawing pixel using the given color in the given (X, Y) location.\n
      //! \n
      //! \note All coordinates are in cartesian space.
      void PutPixel(int32_t n32X, int32_t n32Y, uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Used for drawing pixel using the given color in the given (X, Y) location
      //! <b>but</b> only if the location contains a pure black pixel.\n
      //! \n
      //! \note All coordinates are in cartesian space.
      void SetPixel(int32_t n32X, int32_t n32Y, uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Used for getting the pixel color of the given (X, Y) location.\n
      //! \n
      //! \note All coordinates are in cartesian space.
      void GetPixel(int32_t n32X, int32_t n32Y, uint8_t &un8R, uint8_t &un8G, uint8_t &un8B) const;
      //!
      //! Used for drawing line from (X1, Y1) to (X2, Y2) using the active color.\n
      //! \n
      //! \note All coordinates are in cartesian space.
      void DrawLine(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2);
      //!
      //! Used for drawing polygon using the active color.
      //! \param[in] n32X[]    Array of X coordinates.
      //! \param[in] n32Y[]    Array of Y coordinates.
      //! \param[in] un32Count Number of coordinates in the array.
      //!
      //! \note The last point in the array will be connected with the 1st one.\n
      //! \note All coordinates are in cartesian space.
      void DrawPolygon(const int32_t n32X[], const int32_t n32Y[], uint32_t un32Count);

      // Data
      CRandomCoreUInt32 _rnGen;              // Core random number generator
      CRandomFloat      _rnFloat;            // Floating points random number generator
      CImageBuilder    *_pcsIBuild;          // Used image builder class
      uint8_t           _un8R, _un8G, _un8B; // Active color
  };


  //!
  //! \brief The base of all definite area fractal classes
  //!
  //! \api experiment
  class VGE_EXPORT CDefiniteAreaFractal : public CBasicFractal {
    public:
      //!
      //! Derivable drawing function for definite area fractals.\n
      //! \n
      //! \note All coordinates are in standard screen coordinates.
      virtual void Draw(CImageBuilder *pcsDstIBuild, int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2) = 0;
  };


  //!
  //! \brief The base of all indefinite area fractal classes
  //!
  //! \api experiment
  class VGE_EXPORT CIndefiniteAreaFractal : public CBasicFractal {
    public:
      //! Derivable drawing function for indefinite area fractals.\n
      //! \n
      //! \note All coordinates are in standard screen coordinates.
      virtual void Draw(CImageBuilder *pcsDstIBuild, int32_t n32XStart, int32_t n32YStart) = 0;
  };

} // namespace vge
#endif
