// ############################################################################################
//  IndefiniteAreaFractals.cc
//  Implementation of various CCNBFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Pemrograman Fraktal Resolusi Tinggi untuk Kartu Trident SVGA
//    Ir. P. Insap Santosa, M.Sc. & Ir. R. Tory K.
//    Penerbit ANDI Yogyakarta
//    Indonesia, 1997
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "ComplexNumberBasedFractals.h"

// C++ headers
#include <complex>


namespace vge {

//
// Mandelbrot
//
uint32_t FractalMandelbrot::IterateEqu(float fX, float fY, uint32_t un32MaxIter)
{
  // Iterate the equation : z(n) = z(n-1)^2 + c
  //                        with : z(0) = 0
  //                               c    = x + iy0   (from fX and fY parameters)
  //
  //                        Exit point : when |z(n)| >= 4.0
  float    fXc     = 0.0f;
  float    fYc     = 0.0f;
  float    fSq     = 0.0f;
  uint32_t un32Cnt = 0;
  while((un32Cnt < un32MaxIter) && (fSq < 4.0f)) {
    float fTc = fXc;
    fXc = (fXc * fXc) - (fYc * fYc) + fX;
    fYc = (2.0f * fTc * fYc) + fY;
    fSq = (fXc * fXc) + (fYc * fYc);
    un32Cnt++;
  }
  return(un32Cnt);
}

void FractalMandelbrot::SelectDefaultWindow()
{
  SelectWindow(-2.05f, -1.20f, 0.55f, 1.20f);
}


//
// Julia
//
uint32_t FractalJulia::IterateEqu(float fX, float fY, uint32_t un32MaxIter)
{
  // Iterate the equation : z(n) = z(n-1)^2
  //                        with : z(0) = x + iy   (from fX and fY parameters)
  //
  //                        Exit point : when |z(n)| >= 4.0
  float    fPv     = _tW.fPv;
  float    fQv     = _tW.fQv;
  float    fXc     = fX;
  float    fYc     = fY;
  float    fSq     = 0.0f;
  uint32_t un32Cnt = 0;
  while((un32Cnt < un32MaxIter) && (fSq < 4.0f)) {
    float fTc = fXc;
    fXc = (fXc * fXc) - (fYc * fYc) + fPv;
    fYc = (2.0f * fTc * fYc) + fQv;
    fSq = (fXc * fXc) + (fYc * fYc);
    un32Cnt++;
  }
  return(un32Cnt);
}

void FractalJulia::SelectDefaultWindow()
{
  SelectWindow(-1.35f, -1.45f, 1.35f, 1.45f);
}

void FractalJulia::SelectDefaultSeed()
{
  SelectSeed(0.30f, 0.60f);
}


//
// Dragon
//
uint32_t FractalDragon::IterateEqu(float fX, float fY, uint32_t un32MaxIter)
{

  // Iterate the equation : z(n) = cz(n-1) - cz(n-1)^2
  //                        with : z(0) = x + iy   (from fX and fY parameters)
  //                               c    = p + iq   (from _tW.fPv and _tW.fQv)
  //
  //                        Exit point : when |z(n)| >= 4.0
  float    fPv     = _tW.fPv;
  float    fQv     = _tW.fQv;
  float    fXc     = fX;
  float    fYc     = fY;
  float    fSq     = 0.0f;
  uint32_t un32Cnt = 0;
  while((un32Cnt < un32MaxIter) && (fSq < 4.0f)) {
    float fXq = fXc * fXc;
    float fYq = fYc * fYc;
    float fAq = fXc - fXq + fYq;
    float fBq = fYc - (2 * fXc * fYc);
    fXc = (fPv * fAq) - (fQv * fBq);
    fYc = (fPv * fBq) + (fQv * fAq);
    fSq = fXq + fYq;
    un32Cnt++;
  }
  return(un32Cnt);
}

void FractalDragon::SelectDefaultWindow()
{
  SelectWindow(-0.15f, -0.62f, 1.15f, 0.62f);
}

void FractalDragon::SelectDefaultSeed()
{
  SelectSeed(1.646009f, 0.967049f);
}


//
// Sierspinski
//
uint32_t FractalSierspinski::IterateEqu(float fX, float fY, uint32_t un32MaxIter)
{
  //                                / 2z(n-1) - i   (if Im[z(n-1)] > 0.5)
  // Iterate the equation : z(n) = |  2z(n-1) - 1   (if Re[z(n-1)] > 0.5)
  //                                \ 2z(n-1)       (if others)
  //                        with : z(0) = x + iy    (from fX and fY parameters)
  //
  //                        Exit point : when |z(n)| >= 4.0
  float    fXc     = fX;
  float    fYc     = fY;
  float    fSq     = 0.0f;
  uint32_t un32Cnt = 0;
  while((un32Cnt < un32MaxIter) && (fSq < 4.0f)) {
    if(fYc > 0.5f) {
      fXc = 2 * fXc;
      fYc = 2 * fYc - 1;
    }
    else if(fXc > 0.5f) {
      fXc = 2 * fXc - 1;
      fYc = 2 * fYc;
    }
    else {
      fXc *= 2;
      fYc *= 2;
    }
    fSq = (fXc * fXc) + (fYc * fYc);
    un32Cnt++;
  }
  return(un32Cnt);
}

void FractalSierspinski::SelectDefaultWindow()
{
  SelectWindow(-0.05f, -0.05f, 1.05f, 1.05f);
}


//
// Sun
//
uint32_t FractalSun::IterateEqu(float fX, float fY, uint32_t un32MaxIter)
{
  // Iterate the equation : z(n) = z(n-1)^2
  //                        with : z(0) = x + iy   (from fX and fY parameters)
  //
  //                        Exit point : when Re[z(n)] * Im[z(n)] >= 4.0
  float    fXc     = fX;
  float    fYc     = fY;
  float    fSq     = 0.0f;
  uint32_t un32Cnt = 0;
  while((un32Cnt < un32MaxIter) && (fSq < 4.0f)) {
    float fTc = fXc;
    fXc = (fXc * fXc) - (fYc * fYc);
    fYc = (2.0f * fTc * fYc);
    fSq = (fXc * fYc);
    un32Cnt++;
  }
  return(un32Cnt);
}

void FractalSun::SelectDefaultWindow()
{
  SelectWindow(-2.00f, -2.00f, 2.00f, 2.00f);
}


//
// Transcendental-function-based fractal derivatives
//
void FractalFSin::CalculateEqu(complex<float> &tZ, const complex<float> &tC)  { tZ = sin(tZ) + tC; }
void FractalFCos::CalculateEqu(complex<float> &tZ, const complex<float> &tC)  { tZ = cos(tZ) + tC; }
void FractalFSinh::CalculateEqu(complex<float> &tZ, const complex<float> &tC) { tZ = sinh(tZ) + tC; }
void FractalFCosh::CalculateEqu(complex<float> &tZ, const complex<float> &tC) { tZ = cosh(tZ) + tC; }
void FractalFExp::CalculateEqu(complex<float> &tZ, const complex<float> &tC)  { tZ = tC * exp(tZ); }

} // namespace vge
