// ############################################################################################
//  IndefiniteAreaFractals.cc
//  Implementation of CIndefiniteAreaFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Fractal Vision : Put Fractals to Work for You
//    Dick Oliver
//    Sams Publishing
//    1992
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "IndefiniteAreaFractals.h"


namespace vge {

//
// Helper functions
//
static uint32_t _CalcTotalTrans_(uint32_t un32Level, uint32_t un32ATot)
{
  uint32_t un32CurA;
  if     (un32Level == 0) un32CurA = 1;
  else if(un32Level == 1) un32CurA = un32ATot;
  else {
    un32CurA = un32ATot;
    for(uint32_t un32CntA = 0; un32CntA < (un32Level - 1); un32CntA++)
      un32CurA *= un32ATot;
  }
  return(un32CurA);
}

static uint32_t _CalcTotalCells_(uint32_t un32Level, uint32_t un32ATot)
{
  uint32_t un32TotCells = 0;
  for(uint32_t un32CntC = 0; un32CntC <= un32Level; un32CntC++)
    un32TotCells += _CalcTotalTrans_(un32CntC, un32ATot);
  return(un32TotCells);
}


//
// The base of all seed-replicator fractals
//
CSeedReplicatorFractal::CSeedReplicatorFractal()
: _ptTrns(NULL), _un32TCnt(0)
{
  memset(&_tSeed, 0, sizeof(_tSeed));
}

CSeedReplicatorFractal::~CSeedReplicatorFractal() throw()
{
  delete [] _tSeed.pn32X;
  delete [] _tSeed.pn32Y;
  delete [] _ptTrns;
}

bool CSeedReplicatorFractal::SEED_REPLICATOR_FRACTAL_ALLOC(uint32_t un32SeedElemCounts, uint32_t un32SeedPointCounts, uint32_t un32ReplicaCounts)
{
  _tSeed.pn32X = _nothrow_new int32_t[un32SeedElemCounts * un32SeedPointCounts];
  _tSeed.pn32Y = _nothrow_new int32_t[un32SeedElemCounts * un32SeedPointCounts];
  _ptTrns      = _nothrow_new _TSTransform[un32ReplicaCounts];
  if((!_tSeed.pn32X) || (!_tSeed.pn32Y) || (!_ptTrns)) {
    delete [] _tSeed.pn32X;
    delete [] _tSeed.pn32Y;
    delete [] _ptTrns;
    _tSeed.un32NumElements = _tSeed.un32NumPoints = _un32TCnt = 0;
    VWarning(_NFL_, "Initialization of fractal %s has failed; no fractal will be drawn", _N_);
    return(false);
  }
  _tSeed.un32NumElements = un32SeedElemCounts;
  _tSeed.un32NumPoints   = un32SeedPointCounts;
  _un32TCnt              = un32ReplicaCounts;
  return(true);
}

void CSeedReplicatorFractal::SEED_POINT(uint32_t un32EIdx, uint32_t un32PIdx, int32_t n32X, int32_t n32Y)
{
  _tSeed.pn32X[(un32EIdx * _tSeed.un32NumElements) + un32PIdx] = n32X;
  _tSeed.pn32Y[(un32EIdx * _tSeed.un32NumElements) + un32PIdx] = n32Y;
}

void CSeedReplicatorFractal::SEED_REPLICA(uint32_t un32Idx, int32_t n32Sx, int32_t n32Sy, int32_t n32Rx, int32_t n32Ry, int32_t n32Tx, int32_t n32Ty, bool bToBeDrawn)
{
  _ptTrns[un32Idx].n32Sx = n32Sx;
  _ptTrns[un32Idx].n32Sy = n32Sy;
  _ptTrns[un32Idx].n32Rx = n32Rx;
  _ptTrns[un32Idx].n32Ry = n32Ry;
  _ptTrns[un32Idx].n32Tx = n32Tx;
  _ptTrns[un32Idx].n32Ty = n32Ty;
  _ptTrns[un32Idx].bTBD  = bToBeDrawn;
}

bool CSeedReplicatorFractal::CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_6(un32MaxLevel, un32CurLevel, un32TotPart, un32CurPart, un32TotRelPart, un32CurRelPart);
  SetColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, MAX_COLOR_VALUE);
  return(true);
}

bool CSeedReplicatorFractal::ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32MaxLevel, un32CurLevel, un32TotRelPart, un32CurRelPart);
  SetColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, MAX_COLOR_VALUE);
  return(true);
}

uint32_t CSeedReplicatorFractal::TotalReplicationLevel()          const { return(INDEFINITE_AREA_FRACTAL_DEFAULT_REPLICATION_LEVEL); }
bool     CSeedReplicatorFractal::CosmosStageIsEnabled()           const { return(true); }
uint32_t CSeedReplicatorFractal::CosmosStartLevelThreshold()      const { return(INDEFINITE_AREA_FRACTAL_DEFAULT_COSMOS_START); }
uint32_t CSeedReplicatorFractal::CosmosPolygonEndLevelThreshold() const { return(0xFFFFFFFF); }
bool     CSeedReplicatorFractal::ChaosStageIsEnabled()            const { return(true); }
uint32_t CSeedReplicatorFractal::ChaosStartLevelThreshold()       const { return(INDEFINITE_AREA_FRACTAL_DEFAULT_CHAOS_START); }
uint32_t CSeedReplicatorFractal::ChaosEndLevelThreshold()         const { return(0xFFFFFFFF); }
uint8_t  CSeedReplicatorFractal::ChaosFactor()                    const { return(INDEFINITE_AREA_FRACTAL_DEFAULT_CHAOS_FACTOR); }
uint8_t  CSeedReplicatorFractal::ChaosIterationLimit()            const { return(INDEFINITE_AREA_FRACTAL_DEFAULT_CHAOS_ITERATION); }

void CSeedReplicatorFractal::DrawImpl(_TAffineStorage *ptDstAStorage, uint32_t un32StartLT, uint32_t un32PolygonEndLT,
                                                  int32_t n32XStart, int32_t n32YStart, const _TAffine *pcsCurA,
                                                  uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32CurPart)
{
  // Check if exceeding the maximum level
  if(un32CurLevel > un32MaxLevel) return;

  // Generate affine transformation structures
  _TAffine tA[_un32TCnt];
  for(uint32_t un32CntA = 0; un32CntA < _un32TCnt; un32CntA++) {
    GenerateAffineStruct(tA[un32CntA], _ptTrns[un32CntA]);
  }

  // Draw level 0 (the seed)
  if(un32CurLevel == 0) {
    // Convert the n32YStart from standard screen coordinate to cartesian space
    n32YStart = _pcsIBuild->GetHeight() - n32YStart;
    // Save the current transformation
    if(ptDstAStorage) {
      uint32_t un32WP = ptDstAStorage->un32WritePtr;
      ptDstAStorage->ptData[un32WP].tAffine        = *pcsCurA;
      ptDstAStorage->ptData[un32WP].un32CurLevel   = 0;
      ptDstAStorage->ptData[un32WP].un32CurPart    = 1;
      ptDstAStorage->ptData[un32WP].un32CurRelPart = 1;
      ptDstAStorage->un32WritePtr++;
    }
    // Draw the finite stage only if asked
    if(CosmosStageIsEnabled() && (un32StartLT == 0)) {
      // Update color for this level
      CosmosUpdateColor(un32MaxLevel, 0, 1, 0, 1, 0);
      // Call the calback first
      if(!DrawingProgressCallback(un32MaxLevel, 0, 1, 0, 1, 0)) return;
      // Perform transformation and drawing
      for(uint32_t un32CntE = 0; un32CntE < _tSeed.un32NumElements; un32CntE++) {
        // Calculate start index
        uint32_t un32SI = un32CntE * _tSeed.un32NumElements;
        // Transform the points
        int32_t n32X[_tSeed.un32NumPoints];
        int32_t n32Y[_tSeed.un32NumPoints];
        for(uint32_t un32CntP = 0; un32CntP < _tSeed.un32NumPoints; un32CntP++) {
          AffineTransform(n32X[un32SI + un32CntP], n32Y[un32SI + un32CntP], *pcsCurA, _tSeed.pn32X[un32CntP], _tSeed.pn32Y[un32CntP]);
          n32X[un32CntP] += n32XStart;
          n32Y[un32CntP] += n32YStart;
        }
        // Draw the polygon
        if(un32CurLevel < un32PolygonEndLT)
          DrawPolygon(n32X, n32Y, _tSeed.un32NumPoints);
        else {
          for(uint32_t un32CntP = 0; un32CntP < _tSeed.un32NumPoints; un32CntP++)
            PutPixel(n32X[un32CntP], n32Y[un32CntP]);
        }
      }
    }
    // Draw starting from level 1 recursively to the maximum level
    DrawImpl(ptDstAStorage, un32StartLT, un32PolygonEndLT, n32XStart, n32YStart, pcsCurA, un32MaxLevel, 1, 0);
    return;
  }

  // Draw all transformed seeds in this level
  if(un32CurLevel >= 1) {
    // Calculate some values
    uint32_t un32TP = _CalcTotalTrans_(un32CurLevel, _un32TCnt);
    // Draw all parts
    for(uint32_t un32CntA = 0; un32CntA < _un32TCnt; un32CntA++) {
      // Calculate the current part number
      uint32_t un32CP = (un32CurPart * _un32TCnt) + un32CntA;
      // Combine with the previous transformation
      if(pcsCurA) MultiplyAffineStruct(tA[un32CntA], *pcsCurA);
      // Check if this part must not be drawn
      if(!_ptTrns[un32CntA].bTBD) continue;
      // Save the current transformation
      if(ptDstAStorage) {
        uint32_t un32WP = ptDstAStorage->un32WritePtr;
        ptDstAStorage->ptData[un32WP].tAffine        = tA[un32CntA];
        ptDstAStorage->ptData[un32WP].un32CurLevel   = un32CurLevel;
        ptDstAStorage->ptData[un32WP].un32CurPart    = un32CP;
        ptDstAStorage->ptData[un32WP].un32CurRelPart = un32CntA;
        ptDstAStorage->un32WritePtr++;
      }
      // Draw the finite stage only if asked
      if(CosmosStageIsEnabled() && (un32CurLevel >= un32StartLT)) {
        // Call the calback first
        if(!DrawingProgressCallback(un32MaxLevel, un32CurLevel, un32TP, un32CP, _un32TCnt, un32CntA)) return;
        // Perform transformation and drawing
        for(uint32_t un32CntE = 0; un32CntE < _tSeed.un32NumElements; un32CntE++) {
          // Calculate start index
          uint32_t un32SI = un32CntE * _tSeed.un32NumElements;
          // Transform the points
          int32_t n32X[_tSeed.un32NumPoints];
          int32_t n32Y[_tSeed.un32NumPoints];
          for(uint32_t un32CntP = 0; un32CntP < _tSeed.un32NumPoints; un32CntP++) {
            AffineTransform(n32X[un32SI + un32CntP], n32Y[un32SI + un32CntP], tA[un32CntA], _tSeed.pn32X[un32CntP], _tSeed.pn32Y[un32CntP]);
            n32X[un32CntP] += n32XStart;
            n32Y[un32CntP] += n32YStart;
          }
          // Update color for this level and part
          if(!CosmosUpdateColor(un32MaxLevel, un32CurLevel, un32TP, un32CP, _un32TCnt, un32CntA)) continue;
          // Draw the polygon
          if(un32CurLevel < un32PolygonEndLT)
            DrawPolygon(n32X, n32Y, _tSeed.un32NumPoints);
          else {
            for(uint32_t un32CntP = 0; un32CntP < _tSeed.un32NumPoints; un32CntP++)
              PutPixel(n32X[un32CntP], n32Y[un32CntP]);
          }
        }
      }
    }
  }

  // Draw all the next levels recursively
  for(uint32_t un32CntA = 0; un32CntA < _un32TCnt; un32CntA++)
    DrawImpl(ptDstAStorage, un32StartLT, un32PolygonEndLT, n32XStart, n32YStart, &tA[un32CntA],
              un32MaxLevel, un32CurLevel + 1, (un32CurPart * _un32TCnt) + un32CntA);
}

void CSeedReplicatorFractal::PaintImpl(int32_t n32XStart, int32_t n32YStart,
                                                   _TAffineStorage *ptDstAStorage,
                                                   uint32_t un32StartLT, uint32_t un32EndLT,
                                                   uint8_t un8ChaosFactor, uint8_t un8ChaosIter)
{
  // Check for the minimum level
  if(ptDstAStorage->un32MaxLevel < un32StartLT) return;

  // Convert the n32YStart from standard screen coordinate to cartesian space
  n32YStart = _pcsIBuild->GetHeight() - n32YStart;

  // Draw the points
  for(uint32_t un32CntC = 2; un32CntC < ptDstAStorage->un32WritePtr; un32CntC++) {
    // Check
    if(ptDstAStorage->ptData[un32CntC].un32CurLevel < un32StartLT) continue;
    if(ptDstAStorage->ptData[un32CntC].un32CurLevel > un32EndLT) continue;
    // Update color
    if(!ChaosUpdateColor(ptDstAStorage->un32MaxLevel, ptDstAStorage->ptData[un32CntC].un32CurLevel,
                          _un32TCnt, ptDstAStorage->ptData[un32CntC].un32CurRelPart))
      continue;
    // Draw all points
    int32_t n32X, n32Y;
    for(uint32_t un32CntE = 0; un32CntE < _tSeed.un32NumElements; un32CntE++) {
      for(uint32_t un32CntP = 0; un32CntP < _tSeed.un32NumPoints; un32CntP++) {
        // Perform transformation
        AffineTransform(n32X, n32Y, ptDstAStorage->ptData[un32CntC].tAffine, _tSeed.pn32X[un32CntP], _tSeed.pn32Y[un32CntP]);
        // Draw the pixels
        for(uint32_t un32S = 0; un32S < uint32_t(un8ChaosIter); un32S++) {
          if(un8ChaosFactor) {
            //float   fCF   = float(ptDstAStorage->un32MaxLevel - ptDstAStorage->ptData[un32CntC].un32CurLevel) / float(ptDstAStorage->un32MaxLevel);
            //int32_t n32CF = fCF * float(un8ChaosFactor);
            //        n32CF = (n32CF == 0) ? 1 : n32CF;
            int32_t n32CF = un8ChaosFactor;
            int32_t n32XC = int32_t(_rnFloat.GenerateBipolar(-n32CF, n32CF));
            int32_t n32YC = int32_t(_rnFloat.GenerateBipolar(-n32CF, n32CF));
            PutPixel(n32X + n32XC + n32XStart, n32Y + n32YC + n32YStart);
          }
          else
            PutPixel(n32X + n32XStart, n32Y + n32YStart);
        }
      }
    }
  }
}

void CSeedReplicatorFractal::Draw(CImageBuilder *pcsDstIBuild, int32_t n32XStart, int32_t n32YStart)
{
  Draw(pcsDstIBuild, n32XStart, n32YStart, 100, 0);
}

void CSeedReplicatorFractal::Draw(CImageBuilder *pcsDstIBuild, int32_t n32XStart, int32_t n32YStart, int32_t n32InitialScale, int32_t n32InitialRotation)
{
  // Check
  if(_tSeed.un32NumPoints < 2) return;

  // Copy the target image builder
  _pcsIBuild = pcsDstIBuild;

  // Generate the initial transform
  _TSTransform tS;
  _TAffine     tA;
  tS.n32Sx = n32InitialScale;    tS.n32Sy = n32InitialScale;
  tS.n32Rx = n32InitialRotation; tS.n32Ry = n32InitialRotation;
  tS.n32Tx = 0;                  tS.n32Ty = 0;
  GenerateAffineStruct(tA, tS);

  // Allocate space for storage
  _TAffineStorage tAS, *ptAS = NULL;
  if(ChaosStageIsEnabled()) {
    tAS.un32DataSize = _CalcTotalCells_(TotalReplicationLevel(), _un32TCnt);
    tAS.ptData       = _nothrow_new _TAffineStorage::_Element[tAS.un32DataSize];
    tAS.un32WritePtr = 0;
    tAS.un32MaxLevel = TotalReplicationLevel();
    if(tAS.ptData) ptAS = &tAS;
  }

  // Draw the fractal
  DrawImpl(ptAS, CosmosStartLevelThreshold(), CosmosPolygonEndLevelThreshold(),
            n32XStart, n32YStart, &tA, TotalReplicationLevel());
  if(ptAS) {
    PaintImpl(n32XStart, n32YStart, ptAS,
               ChaosStartLevelThreshold(), ChaosEndLevelThreshold(),
               ChaosFactor(), ChaosIterationLimit());
    delete [] tAS.ptData;
  }
}

// Drawing callback function
// NOTE: * Override this function to display/control the replication progress.
//       * Returning 'false' will stop the drawing.
bool CSeedReplicatorFractal::DrawingProgressCallback(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_6(un32MaxLevel, un32CurLevel, un32TotPart, un32CurPart, un32TotRelPart, un32CurRelPart);
  return(true);
}

} // namespace vge
