/** \file IndefiniteAreaFractals.h */
// ############################################################################################
//  IndefiniteAreaFractals.h
//  Definition of CIndefiniteAreaFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Fractal Vision : Put Fractals to Work for You
//    Dick Oliver
//    Sams Publishing
//    1992
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___FRACTAL___INDEFINITE_AREA_FRACTALS_H_
#define _2D_MANIPULATION___FRACTAL___INDEFINITE_AREA_FRACTALS_H_

// VortexGE headers
#include "BasicFractals.h"

namespace vge {

  // Indefinite area fractal default paramaters
  const uint32_t INDEFINITE_AREA_FRACTAL_DEFAULT_REPLICATION_LEVEL = uint32_t(7); //!< Default replication level
  const uint32_t INDEFINITE_AREA_FRACTAL_DEFAULT_COSMOS_START      = uint32_t(0); //!< Default level in which the cosmos drawing stage begin
  const uint32_t INDEFINITE_AREA_FRACTAL_DEFAULT_CHAOS_START       = uint32_t(2); //!< Default level in which the chaos drawing stage begin
  const uint8_t  INDEFINITE_AREA_FRACTAL_DEFAULT_CHAOS_FACTOR      =  uint8_t(1); //!< Default chaos factor
  const uint8_t  INDEFINITE_AREA_FRACTAL_DEFAULT_CHAOS_ITERATION   =  uint8_t(1); //!< Default chaos iteration

  //!
  //! \brief Seed (polygon) structure
  //!
  //! The X and Y coordinates are in cartesian space which means
  //! the (0, 0) point is in the bottom left of the monitor
  //!
  //! \api experiment
  struct _TSeed {
    uint32_t  un32NumElements; //!< Number of elements
    uint32_t  un32NumPoints;   //!< Number of points in each element
    int32_t  *pn32X;           //!< List of X coordinates
    int32_t  *pn32Y;           //!< List of Y coordinates
  };

  //!
  //! \internal
  //! \brief Cumulative affine transformation data storage structure
  //!
  //! Used for storing affine transformation data of all fractal's parts.
  //!
  //! \api experiment
  struct _TAffineStorage {
    struct _Element {
      _TAffine tAffine;        // Affine transformation for the current part
      uint32_t un32CurLevel;   // Current level
      uint32_t un32CurPart;    // Current part
      uint32_t un32CurRelPart; // Current relative part
    };
    _Element *ptData;
    uint32_t un32DataSize; // Allocated size of ptData
    uint32_t un32WritePtr; // Usable size of ptData
    uint32_t un32MaxLevel; // Value from TotalReplicationLevel()
  };


  //!
  //! \brief The base of all seed replicator fractal classes
  //!
  //! See the page \ref More_Info_About_CSeedReplicatorFractal "More Information About CSeedReplicatorFractal Class" for more explanations.
  //!
  //! \api experiment
  class VGE_EXPORT CSeedReplicatorFractal : public CIndefiniteAreaFractal {
    public:
      CSeedReplicatorFractal();
      virtual ~CSeedReplicatorFractal() throw();

      // Drawing functions (all coordinates are in standard screen coordinates)
      //!
      //! Derivable drawing function for seed-replicator fractals.
      //! \param[in] pcsDstIBuild        Destination CImageBuilder instance.
      //! \param[in] n32XStart,n32YStart Starting coordinates for drawing.
      //!
      //! \note All coordinates are in standard screen coordinates.
      virtual void Draw(CImageBuilder *pcsDstIBuild, int32_t n32XStart, int32_t n32YStart);
      //!
      //! Derivable drawing function for seed-replicator fractals.
      //! \param[in] pcsDstIBuild        Destination CImageBuilder instance.
      //! \param[in] n32XStart,n32YStart Starting coordinates for drawing.
      //! \param[in] n32InitialScale     Initial scalling factor to be added.
      //! \param[in] n32InitialRotation  Initial rotation factor to be added.
      //!
      //! \note All coordinates are in standard screen coordinates.
      virtual void Draw(CImageBuilder *pcsDstIBuild, int32_t n32XStart, int32_t n32YStart, int32_t n32InitialScale, int32_t n32InitialRotation = 0);
      // Drawing callback functions
      //!
      //! Override this callback to follow the progress of the drawing.\n
      //! See the page \ref More_Info_About_CSeedReplicatorFractal "More Information AboutCSeedReplicatorFractal Class" for more explanations.
      virtual bool DrawingProgressCallback(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart);

    protected:
      // Fractal initialization functions
      //!
      //! Allocate the necessary memory for storing the fractal seed and replicas data.
      //! \param[in] un32SeedElemCounts  Number of elements in the seed.
      //! \param[in] un32SeedPointCounts Number of points for each element (currently all elements must have the same number of points).
      //! \param[in] un32ReplicaCounts   Number of replicas for the 1st level.
      //!
      //! \note This function must be called in the constructor of the derivative classes.
      bool SEED_REPLICATOR_FRACTAL_ALLOC(uint32_t un32SeedElemCounts, uint32_t un32SeedPointCounts, uint32_t un32ReplicaCounts);
      //!
      //! Define a point for the seed.
      //! \param[in] un32EIdx  Index of the element in which the point will be defined.
      //! \param[in] un32PIdx  Index of the point to be defined.
      //! \param[in] n32X,n32Y The point.
      //!
      //! \note All coordinates are in cartesian space.\n
      //! \note This function must be called in the constructor of the derivative classes.
      void SEED_POINT(uint32_t un32EIdx, uint32_t un32PIdx, int32_t  n32X, int32_t n32Y);
      //!
      //! Define a seed replica.
      //! \param[in] un32Idx     Index of the replica to be defined.
      //! \param[in] n32Sx,n32Sy Scalling factor for this replica.
      //! \param[in] n32Rx,n32Ry Rotation factor for this replica.
      //! \param[in] n32Tx,n32Ty Translation factor for this replica.
      //! \param[in] bToBeDrawn  Shall this replica be drawn? Or it just for "container" for other replicas?
      //!
      //! \note All coordinates are in cartesian space.\n
      //! \note This function must be called in the constructor of the derivative classes.
      void SEED_REPLICA(uint32_t un32Idx, int32_t n32Sx, int32_t n32Sy, int32_t n32Rx, int32_t n32Ry, int32_t n32Tx, int32_t n32Ty, bool bToBeDrawn = true);

      // Parameter-control functions
      //!
      //! Override this to control the replication level.
      virtual uint32_t TotalReplicationLevel() const;
      //!
      //! Override this to enable/disable the cosmos drawing stage.
      virtual bool CosmosStageIsEnabled() const;
      //!
      //! Override this to control the starting level in which the cosmos stage will be started to be drawn.
      virtual uint32_t CosmosStartLevelThreshold() const;
      //!
      //! Override this to control the ending level in which the cosmos stage will started to be drawn just using points instead of polygons.
      virtual uint32_t CosmosPolygonEndLevelThreshold() const;
      //!
      //! Override this callback to control the color of the seed and replicas drawn in the cosmos stage.\n
      //! See the page \ref More_Info_About_CSeedReplicatorFractal "More Information AboutCSeedReplicatorFractal Class" for more explanations.
      virtual bool CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
      //!
      //! Override this to enable/disable the chaos drawing stage.
      virtual bool ChaosStageIsEnabled() const;
      //!
      //! Override this to control the starting level in which the chaos stage will started to be drawn.
      virtual uint32_t ChaosStartLevelThreshold() const;
      //!
      //! Override this to control the ending level in which the cosmos stage will stopped from being drawn.
      virtual uint32_t ChaosEndLevelThreshold() const;
      //!
      //! Override this callback to control the color of the seed and replicas drawn in the chaos stage.\n
      //! See the page \ref More_Info_About_CSeedReplicatorFractal "More Information AboutCSeedReplicatorFractal Class" for more explanations.
      virtual bool ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
      //!
      //! Override this to change the chaos factor.
      virtual uint8_t ChaosFactor() const;
      //!
      //! Override this to change the number of chaos iteration.
      virtual uint8_t ChaosIterationLimit() const;

      // Drawing functions
      //! Cosmos drawing stage implementation for seed-replicator fractals.\n
      //! \n
      //! \note All coordinates are in standard screen coordinates.
      void DrawImpl(_TAffineStorage *ptDstAStorage, uint32_t un32StartLT, uint32_t un32PolygonEndLT,
                     int32_t n32XStart, int32_t n32YStart, const _TAffine *pcsCurA,
                     uint32_t un32MaxLevel, uint32_t un32CurLevel = 0, uint32_t un32CurPart = 0);
      //! Chaos drawing stage implementation for seed-replicator fractals.\n
      //! \n
      //! \note All coordinates are in standard screen coordinates.
      void PaintImpl(int32_t n32XStart, int32_t n32YStart, _TAffineStorage *ptDstAStorage,
                      uint32_t un32StartLT, uint32_t un32EndLT, uint8_t un8ChaosFactor, uint8_t un8ChaosIter);

      // Data
      _TSeed        _tSeed;    // The seed
      _TSTransform *_ptTrns;   // Transformations array
      uint32_t      _un32TCnt; // Number of transformations in the array = number of seed replicas
  };

} // namespace vge
#endif


/**
\page More_Info_About_CSeedReplicatorFractal More Information About CSeedReplicatorFractal Class

The member functions CosmosUpdateColor(), ChaosUpdateColor() and
DrawingProgressCallback() have these parameters :
  - un32MaxLevel,
  - un32CurLevel,
  - un32TotPart,
  - un32CurPart,
  - un32TotRelPart, and
  - un32CurRelPart

\n
See the ASCII art picture below and assume that it is a tree fractal :
<PRE>
    \ /   \ /   \ /   \ /   \ /   \ /   \ /   \ /
    H|    I|    J|    K|    L|    M|    N|    O|
     |     |     |     |     |     |     |     | .............. Level 3
      \   /       \   /       \   /       \   /
       \ /         \ /         \ /         \ /
        |           |           |           |
       D|          E|          F|          G|
        |           |           |           |
        |           |           |           | ................. Level 2
         \         /             \         /
           \     /                 \     /
             \ /                     \ /
              |                       |
             B|                      C|
              |                       |
              |                       |
              |                       |
              |                       | ....................... Level 1
               \                     /
                 \                 /
                   \             /
                     \         /
                       \     /
                         \ /
                          |
                         A|
                          |
                          |
                          |
                          |
                          |
                          | ................................... Level 0
</PRE>
In the picture above, level 0 is the seed. A replica of the seed is put
on each branch of the seeed. In this example, the fractal is drawn up to
level 3.
\n
\n
Before drawing each part (seed and the seed replicas), the callback
function is called. The parameters of the above fractal will be :
<PRE>
    +--------+----------+----------+---------+---------+------------+------------+
    | Branch | MaxLevel | CurLevel | MaxPart | CurPart | MaxRelPart | CurRelPart |
    +--------+----------+----------+---------+---------+------------+------------+
    |   A    |    3     |    0     |    1    |    0    |     1      |     0      |
    +--------+----------+----------+---------+---------+------------+------------+
    |   B    |    3     |    1     |    2    |    0    |     2      |     0      |
    |   C    |    3     |    1     |    2    |    1    |     2      |     1      |
    +--------+----------+----------+---------+---------+------------+------------+
    |   D    |    3     |    2     |    4    |    0    |     2      |     0      |
    |   E    |    3     |    2     |    4    |    1    |     2      |     1      |
    |   F    |    3     |    2     |    4    |    2    |     2      |     0      |
    |   G    |    3     |    2     |    4    |    3    |     2      |     1      |
    +--------+----------+----------+---------+---------+------------+------------+
    |   H    |    3     |    3     |    8    |    0    |     2      |     0      |
    |   I    |    3     |    3     |    8    |    1    |     2      |     1      |
    |   J    |    3     |    3     |    8    |    2    |     2      |     0      |
    |   K    |    3     |    3     |    8    |    3    |     2      |     1      |
    |   L    |    3     |    3     |    8    |    4    |     2      |     0      |
    |   M    |    3     |    3     |    8    |    5    |     2      |     1      |
    |   N    |    3     |    3     |    8    |    6    |     2      |     0      |
    |   O    |    3     |    3     |    8    |    7    |     2      |     1      |
    +--------+----------+----------+---------+---------+------------+------------+
</PRE>
As seen in the table, there are two values that always constants:
  -# The "MaxLevel" which always contains 3. This value is came from the
     return value of the virtual member function, TotalReplicationLevel().
  -# The "MaxRelPart" which always contains 2 (except for branch A, the seed
     itself). This value is came from the member variable "_un32TCnt" of
     the class.

The value of "MaxPart" can be calculated manually using this formula :\n
<span style="margin-left:2em;"><b>
MaxPart = (MaxRelPart)<sup>CurLevel</sup>
</b></span>\n
with the ranges :
<PRE>
  CurLevel   : from 0 to MaxLevel
  CurPart    : from 0 to (MaxPart - 1)
  CurRelPart : from 0 to (MaxRelPart - 1)
</PRE>

\n
Returning false from DrawingProgressCallback() will make the current stage
stopped.
\n
\n
The value returned by the _cosmosPolygonLevelThreshold() virtual member function
will be used to determine whether the seed replicas will be drawn using lines or
just a dot.
  -# If the value of "CurLevel" is smaller than the value returned by this
     function then the replicas will be drawn using polygon (lines).
  -# If the value of "CurLevel" is bigger than the value returned by this
     function then the replicas will be drawn using just a dot.

Cosmos (finite) stage is drawn using the DrawImpl() member function.
These parameter-control functions can be used to control this stage:
  - CosmosStageIsEnabled(), returning false will make the cosmos stage skipped.
  - CosmosStartLevelThreshold()
  - _cosmosPolygonLevelThreshold()
  - CosmosUpdateColor(), returning false will make the current part skipped.
  - TotalReplicationLevel()

Chaos(infinite) stage is drawn using the PaintImpl() member function.
These parameter-control functions can be used to control this stage:
  - ChaosStageIsEnabled(), returning false will make the chaos stage skipped.
  - ChaosStartLevelThreshold()
  - ChaosEndLevelThreshold()
  - ChaosUpdateColor(), returning false will make the current part skipped.
  - ChaosFactor()
  - ChaosIterationLimit()
  - TotalReplicationLevel()

Note that except CosmosUpdateColor() and ChaosUpdateColor() which will
be called in each iteration, the other functions will be called only once.
\n
<hr>
Back to vge::CSeedReplicatorFractal.
*/
