// ############################################################################################
//  ImageBuilder2.cc
//  Implementation of CImageBuilder class (2nd part)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "ImageBuilder.h"


namespace vge {

//
// External image helper functions (defined inside "lib/2DManioulation/ImageBuilder1.cc"
//
extern void _OrderedDitherRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height);
extern void _ClipRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height);


//
// General effect functions
//
void CImageBuilder::SmearEffect(uint8_t un8Radius)
{
  // Temporary variables
  TUN8RGB *ptNew;
  uint32_t       un32CntX, un32CntY, un32PosD, un32PosS;
  int32_t        n32CntX, n32CntY;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if(!un8Radius)
     VError(VERR_InvalidParameter, _NFL_, "The given radius must not zero");

  // Allocate memory for new image
  ptNew = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * _un32Width * _un32Height));
  if(ptNew == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  // Smear image
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32PosD = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      // Calculate position
      n32CntX = int32_t(un32CntX) + int32_t(float(random()) / float(RAND_MAX) * float(un8Radius * 2) - float(un8Radius));
      n32CntY = int32_t(un32CntY) + int32_t(float(random()) / float(RAND_MAX) * float(un8Radius * 2) - float(un8Radius));
      // Copy the color into buffer
      if((n32CntY >= 0) && (n32CntX >= 0) &&
      (n32CntY < int32_t(_un32Height)) && (n32CntX < int32_t(_un32Width))) {
        un32PosS = uint32_t(n32CntY) * _un32Width;
        ptNew[un32PosD + un32CntX].un8R = _ptBuffer[un32PosS + n32CntX].un8R;
        ptNew[un32PosD + un32CntX].un8G = _ptBuffer[un32PosS + n32CntX].un8G;
        ptNew[un32PosD + un32CntX].un8B = _ptBuffer[un32PosS + n32CntX].un8B;
      }
      else {
        un32PosS = uint32_t(un32CntY) * _un32Width;
        ptNew[un32PosD + un32CntX].un8R = _ptBuffer[un32PosS + un32CntX].un8R;
        ptNew[un32PosD + un32CntX].un8G = _ptBuffer[un32PosS + un32CntX].un8G;
        ptNew[un32PosD + un32CntX].un8B = _ptBuffer[un32PosS + un32CntX].un8B;
      }
    }
  }

  // Change pointer
  free(_ptBuffer);
  _ptBuffer = ptNew;
}

void CImageBuilder::BlurEffect(uint8_t un8Radius, bool bUseDithering)
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if(!un8Radius)
     VError(VERR_InvalidParameter, _NFL_, "The given radius must not zero");

  // Allocate memory for new image
  TUN8RGB *ptNew = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * _un32Width * _un32Height));
  if(ptNew == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  // Blur image
  for(uint32_t yc = 0; yc < _un32Height; yc++) {
    const uint32_t posD = yc * _un32Width;
    for(uint32_t xc = 0; xc < _un32Width; xc++) {
      // Grab pixels
      uint32_t r = 0;
      uint32_t g = 0;
      uint32_t b = 0;
      uint32_t t = 0;
      for(int32_t yg = int32_t(yc) - int32_t(un8Radius); yg <= int32_t(yc + un8Radius); yg++) {
        const uint32_t posS = uint32_t(yg) * _un32Width;
        for(int32_t xg = int32_t(xc) - int32_t(un8Radius); xg <= int32_t(xc + un8Radius); xg++) {
          if((yg >= 0) && (xg >= 0) &&
             (yg < int32_t(_un32Height)) && (xg < int32_t(_un32Width))) {
            r += (_ptBuffer[posS + xg].un8R << 3);
            g += (_ptBuffer[posS + xg].un8G << 3);
            b += (_ptBuffer[posS + xg].un8B << 3);
            t++;
          }
        }
      }
      // Put the mean colors into buffer
      ptNew[posD + xc].un8R = uint8_t(r / t);
      ptNew[posD + xc].un8G = uint8_t(g / t);
      ptNew[posD + xc].un8B = uint8_t(b / t);
    }
  }

  // Dither or clip the color values
  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(ptNew, _un32Width, _un32Height);
  else
    _ClipRGB24ToVRGB_(ptNew, _un32Width, _un32Height);

  // Change pointer
  free(_ptBuffer);
  _ptBuffer = ptNew;
}

void CImageBuilder::OilPaintingEffect(uint8_t un8Radius)
{
  // Temporary variables
  TUN8RGB  *ptNew;
  uint32_t  un32CntX, un32CntY, un32PosD, un32PosS;
  int32_t   n32CntX, n32CntY;
  uint8_t   un8R[128], un8G[128], un8B[128], un8C[128], un8Pos, un8Tot, un8Cnt;
  bool      bF;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if(!un8Radius || un8Radius>5)
     VError(VERR_InvalidParameter, _NFL_, "The given radius must between 1 and 5");

  // Allocate memory for new image
  ptNew = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * _un32Width * _un32Height));
  if(ptNew == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  // Perform the effect
  for(un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
    un32PosD = un32CntY * _un32Width;
    for(un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
      // Grab pixels
      un8Tot = 0;
      for(n32CntY = int32_t(un32CntY) - int32_t(un8Radius); n32CntY <= int32_t(un32CntY + un8Radius); n32CntY++) {
        un32PosS = uint32_t(n32CntY) * _un32Width;
        for(n32CntX = int32_t(un32CntX) - int32_t(un8Radius); n32CntX <= int32_t(un32CntX + un8Radius); n32CntX++) {
          if((n32CntY >= 0) && (n32CntX >= 0) &&
          (n32CntY < int32_t(_un32Height)) && (n32CntX < int32_t(_un32Width))) {
            // Check if the pixel already in the list
            bF = false;
            for(un8Cnt = 0; un8Cnt < un8Tot; un8Cnt++) {
              if((un8R[un8Cnt] == _ptBuffer[un32PosS + n32CntX].un8R) &&
                (un8G[un8Cnt] == _ptBuffer[un32PosS + n32CntX].un8G) &&
              (un8B[un8Cnt] == _ptBuffer[un32PosS + n32CntX].un8B)) {
                un8C[un8Cnt]++;
                bF = true;
                break;
              }
            }
            // Add the pixel to the list
            if(!bF) {
              un8R[un8Tot] = _ptBuffer[un32PosS + n32CntX].un8R;
              un8G[un8Tot] = _ptBuffer[un32PosS + n32CntX].un8G;
              un8B[un8Tot] = _ptBuffer[un32PosS + n32CntX].un8B;
              un8C[un8Tot] = 1;
              un8Tot++;
            }
          }
        }
      }
      // Find the modulus of the color
      un8Pos = 0;
      for(un8Cnt = 1; un8Cnt < un8Tot; un8Cnt++) if(un8C[un8Cnt] > un8C[un8Pos]) un8Pos = un8Cnt;
      // Put the modulus colors into buffer
      ptNew[un32PosD + un32CntX].un8R = un8R[un8Pos];
      ptNew[un32PosD + un32CntX].un8G = un8G[un8Pos];
      ptNew[un32PosD + un32CntX].un8B = un8B[un8Pos];
    }
  }

  // Change pointer
  free(_ptBuffer);
  _ptBuffer = ptNew;
}

void CImageBuilder::AdjustBrightness(float fMultFactor, int8_t un8AddFactor,
                                     bool bAffectRed, bool bAffectGreen, bool bAffectBlue)
{
  // Temporary variables
  int32_t   n32Tmp;
  uint32_t  un32Cnt;
  TUN8RGB  *ptTmp = _ptBuffer;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ image");

  // Adjust brightness
  for(un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    if(bAffectRed) {
      n32Tmp = int32_t(float(ptTmp->un8R) * fMultFactor + float(un8AddFactor));
      if(n32Tmp < 0) n32Tmp = 0;
      if(n32Tmp > MAX_COLOR_VALUE) n32Tmp = MAX_COLOR_VALUE;
      ptTmp->un8R = uint8_t(n32Tmp);
    }
    if(bAffectGreen) {
      n32Tmp = int32_t(float(ptTmp->un8G) * fMultFactor + float(un8AddFactor));
      if(n32Tmp < 0) n32Tmp = 0;
      if(n32Tmp > MAX_COLOR_VALUE) n32Tmp = MAX_COLOR_VALUE;
      ptTmp->un8G = uint8_t(n32Tmp);
    }
    if(bAffectBlue) {
      n32Tmp = int32_t(float(ptTmp->un8B) * fMultFactor + float(un8AddFactor));
      if(n32Tmp < 0) n32Tmp = 0;
      if(n32Tmp > MAX_COLOR_VALUE) n32Tmp = MAX_COLOR_VALUE;
      ptTmp->un8B = uint8_t(n32Tmp);
    }
    ptTmp++;
  }

}

void CImageBuilder::Grayscale(bool bUseLinearEquation)
{
  // Temporary variables
  uint32_t un32Cnt, un32R, un32G, un32B;
  TUN8RGB *ptTmp = _ptBuffer;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ image");

  // Cconvert to grayscale
  for(un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    un32R = ptTmp->un8R;
    un32G = ptTmp->un8G;
    un32B = ptTmp->un8B;
    if(bUseLinearEquation)
      un32R = (un32R + un32G + un32B) / 3;
    else
      un32R = ((un32R * 3) + (un32G * 5) + (un32B * 2)) / 10;
    ptTmp->un8R = un32R;
    ptTmp->un8G = un32R;
    ptTmp->un8B = un32R;
    ptTmp++;
  }
}


//
// Image resizing functions
//
void CImageBuilder::RoughResize(uint32_t un32NewWidth, uint32_t un32NewHeight)
{
  // Temporary variables
  TUN8RGB  *ptNew, *ptOld = _ptBuffer;
  uint32_t  un32X, un32Y, un32S, un32D;
  float     fRX, fRY, fX, fY;

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if((un32NewWidth == _un32Width) && (un32NewHeight == _un32Height)) return;

  // Determine size
  if(un32NewWidth) {
    if(un32NewHeight) {
      if((un32NewWidth < 2) || (un32NewHeight < 2))
        VError(VERR_InvalidParameter, _NFL_, "The new image size too small, minimum width and height are 2 pixels");
    }
    else
      un32NewHeight = uint32_t(float(_un32Height) / float(_un32Width) * float(un32NewWidth));
  }
  else {
    if(un32NewHeight)
      un32NewWidth = uint32_t(float(_un32Width) / float(_un32Height) * float(un32NewHeight));
    else
      VError(VERR_InvalidParameter, _NFL_, "Invalid new image size");
  }

  // Allocate memory for the new image
  ptNew = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * un32NewWidth * un32NewHeight));
  if(ptNew == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image resizing buffer");

  // Calculate image ratio
  fRX  = float(_un32Width) / float(un32NewWidth);
  fRY  = float(_un32Height) / float(un32NewHeight);

  // Do rough scalling
  fY = 0;
  for(un32Y = 0; un32Y < un32NewHeight; un32Y++) {
    un32D = un32Y * un32NewWidth;
    un32S = int32_t(fY) * _un32Width;
    fX = 0;
    for(un32X = 0; un32X < un32NewWidth; un32X++) {
      ptNew[un32D + un32X].un8R = ptOld[un32S + int32_t(fX)].un8R;
      ptNew[un32D + un32X].un8G = ptOld[un32S + int32_t(fX)].un8G;
      ptNew[un32D + un32X].un8B = ptOld[un32S + int32_t(fX)].un8B;
      fX += fRX;
    }
    fY += fRY;
  }

  // Destroy old image
  free(_ptBuffer);

  // Copy variables
  _ptBuffer   = ptNew;
  _un32Width  = un32NewWidth;
  _un32Height = un32NewHeight;
}

void CImageBuilder::Resize(uint32_t un32NewWidth, uint32_t un32NewHeight, bool bUseDithering)
{
  // Helper class for cubic scale
  class CubicScaleHelper {
    private:
      TUN8RGB* _sb;
      int32_t  _sw;
      int32_t  _sh;

      inline float cc1D(float f0, float f1, float f2, float f3, float ds)
      {
        float d1 = ds;
        float d2 = d1 * ds;
        float d3 = d2 * ds;
        return( (    -f0 +      f1 - f2 + f3)*d3 +
                (2.0f*f0 - 2.0f*f1 + f2 - f3)*d2 +
                (    -f0           + f2     )*d1 + f1);
      }

    public:    
      inline CubicScaleHelper(TUN8RGB* srcBuff, int32_t srcWidth, int32_t srcHeight)
      : _sb(srcBuff), _sw(srcWidth), _sh(srcHeight)
      {}
      
      inline void grabPixel(float dv[3], int32_t sy, int32_t sx)
      {
        if(sx < 0)    sx = 0;
        if(sy < 0)    sy = 0;
        if(sx >= _sw) sx = _sw - 1;
        if(sy >= _sh) sy = _sh - 1;
        dv[0] = _sb[sy*_sw + sx].un8R << 3;
        dv[1] = _sb[sy*_sw + sx].un8G << 3;
        dv[2] = _sb[sy*_sw + sx].un8B << 3;
      }
      
      inline void cubicConvolution(float dv[3], float sv[4][3], float ds)
      {
        dv[0] = cc1D(sv[0][0], sv[1][0], sv[2][0], sv[3][0], ds);
        dv[1] = cc1D(sv[0][1], sv[1][1], sv[2][1], sv[3][1], ds);
        dv[2] = cc1D(sv[0][2], sv[1][2], sv[2][2], sv[3][2], ds);
      }
  };

  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be processed (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) && (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_InvalidCall, _NFL_, "Can't process a non _IMG_TYPE_RGB5_ or _IMG_TYPE_FYUV_ image");
  if((un32NewWidth == _un32Width) && (un32NewHeight == _un32Height)) return;

  // Determine the new size
  if(un32NewWidth) {
    if(un32NewHeight) {
      if((un32NewWidth < 2) || (un32NewHeight < 2))
        VError(VERR_InvalidParameter, _NFL_, "The new image size too small, minimum width and height are 2 pixels");
    }
    else
      un32NewHeight = uint32_t(float(_un32Height) / float(_un32Width) * float(un32NewWidth));
  }
  else {
    if(un32NewHeight)
      un32NewWidth = uint32_t(float(_un32Width) / float(_un32Height) * float(un32NewHeight));
    else
      VError(VERR_InvalidParameter, _NFL_, "Invalid new image size");
  }

  // Allocate memory for the new image
  TUN8RGB *ptNew = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * un32NewWidth * un32NewHeight));
  if(ptNew == NULL) VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image resizing buffer");

  // Calculate the scalling ratio
  float rx = float(_un32Width) / float(un32NewWidth);
  float ry = float(_un32Height) / float(un32NewHeight);

  // Cubic scale the image
  CubicScaleHelper csh(_ptBuffer, _un32Width, _un32Height);
  for(int32_t y = 0; y < int32_t(un32NewHeight); y++) {
    float   fy = float(y) * ry;
    int32_t ny = int32_t(fy);
    float   dy = fy - float(ny);
    
    for(int32_t x = 0; x < int32_t(un32NewWidth); x++) {
      float   fx = float(x) * rx;
      int32_t nx = int32_t(fx);
      float   dx = fx - float(nx);
      
      // Grab the source pixels, like this :
      //      0 1 2 3 (X)
      //
      //   0  N N N N
      //   1  N P N N
      //   2  N N N N
      //   3  N N N N
      //  (Y)
      //
      // P is the current pixel to be processed, N are neighboring pixels
      //
      float pixels[4][4][3];
      for(int cy = 0; cy < 4; cy++) {
         for(int cx = 0; cx < 4; cx++)
           csh.grabPixel(pixels[cy][cx], ny+(cy-1), nx+(cx-1));
      }
      
      // Perform cubic convolution on the horizontal direction
      float tmp[4][3];
      for(int i = 0; i < 4; i++)
        csh.cubicConvolution(tmp[i], pixels[i], dx);

      // Perform cubic convolution on the vertical direction
      float res[3];
      csh.cubicConvolution(res, tmp, dy);

      // Put the destination pixel
      ptNew[y*un32NewWidth + x].un8R = uint8_t(clip_value(0.0f, res[0], 255.0f));
      ptNew[y*un32NewWidth + x].un8G = uint8_t(clip_value(0.0f, res[1], 255.0f));
      ptNew[y*un32NewWidth + x].un8B = uint8_t(clip_value(0.0f, res[2], 255.0f));
    }
  }

  // Destroy old image
  free(_ptBuffer);

  // Dither or clip the color values
  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(ptNew, un32NewWidth, un32NewHeight);
  else
    _ClipRGB24ToVRGB_(ptNew, un32NewWidth, un32NewHeight);

  // Copy variables
  _ptBuffer   = ptNew;
  _un32Width  = un32NewWidth;
  _un32Height = un32NewHeight;
}

void CImageBuilder::RoughResizeToFitBackground(C3DEnv *pcsV3DEnv)
{
  // Use the supplied 3d environment class
  if(pcsV3DEnv)
    return(RoughResize(pcsV3DEnv->_tRB->un32Width, pcsV3DEnv->_tRB->un32Height));

  // Use the active 3D environment
  if(C3DEnv::Active3DEnv())
    return(RoughResize(C3DEnv::Active3DEnv()->_tRB->un32Width, C3DEnv::Active3DEnv()->_tRB->un32Height));

  // Error
  VError(VERR_NotFound, _NFL_, "No 3D environment can be used as reference for resizing image");
}

void CImageBuilder::ResizeToFitBackground(C3DEnv *pcsV3DEnv, bool bUseDithering)
{
  // Use the supplied 3d environment class
  if(pcsV3DEnv)
    return(Resize(pcsV3DEnv->_tRB->un32Width, pcsV3DEnv->_tRB->un32Height, bUseDithering));

  // Use the active 3D environment
  if(C3DEnv::Active3DEnv())
    return(Resize(C3DEnv::Active3DEnv()->_tRB->un32Width, C3DEnv::Active3DEnv()->_tRB->un32Height, bUseDithering));

  // Error
  VError(VERR_NotFound, _NFL_, "No 3D environment can be used as reference for resizing image");
}

void CImageBuilder::RoughResizeToFitTexture(uint32_t un32TexWidth, uint32_t un32TexHeight)
{
  _CalcTextureSize_(un32TexWidth, un32TexHeight);
  return(RoughResize(un32TexWidth, un32TexHeight));
}

void CImageBuilder::ResizeToFitTexture(uint32_t un32TexWidth, uint32_t un32TexHeight, bool bUseDithering)
{
  _CalcTextureSize_(un32TexWidth, un32TexHeight);
  return(Resize(un32TexWidth, un32TexHeight, bUseDithering));
}

void CImageBuilder::_CalcTextureSize_(uint32_t &un32TexWidth, uint32_t &un32TexHeight)
{
  // Determine the texture width
  if(un32TexWidth) {
    if((un32TexWidth !=  16) && (un32TexWidth !=  32) &&
       (un32TexWidth !=  64) && (un32TexWidth != 128) && 
       (un32TexWidth != 256) && (un32TexWidth != 512))
      VError(VERR_InvalidParameter, _NFL_, "Texture width must be one of 16/32/64/128/256/512 pixels");
  }
  else {
    if(_un32Width > 256)
      un32TexWidth = 512;
    else if(_un32Width > 128)
      un32TexWidth = 256;
    else if(_un32Width > 64)
      un32TexWidth = 128;
    else if(_un32Width > 32)
      un32TexWidth = 64;
    else if(_un32Width > 16)
      un32TexWidth = 32;
    else
      un32TexWidth = 16;
  }

  // Determine the texture height
  if(un32TexHeight) {
    if((un32TexHeight < 16) || (un32TexHeight > 512))
      VError(VERR_InvalidParameter, _NFL_, "Texture height must between 16 to 512 pixels");
  }
  else {
    un32TexHeight = uint32_t(float(_un32Height) / float(_un32Width) * float(un32TexWidth));
    if(un32TexHeight < 16)  un32TexHeight = 16;
    if(un32TexHeight > 512) un32TexHeight = 512;
  }
}

} // namespace vge
