// ############################################################################################
//  ImageBuilder5.cc
//  Implementation of CImageBuilder class (5th part)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../VideoEnvironment/VideoEnvironment.h"
#include "BitmapStructs.h"
#include "ImageBuilder.h"

// C++ headers
#include <fstream>

// Image libraries headers
#ifdef __USE_LIBPNG__
#include <png.h>
#endif

#ifdef __USE_LIBJPEG__
extern "C" { // Some buggy libjpeg headers need this to be useable under C++
#include <jerror.h>
#include <jpeglib.h>
}
#endif

#ifdef __USE_LIBTIFF__
#include <tiffio.h>
#endif

#ifdef __USE_LIBUNGIF__
#include <gif_lib.h>
#endif

#include <zlib.h>


namespace vge {

//
// External image helper functions (defined inside "lib/2DManioulation/ImageBuilder1.cc"
//
extern void _OrderedDitherRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height);
extern void _ClipRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height);


//
// PNG Loading
//
#ifdef __USE_LIBPNG__

static void _pngReadBytesCallback(png_struct* ps, png_byte* pb, png_size_t pz)
{
  istream &is = reinterpret_cast<const CDataSource*>(ps->io_ptr)->Stream();
  is.read(reinterpret_cast<char*>(pb), pz);
}

void CImageBuilder::LoadPNG(const CDataSource &source, bool bUseDithering, XColor15 backgroundColor, float screenGamma)
{
  // Get the stream name
  const char *strmName = source.Name().c_str();

  // Prepare the needed structures
  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if(!png_ptr)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate PNG read structure");
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if(!info_ptr) {
    png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate PNG information structure");
  }
  png_infop end_info = png_create_info_struct(png_ptr);
  if(!end_info) {
    png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate PNG information structure");
  }

  // Set the error return point
  if(setjmp(png_jmpbuf(png_ptr)))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    VError(VERR_ReadStream, _NFL_, "Error when loading PNG stream \"%s\"", strmName);
  }

  // Init the libpng I/O
  png_set_read_fn(png_ptr, reinterpret_cast<void*>(const_cast<CDataSource*>(&source)), _pngReadBytesCallback);

  // Read and parse the header
  png_read_info(png_ptr, info_ptr);

  png_uint_32 width, height;
  int32_t bit_depth, color_type, interlace_type;
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
               &interlace_type, int_p_NULL, int_p_NULL);

  // These lines should make sure that whatever the input file format is,
  // libpng will always generate RGB or ARGB output dta
  png_set_packing(png_ptr);
  png_set_gray_to_rgb(png_ptr);
  png_set_expand(png_ptr);
  png_set_strip_16(png_ptr);
  if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) png_set_tRNS_to_alpha(png_ptr);

  // Try to set the background color if the user wants it
  // (only if the file itself specify the background region)
  if(backgroundColor != INVALID_COLOR15_VALUE) {
    uint8_t r, g, b;
    V_RGB_Extract(r, g, b, backgroundColor);

    png_color_16 my_background;
    memset(&my_background, 0, sizeof(my_background));
    my_background.red   = r << 3;
    my_background.green = g << 3;
    my_background.blue  = b << 3;
    png_set_background(png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
  }
  // Try to set the background color using the color specified in the file
  // or just use black if not available
  else {
    png_color_16 *image_background;
    png_color_16 my_background;
    memset(&my_background, 0, sizeof(my_background));
    if(png_get_bKGD(png_ptr, info_ptr, &image_background))
      png_set_background(png_ptr, image_background, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
    else
      png_set_background(png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
  }

  // Adjust the gama if the user wants it
  if(screenGamma > 0.0) {
    double gamma;
    if (png_get_gAMA(png_ptr, info_ptr, &gamma))
      png_set_gamma(png_ptr, screenGamma, gamma);
    else
      png_set_gamma(png_ptr, screenGamma, 0.45455);
  }

  // Make sure that the format is ARGB
  png_set_swap_alpha(png_ptr);

  // Make sure libpng handle deinterlacing automatically
  png_set_interlace_handling(png_ptr);

  // Update the info struct
  png_read_update_info(png_ptr, info_ptr);

  // Allocate memory for PNG buffer
  png_bytep row_pointers[height];
  for (uint32_t row = 0; row < height; row++)
    row_pointers[row] = reinterpret_cast<png_byte*>(png_malloc(png_ptr, png_get_rowbytes(png_ptr, info_ptr)));

  // Read the image data
  png_read_image(png_ptr, row_pointers);
  png_read_end(png_ptr, info_ptr);

  // Allocate memory for new image buffer
  TUN8RGB *tmpBuff = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * width * height));
  if(!tmpBuff) {
    for(uint32_t row = 0; row < height; row++) png_free(png_ptr, row_pointers[row]);
      png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");
  }

  // Parse the image
  TUN8RGB *dst = tmpBuff;

  for(uint32_t y = 0; y < height; y++) {
    png_bytep src = row_pointers[y];

    for(uint32_t x = 0; x < width; x++) {
      uint8_t r = *src++;
      uint8_t g = *src++;
      uint8_t b = *src++;

       dst->un8R = r;
       dst->un8G = g;
       dst->un8B = b;
       dst++;
    }
  }

  // Perform dithering (or clipping)
  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(tmpBuff, width, height);
  else
    _ClipRGB24ToVRGB_(tmpBuff, width, height);

  // Destroy old image, if any
  FreeImage();

  // Set variables to the new image
  _un32Width  = width;
  _un32Height = height;
  _ptBuffer   = tmpBuff;

  // Perform cleanups
  for(uint32_t row = 0; row < height; row++) png_free(png_ptr, row_pointers[row]);
  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
}

#else

void CImageBuilder::LoadPNG(const CDataSource &source, bool bUseDithering, XColor15 backgroundColor, float screenGamma)
{
  UNUSED_PARAMS_4(source, bUseDithering, backgroundColor, screenGamma);
  VError(VERR_NotSupported, _NFL_, "PNG loading support is disabled in configuration");
}

#endif


//
// JPEG Loading
//
#ifdef __USE_LIBJPEG__

#define _JPEG_INPUT_BUFF_SIZE 4096

struct _vgeJpegErrorMgr {
  jpeg_error_mgr pub;
  jmp_buf        setjmp_buffer;
};

struct _vgeJpegSourceMgr {
  jpeg_source_mgr    pub;
  const CDataSource *instream;      // Source stream
  JOCTET            *buffer;        // Start of buffer
  boolean            start_of_file; // Have we gotten any data yet?
};

static void _jpegErrorExit(j_common_ptr cinfo) _attr_noreturn;
static void _jpegErrorExit(j_common_ptr cinfo)
{
  _vgeJpegErrorMgr *myerr = reinterpret_cast<_vgeJpegErrorMgr*>(cinfo->err);
  (*cinfo->err->output_message)(cinfo);
  longjmp(myerr->setjmp_buffer, 1);
}

static void _jpegInitSource(j_decompress_ptr cinfo)
{
  _vgeJpegSourceMgr *src = reinterpret_cast<_vgeJpegSourceMgr*>(cinfo->src);
  src->start_of_file = TRUE;
}

static boolean _jpegFillInputBuffer(j_decompress_ptr cinfo)
{
  _vgeJpegSourceMgr *src = reinterpret_cast<_vgeJpegSourceMgr*>(cinfo->src);

  istream &is = src->instream->Stream();
  is.readsome(reinterpret_cast<char*>(src->buffer), _JPEG_INPUT_BUFF_SIZE * sizeof(JOCTET));

  size_t nbytes = is.gcount();

  if(nbytes <= 0) {
    // Treat empty input file as fatal error
    if(src->start_of_file) ERREXIT(cinfo, JERR_INPUT_EMPTY);
    WARNMS(cinfo, JWRN_JPEG_EOF);
    // Insert a fake EOI marker
    src->buffer[0] = static_cast<JOCTET>(0xFF);
    src->buffer[1] = static_cast<JOCTET>(JPEG_EOI);
    nbytes = 2;
  }

  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->start_of_file       = FALSE;

  return(TRUE);
}

static void _jpegSkipInputData(j_decompress_ptr cinfo, long num_bytes)
{
  _vgeJpegSourceMgr *src = reinterpret_cast<_vgeJpegSourceMgr*>(cinfo->src);

  if(num_bytes > 0) {
    while(num_bytes > long(src->pub.bytes_in_buffer)) {
      num_bytes -= long(src->pub.bytes_in_buffer);
      _jpegFillInputBuffer(cinfo);
    }
    src->pub.next_input_byte += size_t(num_bytes);
    src->pub.bytes_in_buffer -= size_t(num_bytes);
  }
}

static void _jpegTermSource(j_decompress_ptr cinfo)
{
  UNUSED_PARAM(cinfo);
}

static void _jpegStreamSrc(j_decompress_ptr cinfo, const CDataSource *instream)
{
  _vgeJpegSourceMgr *src;

  if(cinfo->src == NULL) {
    cinfo->src = reinterpret_cast<struct jpeg_source_mgr*>((*cinfo->mem->alloc_small)(reinterpret_cast<j_common_ptr>(cinfo), JPOOL_PERMANENT, sizeof(_vgeJpegSourceMgr)));
    src = reinterpret_cast<_vgeJpegSourceMgr*>(cinfo->src);
    src->buffer = reinterpret_cast<JOCTET*>((*cinfo->mem->alloc_small)(reinterpret_cast<j_common_ptr>(cinfo), JPOOL_PERMANENT, _JPEG_INPUT_BUFF_SIZE * sizeof(JOCTET)));
  }
  src = reinterpret_cast<_vgeJpegSourceMgr*>(cinfo->src);

  src->pub.init_source       = _jpegInitSource;
  src->pub.fill_input_buffer = _jpegFillInputBuffer;
  src->pub.skip_input_data   = _jpegSkipInputData;
  src->pub.resync_to_restart = jpeg_resync_to_restart; // Use the default method from libjpeg
  src->pub.term_source       = _jpegTermSource;
  src->instream              = instream;
  src->pub.bytes_in_buffer   = 0;    // Uorces fill_input_buffer on first read
  src->pub.next_input_byte   = NULL; // Until buffer loaded
}

void CImageBuilder::LoadJPG(const CDataSource &source, bool bUseDithering)
{
  // Get the stream name
  const char *strmName = source.Name().c_str();

  // Variables needed in decompression
  jpeg_decompress_struct cinfo;
  _vgeJpegErrorMgr       jerr;
  TUN8RGB               *tmpBuff = NULL;

  // Initialize the error handler
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit =  _jpegErrorExit;

  if (setjmp(jerr.setjmp_buffer)) {
    jpeg_destroy_decompress(&cinfo);
    if(tmpBuff) free(tmpBuff);
    VError(VERR_ReadStream, _NFL_, "Error when loading JPEG stream \"%s\"", strmName);
  }

  // Prepare decompression
  jpeg_create_decompress(&cinfo);
 _jpegStreamSrc(&cinfo, &source);
  jpeg_read_header(&cinfo, TRUE);
  jpeg_start_decompress(&cinfo);

  // Get some information
  uint32_t row_stride = cinfo.output_width * cinfo.output_components;
  uint32_t width      = cinfo.output_width;
  uint32_t height     = cinfo.output_height;

  // Allocate memory for new image buffer
  tmpBuff = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * width * height));
  if(!tmpBuff) {
    jpeg_destroy_decompress(&cinfo);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");
  }

  // Convert the image
  bool       error  = false;
  JSAMPARRAY buffer = (*cinfo.mem->alloc_sarray)(reinterpret_cast<j_common_ptr>(&cinfo), JPOOL_IMAGE, row_stride, 1);

  if((row_stride % width) == 0) {
    uint32_t         y = 0;
    TUN8RGB *dst = tmpBuff;

    while (cinfo.output_scanline < cinfo.output_height) {
      jpeg_read_scanlines(&cinfo, buffer, 1);

      u_char *src = buffer[0];

      for(uint32_t x = 0; x < width; x ++) {
        uint8_t r = *src++;
        uint8_t g = *src++;
        uint8_t b = *src++;

        dst->un8R = r;
        dst->un8G = g;
        dst->un8B = b;
        dst++;
      }

      y++;
    }

  }
  else
    error = true;

  // Perform dithering (or clipping)
  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(tmpBuff, width, cinfo.output_height);
  else
    _ClipRGB24ToVRGB_(tmpBuff, width, cinfo.output_height);

  // Cleanup
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  // Check for error
  if(error) {
    if(tmpBuff) free(tmpBuff);
    VError(VERR_NotSupported, _NFL_, "The stream \"%s\" has an unsupported format", strmName);
  }

  // Destroy old image, if any
  FreeImage();

  // Set variables to the new image
  _un32Width  = width;
  _un32Height = height;
  _ptBuffer   = tmpBuff;
}

#else

void CImageBuilder::LoadJPG(const CDataSource &source, bool bUseDithering)
{
  UNUSED_PARAMS_2(source, bUseDithering);
  VError(VERR_NotSupported, _NFL_, "JPEG loading support is disabled in configuration");
}

#endif


//
// TIFF Loading
//
#ifdef __USE_LIBTIFF__

tsize_t _tiffReadProc(thandle_t h, tdata_t b, tsize_t z)
{
  istream &is = reinterpret_cast<const CDataSource*>(h)->Stream();
  return(is.read(reinterpret_cast<char*>(b), z).gcount());
}

tsize_t _tiffWriteProc(thandle_t h, tdata_t b, tsize_t z)
{
  UNUSED_PARAMS_3(h, b, z);
  return(0);
}

toff_t _tiffSeekProc(thandle_t h, toff_t o, int n)
{
  const CDataSource       &ds = *reinterpret_cast<const CDataSource*>(h);
  const istream::off_type  sf = ds.GetStartOffsetOfData();
  const istream::off_type  ef = ds.GetStartOffsetOfData() + ds.Size();
        pistream          &is = ds.PStream();

  // Does the data has start offset ?
  if(sf) {
    switch(n) {
      case SEEK_SET: is.seekg(sf + o, std::ios_base::beg); break;
      case SEEK_CUR: is.seekg(     o, std::ios_base::cur); break;
      case SEEK_END: is.seekg(ef + o, std::ios_base::beg); break;
    }
    return(static_cast<toff_t>(is.tellg()) - sf);
  }
  // The data has no start offset
  else {
    switch(n) {
      case SEEK_SET: is.seekg(o, std::ios_base::beg); break;
      case SEEK_CUR: is.seekg(o, std::ios_base::cur); break;
      case SEEK_END: is.seekg(o, std::ios_base::end); break;
    }
    return(is.tellg());
  }
}

int _tiffCloseProc(thandle_t h)
{
  UNUSED_PARAM(h);
  return(0);
}

toff_t _tiffSizeProc(thandle_t h)
{
  return(reinterpret_cast<const CDataSource*>(h)->Size());
}

void CImageBuilder::LoadTIF(const CDataSource &source, bool bUseDithering)
{
  // Get the stream name
  const char *strmName = source.Name().c_str();

  // Open TIFF client connection
  TIFF *tif = TIFFClientOpen(strmName, "r", reinterpret_cast<thandle_t>(const_cast<CDataSource*>(&source)),
                             _tiffReadProc, _tiffWriteProc,
                             _tiffSeekProc, _tiffCloseProc,
                             _tiffSizeProc, 0, 0);
  if(!tif) VError(VERR_OpenFromStream, _NFL_, "Error opening TIFF client context from stream \"%s\"", strmName);

  // Get the image width, height and size
  uint32_t w, h;
  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);

  size_t npixels = w * h;

  // Allocate buffer
  uint32_t* raster = reinterpret_cast<uint32_t*>(_TIFFmalloc(npixels * sizeof(uint32_t)));
  if(!raster) {
    TIFFClose(tif);
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for loading TIFF stream \"%s\"", strmName);
  }

  // Load the image
  if(!TIFFReadRGBAImage(tif, w, h, reinterpret_cast<uint32*>(raster), 0)) {
    _TIFFfree(raster);
    TIFFClose(tif);
    VError(VERR_ReadStream, _NFL_, "Error when loading TIFF stream \"%s\"", strmName);
  }

  // Allocate memory for new image buffer
  TUN8RGB *tmpBuff = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * w * h));
  if(!tmpBuff) {
    _TIFFfree(raster);
    TIFFClose(tif);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");
  }

  // Convert the image
  u_char *src = reinterpret_cast<u_char*>(raster);

  for(uint32_t y = 0; y < h; y ++) {
    TUN8RGB *dst = &tmpBuff[(h-y-1) * w];

    for(uint32_t x = 0; x < w; x ++) {
      uint8_t r = *src++;
      uint8_t g = *src++;
      uint8_t b = *src++;
      float   a = float(*src++) / 255.0f;
      r = uint8_t(float(r) * a);
      g = uint8_t(float(g) * a);
      b = uint8_t(float(b) * a);

      dst->un8R = r;
      dst->un8G = g;
      dst->un8B = b;
      dst++;
    }
  }

  // Perform dithering (or clipping)
  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(tmpBuff, w, h);
  else
    _ClipRGB24ToVRGB_(tmpBuff, w, h);

  // Cleanup TIFF
  _TIFFfree(raster);
  TIFFClose(tif);

  // Destroy old image, if any
  FreeImage();

  // Set variables to the new image
  _un32Width  = w;
  _un32Height = h;
  _ptBuffer   = tmpBuff;
}

#else

void CImageBuilder::LoadTIF(const CDataSource &source, bool bUseDithering)
{
  UNUSED_PARAMS_2(source, bUseDithering);
  VError(VERR_NotSupported, _NFL_, "TIFF loading support is disabled in configuration");
}

#endif


//
// GIF Loading
//
#ifdef __USE_LIBUNGIF__

static int _gif_readCallback(GifFileType* h, GifByteType* b, int z)
{
  istream &is = reinterpret_cast<const CDataSource*>(h->UserData)->Stream();
  return(is.read(reinterpret_cast<char*>(b), z).gcount());
}

void CImageBuilder::LoadGIF(const CDataSource &source, bool bUseDithering)
{
  // Get the stream name
  const char *strmName = source.Name().c_str();

  // Preparation
  GifFileType *gif = DGifOpen(reinterpret_cast<void*>(const_cast<CDataSource*>(&source)), _gif_readCallback);
  if(!gif) VError(VERR_OpenFromStream, _NFL_, "Error opening GIF decompression context from stream \"%s\"", strmName);

  // Load the image
  if(DGifSlurp(gif) !=GIF_OK) {
    DGifCloseFile(gif);
    VError(VERR_ReadStream, _NFL_, "Error when loading GIF stream \"%s\"", strmName);
  }

  // Check if empty
  if(!gif->ImageCount) {
    DGifCloseFile(gif);
    VError(VERR_ReadStream, _NFL_, "GIF stream \"%s\" contains no image", strmName);
  }

  // Get the image data (just the 1st image if there are multiple images)
  SavedImage &sim = gif->SavedImages[0];
  uint32_t    w   = sim.ImageDesc.Width;
  uint32_t    h   = sim.ImageDesc.Height;
  u_char     *src = reinterpret_cast<u_char*>(sim.RasterBits);

  // Allocate memory for new image buffer
  TUN8RGB *tmpBuff = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * w * h));
  if(!tmpBuff) {
    DGifCloseFile(gif);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");
  }

  // Convert the image
  TUN8RGB *dst = tmpBuff;

  if(gif->SColorMap) {
    for(uint32_t y = 0; y < h; y ++) {
      for(uint32_t x = 0; x < w; x ++) {
        uint8_t c = *src++;
        uint8_t r = gif->SColorMap->Colors[c].Red;
        uint8_t g = gif->SColorMap->Colors[c].Green;
        uint8_t b = gif->SColorMap->Colors[c].Blue;

        dst->un8R = r;
        dst->un8G = g;
        dst->un8B = b;
        dst++;
      }
    }
  }
  else {
    for(uint32_t y = 0; y < h; y ++) {
      for(uint32_t x = 0; x < w; x ++) {
        uint8_t c = *src++;
        uint8_t r = c;
        uint8_t g = c;
        uint8_t b = c;

        dst->un8R = r;
        dst->un8G = g;
        dst->un8B = b;
        dst++;
      }
    }
  }

  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(tmpBuff, w, h);
  else
    _ClipRGB24ToVRGB_(tmpBuff, w, h);

  // Cleanup GIF
  DGifCloseFile(gif);

  // Destroy old image, if any
  FreeImage();

  // Set variables to the new image
  _un32Width  = w;
  _un32Height = h;
  _ptBuffer   = tmpBuff;
}

#else

void CImageBuilder::LoadGIF(const CDataSource &source, bool bUseDithering)
{
  UNUSED_PARAMS_2(source, bUseDithering);
  VError(VERR_NotSupported, _NFL_, "GIF loading support is disabled in configuration");
}

#endif


//
// BMP Loading
//
void CImageBuilder::LoadBMP(const CDataSource &source, bool bUseDithering)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Load the file header
  _TBITMAPFILEHEADER tFileHeader;
  is.read(reinterpret_cast<char*>(&tFileHeader), sizeof(tFileHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read BMP stream header data from file \"%s\"", strmName);  

  // Check the file header
  if((tFileHeader.aucBM[0] != 'B') || (tFileHeader.aucBM[1] != 'M'))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid BMP stream", strmName);

  // Load the information header data
  _TBITMAPINFOHEADER tInfoHeader;
  is.read(reinterpret_cast<char*>(&tInfoHeader), sizeof(tInfoHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read BMP information header data from stream \"%s\"", strmName);  

  // Check the information header
  if((tInfoHeader.un16BitCount != 1) && (tInfoHeader.un16BitCount != 4) &&
     (tInfoHeader.un16BitCount != 8) && (tInfoHeader.un16BitCount != 24))
    VError(VERR_NotSupported, _NFL_, "LoadBMP() can only load 1/4/8/24-bit bitmap\n\"%s\" contains a %d-bit bitmap\n", strmName, tInfoHeader.un16BitCount);
  if(tInfoHeader.un32Compression != __BI_RGB)
    VError(VERR_NotSupported, _NFL_, "LoadBMP() can only load uncompressed bitmaps");

  // Calculate some information
  int32_t n32ColorCount, n32BitCount, n32Width, n32Height, n32ImageSize;
  int32_t n32ScanWidth, n32ScanPadding, n32LoadSize;
  if(tInfoHeader.un32ClrUsed)
    n32ColorCount = tInfoHeader.un32ClrUsed;
  else
    n32ColorCount = 1 << tInfoHeader.un16BitCount;
  n32BitCount  = tInfoHeader.un16BitCount;
  n32Width     = tInfoHeader.n32Width;
  n32Height    = tInfoHeader.n32Height;
  n32ImageSize = n32Width * n32Height;
  if(n32BitCount == 1) {
    if(n32Width % 8) {
      n32ScanWidth   = ((n32Width / 8 + 1) + 3) & (~3);
      n32ScanPadding = n32ScanWidth - (n32Width / 8 + 1);
      n32LoadSize    = n32Width / 8 + 1;
    }
    else {
      n32ScanWidth   = ((n32Width / 8) + 3) & (~3);
      n32ScanPadding = n32ScanWidth - (n32Width / 8);
      n32LoadSize    = n32Width / 8;
    }
  }
  else if(n32BitCount == 4) {
    if(n32Width % 2) {
      n32ScanWidth   = ((n32Width / 2 + 1) + 3) & (~3);
      n32ScanPadding = n32ScanWidth - (n32Width / 2 + 1);
      n32LoadSize    = n32Width / 2 + 1;
    }
    else {
      n32ScanWidth   = ((n32Width / 2) + 3) & (~3);
      n32ScanPadding = n32ScanWidth - (n32Width / 2);
      n32LoadSize    = n32Width / 2;
    }
  }
  else {
    n32ScanWidth   = ((n32Width * (n32BitCount / 8)) + 3) & (~3);
    n32ScanPadding = n32ScanWidth - (n32Width * (n32BitCount / 8));
    n32LoadSize    = n32Width * (n32BitCount / 8);
  }

  // Skip some bytes if necessary
  uint32_t un32Skip = tInfoHeader.un32Size - sizeof(tInfoHeader);
  if(un32Skip > 0) {
    is.discard(un32Skip);
    if(is.fail())
      VError(VERR_ReadStream, _NFL_, "Couldn't skip bytes in BMP stream \"%s\"", strmName);  
  }

  // Load the pallete
  vector<_TRGBQUAD> rgbq;
  if(n32BitCount <= 8) {
    // Allocate memory
    rgbq.resize(n32ColorCount);

    // Load it
    is.read(reinterpret_cast<char*>(&rgbq[0]), sizeof(_TRGBQUAD) * n32ColorCount);
    if(is.fail())
      VError(VERR_ReadStream, _NFL_, "Couldn't read BMP information header data from stream \"%s\"", strmName);  
  }

  // Allocate memory for image
  u_char *pucData = NULL;
  if(n32BitCount == 1)
    pucData = reinterpret_cast<u_char*>(malloc(n32ImageSize));
  else if(n32BitCount == 4)
    pucData = reinterpret_cast<u_char*>(malloc(n32ImageSize / 2));
  else
    pucData = reinterpret_cast<u_char*>(malloc(n32ImageSize * n32BitCount / 8));
  if(!pucData)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for loading bitmap");

  // Load the bitmap
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    char discard;

    if(n32Height < 0) {
      for(int32_t n32Y = 0; n32Y < (-n32Height); n32Y++) {
        int32_t n32YPos = n32Y * n32LoadSize;
        is.read(reinterpret_cast<char*>(&pucData[n32YPos]), n32LoadSize);
        for(int32_t i = 0; i < n32ScanPadding; i++) is.get(discard);
      }
    }
    else {
      for(int32_t n32Y = 0; n32Y < n32Height; n32Y++) {
        int32_t n32YPos = ((n32Height - n32Y) * n32LoadSize) - n32LoadSize;
        is.read(reinterpret_cast<char*>(&pucData[n32YPos]), n32LoadSize);
        for(int32_t i = 0; i < n32ScanPadding; i++) is.get(discard);
      }
    }
  }
  catch(ios_base::failure&) {
    free(pucData);
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    free(pucData);
    throw;
  }

  // Destroy old image, if any
  FreeImage();

  // Allocate memory for new image buffer
  _un32Width  = uint32_t(n32Width);
  _un32Height = uint32_t((n32Height < 0) ? -n32Height : n32Height);
  _ptBuffer   = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * _un32Width * _un32Height));
  if(!_ptBuffer) {
    free(pucData);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image buffer");
  }

  // Convert image
  TUN8RGB *ptTmp  = _ptBuffer;
  u_char  *pucTmp = pucData;
  if(n32BitCount == 1) {
    uint32_t un32AGet = 0;
    uint32_t un32XGet = 0;
    for(uint32_t un32CntY = 0; un32CntY < (uint32_t(n32LoadSize) * _un32Height); un32CntY++) {
      uint8_t un8R = *pucTmp++;
      for(uint32_t un32CntX = 0; un32CntX < 8; un32CntX++) {
        if(un8R & (0x80 >> un32CntX)) {
          ptTmp->un8R = rgbq[1].un8R >> 3;
          ptTmp->un8G = rgbq[1].un8G >> 3;
          ptTmp->un8B = rgbq[1].un8B >> 3;
        }
        else {
          ptTmp->un8R = rgbq[0].un8R >> 3;
          ptTmp->un8G = rgbq[0].un8G >> 3;
          ptTmp->un8B = rgbq[0].un8B >> 3;
        }
        ptTmp++;
        un32AGet++;
        if(un32AGet >= (_un32Width * _un32Height)) break;
        un32XGet++;
        if(un32XGet >= _un32Width) {
          un32XGet = 0;
          break;
        }
      }
      if(un32AGet >= (_un32Width * _un32Height)) break;
    }
  }
  else if(n32BitCount == 4) {
    for(uint32_t un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
      for(uint32_t un32CntX = 0; un32CntX < (_un32Width / 2); un32CntX++) {
        uint32_t un32XGet = 0;
        while(un32XGet < 2) {
          // Grab the pixel
          uint8_t un8R, un8G, un8B;
          if(un32XGet == 0) {
            un8R = rgbq[(*pucTmp) & 0xF0].un8R;
            un8G = rgbq[(*pucTmp) & 0xF0].un8G;
            un8B = rgbq[(*pucTmp) & 0xF0].un8B;
            un32XGet = 1;
          }
          else {
            un8R = rgbq[(*pucTmp) & 0x0F].un8R;
            un8G = rgbq[(*pucTmp) & 0x0F].un8G;
            un8B = rgbq[(*pucTmp) & 0x0F].un8B;
            un32XGet = 2;
          }
          // Put the destination pixel
          ptTmp->un8R = un8R;
          ptTmp->un8G = un8G;
          ptTmp->un8B = un8B;
          ptTmp++;
        }
        // Increment the source pointer
        pucTmp++;
      }
    }
    if(bUseDithering)
      _OrderedDitherRGB24ToVRGB_(_ptBuffer, _un32Width, _un32Height);
    else
      _ClipRGB24ToVRGB_(_ptBuffer, _un32Width, _un32Height);
  }
  else {
    for(uint32_t un32CntY = 0; un32CntY < _un32Height; un32CntY++) {
      for(uint32_t un32CntX = 0; un32CntX < _un32Width; un32CntX++) {
        // Grab the pixel
        uint8_t un8R, un8G, un8B;
        if(n32BitCount == 8) {
          un8R = rgbq[*pucTmp].un8R;
          un8G = rgbq[*pucTmp].un8G;
          un8B = rgbq[*pucTmp].un8B;
          pucTmp++;
        }
        else {
          un8B = *pucTmp++;
          un8G = *pucTmp++;
          un8R = *pucTmp++;
        }
        // Put the destination pixel
        ptTmp->un8R = un8R;
        ptTmp->un8G = un8G;
        ptTmp->un8B = un8B;
        ptTmp++;
      }
    }
    if(bUseDithering)
      _OrderedDitherRGB24ToVRGB_(_ptBuffer, _un32Width, _un32Height);
    else
      _ClipRGB24ToVRGB_(_ptBuffer, _un32Width, _un32Height);
  }

  // Free memory
  free(pucData);
}

} // namespace vge
