/** \file ImageClasses.h */
// ############################################################################################
//  ImageClasses.h
//  Definition & implementation of CTexture, CAnimatedTexture,
//  CImage, and CAnimatedImage classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___IMAGE_CLASSES_H_
#define _2D_MANIPULATION___IMAGE_CLASSES_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"


namespace vge {

  //!
  //! \brief A class used for rendering using texture
  //!
  //! \api pretty stable
  class VGE_EXPORT CTexture : VCLASS_UNCOPYABLE, VCLASS_CAN_ABORT_ON_DESTRUCTION {
    public:
      CTexture();
      ~CTexture() throw();

      // Texture functions
      //!
      //! Returns the width of the texture
      inline uint32_t GetWidth() const
      { return(_un32Width); }
      //!
      //! Returns the height of the texture.
      inline uint32_t GetHeight() const
      { return(_un32Height); }
      //!
      //! Returns the raw buffer of the texture.
      inline const uint8_t *GetBuffer() const
      { return(_pun8Buffer); }
      //!
      //! Returns the maximum X coordinate of the texture.
      inline uint32_t XMax() const
      { return(_un32MaxX); }
      //!
      //! Returns the maximum Y coordinate of the texture.
      inline uint32_t YMax() const
      { return(_un32MaxY); }
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeTexture();

      // Friend classes
      friend class CImageBuilder;
      friend class CPanel3DBuffer;
      friend class CPanel3D;

    protected:
      uint8_t *_pun8Buffer;    // The texture buffer (in VRGB format)
      uint32_t _un32Width;     // The texture width
      uint32_t _un32Height;    // The texture height
      uint32_t _un32MaxX;;     // The texture maximum X coordinate
      uint32_t _un32MaxY;      // The texture maximum Y coordinate
      uint32_t _un32Mask;      // The address mask
      uint32_t _un32MultShift; // The pre-multiply shift
      bool     _bAligned;      // Indicates that the buffer address is alignend
  };


  //!
  //! \brief A class used for rendering using animated texture
  //!
  //! \api pretty stable
  class VGE_EXPORT CAnimatedTexture : public CTexture {
    public:
      CAnimatedTexture();
      ~CAnimatedTexture() throw();

      // Texture functions
      //!
      //! Get the total number of animation frames.
      inline uint32_t GetTotalFrame() const
      { return(_un32TotFrame); }
      //!
      //! Get the frame size in bytes.
      inline uint32_t GetFrameSize() const
      { return(_un32FrameSize); }
      //!
      //! Returns the raw buffer array of the animated texture.
      inline const uint8_t *GetBufferArray() const
      { return(_pun8BufferArray); }
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeTexture();

      // Frame functions
      //!
      //! Step to the next frame.
      //! \param[in] bWarp If "true", the frame will be warped back to the 1st frame if past the last frame.
      inline void NextFrame(bool bWarp = false);
      //!
      //! Step to the next frame.
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping to the next frame.
      //! \param[in] bWarp     If "true", the frame will be warped back to the 1st frame after reaching the last frame.
      inline void NextFrame(uint32_t un32Delay, bool bWarp = false);
      //!
      //! Step to the previous frame.
      //! \param[in] bWarp If "true", the frame will be warped back to the last frame if past the 1st frame.
      inline void PrevFrame(bool bWarp = false);
      //!
      //! Step to the previous frame.
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping to the previous frame.
      //! \param[in] bWarp     If "true", the frame will be warped back to the last frame if past the 1st frame.
      inline void PrevFrame(uint32_t un32Delay, bool bWarp = false);
      //!
      //! Step to the next/previous frame continously (warped).
      inline void StepFrame();
      //!
      //! Step to the next/previous frame continously (warped).
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping.
      inline void StepFrame(uint32_t un32Delay);
      //!
      //! Step directly do the given frame number.
      //! \param[in] un32Frame The frame number to be stepped to.
      inline void SetFrame(uint32_t un32Frame);

      // Friend classes
      friend class CImageBuilder;
      friend class CPanel3DBuffer;
      friend class CPanel3D;

    protected:
      uint8_t *_pun8BufferArray; // The image buffer array (in VRGB format)
      uint8_t *_pun8BufferMax;   // The maximum buffer address
      uint32_t _un32FrameSize;   // Size of each frame in bytes
      uint32_t _un32TotFrame;    // Total frames
      uint32_t _un32CurFrame;    // Current frame
      uint32_t _un32StepDelay;   // Stepping delay
      int8_t   _n8Step;          // The step factor
      bool     _bAligned;        // Indicates that the buffer address is alignend
  };

  void CAnimatedTexture::NextFrame(bool bWarp)
  {
    if(_un32CurFrame >= (_un32TotFrame - 1)) {
      if(bWarp) {
        _un32CurFrame = 0;
        _pun8Buffer   = _pun8BufferArray;
      }
      return;
    }
    _un32CurFrame++;
    _pun8Buffer += _un32FrameSize;
  }

  void CAnimatedTexture::NextFrame(uint32_t un32Delay, bool bWarp)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    NextFrame(bWarp);
  }

  void CAnimatedTexture::PrevFrame(bool bWarp)
  {
    if(!_un32CurFrame) {
      if(bWarp) {
        _un32CurFrame = _un32TotFrame - 1;
        _pun8Buffer   = _pun8BufferMax;
      }
      return;
    }
    _un32CurFrame--;
    _pun8Buffer -= _un32FrameSize;
  }

  void CAnimatedTexture::PrevFrame(uint32_t un32Delay, bool bWarp)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    PrevFrame(bWarp);
  }

  void CAnimatedTexture::StepFrame()
  {
    // Take the advantage that uint32_t(-1) is a very big uint32_t
    if(uint32_t(_un32CurFrame + _n8Step) >= _un32TotFrame) _n8Step = -_n8Step;
    _un32CurFrame += _n8Step;
    _pun8Buffer   += int32_t(_un32FrameSize * _n8Step);
  }

  void CAnimatedTexture::StepFrame(uint32_t un32Delay)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    StepFrame();
  }

  void CAnimatedTexture::SetFrame(uint32_t un32Frame)
  {
    _un32CurFrame = min(un32Frame, _un32TotFrame - 1);
    _pun8Buffer   = _pun8BufferArray + (_un32FrameSize * _un32CurFrame);
  }


  //!
  //! \brief A class used for storing image compatible with the render buffer
  //!
  //! \api pretty stable
  class VGE_EXPORT CImage : VCLASS_UNCOPYABLE, VCLASS_CAN_ABORT_ON_DESTRUCTION {
    public:
      CImage();
      ~CImage() throw();

      // Image functions
      //!
      //! Returns the width of the image.
      inline uint32_t GetWidth() const
      { return(_un32Width); }
      //!
      //! Returns the height of the image.
      inline uint32_t GetHeight() const
      { return(_un32Height); }
      //!
      //! Returns the raw buffer of the image.
      inline const uint8_t *GetBuffer() const
      { return(_pun8Buffer); }
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeImage();

      // Friend classes
      friend class CImageBuilder;
      friend class C3DEnv;
      friend class CSprite2D;

    protected:
      uint8_t *_pun8Buffer; // The image buffer (in VRGB format)
      uint32_t _un32Width;  // The image width
      uint32_t _un32Height; // The image height
      uint32_t _un32SizeSL; // The scanline size in bytes
  };


  //!
  //! \brief A class used for storing animated image compatible with the render buffer
  //!
  //! \api pretty stable
  class VGE_EXPORT CAnimatedImage : public CImage {
    public:
      CAnimatedImage();
      ~CAnimatedImage() throw();

      // Image functions
      //!
      //! Get the total number of animation frames.
      inline uint32_t GetTotalFrame() const
      { return(_un32TotFrame); }
      //!
      //! Get the frame size in bytes.
      inline uint32_t GetFrameSize() const
      { return(_un32FrameSize); }
      //!
      //! Returns the raw buffer array of the animated image.
      inline const uint8_t *GetBufferArray() const
      { return(_pun8BufferArray); }
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeImage();

      // Frame functions
      //!
      //! Step to the next frame.
      //! \param[in] bWarp If "true", the frame will be warped back to the 1st frame if past the last frame.
      inline void NextFrame(bool bWarp = false);
      //!
      //! Step to the next frame.
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping to the next frame.
      //! \param[in] bWarp     If "true", the frame will be warped back to the 1st frame after reaching the last frame.
      inline void NextFrame(uint32_t un32Delay, bool bWarp = false);
      //!
      //! Step to the previous frame.
      //! \param[in] bWarp If "true", the frame will be warped back to the last frame if past the 1st frame.
      inline void PrevFrame(bool bWarp = false);
      //!
      //! Step to the previous frame.
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping to the previous frame.
      //! \param[in] bWarp     If "true", the frame will be warped back to the last frame if past the 1st frame.
      inline void PrevFrame(uint32_t un32Delay, bool bWarp = false);
      //!
      //! Step to the next/previous frame continously (warped).
      inline void StepFrame();
      //!
      //! Step to the next/previous frame continously (warped).
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping.
      inline void StepFrame(uint32_t un32Delay);
      //!
      //! Step directly do the given frame number.
      //! \param[in] un32Frame The frame number to be stepped to.
      inline void SetFrame(uint32_t un32Frame);

      // Friend classes
      friend class CImageBuilder;
      friend class C3DEnv;
      friend class CSprite2D;
      friend class CAImgControl;
      friend class CBitmapFont;

    protected:
      uint8_t *_pun8BufferArray; // The image buffer array (in VRGB format)
      uint8_t *_pun8BufferMax;   // The maximum buffer address
      uint32_t _un32FrameSize;   // Size of each frame in bytes
      uint32_t _un32TotFrame;    // Total frames
      uint32_t _un32CurFrame;    // Current frame
      uint32_t _un32StepDelay;   // Stepping delay
      int8_t   _n8Step;          // The step factor
      bool     _bAligned;        // Indicates that the buffer address is alignend
  };

  void CAnimatedImage::NextFrame(bool bWarp)
  {
    if(_un32CurFrame >= (_un32TotFrame - 1)) {
      if(bWarp) {
        _un32CurFrame = 0;
        _pun8Buffer   = _pun8BufferArray;
      }
      return;
    }
    _un32CurFrame++;
    _pun8Buffer += _un32FrameSize;
  }

  void CAnimatedImage::NextFrame(uint32_t un32Delay, bool bWarp)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    NextFrame(bWarp);
  }

  void CAnimatedImage::PrevFrame(bool bWarp)
  {
    if(!_un32CurFrame) {
      if(bWarp) {
        _un32CurFrame = _un32TotFrame - 1;
        _pun8Buffer   = _pun8BufferMax;
      }
      return;
    }
    _un32CurFrame--;
    _pun8Buffer -= _un32FrameSize;
  }

  void CAnimatedImage::PrevFrame(uint32_t un32Delay, bool bWarp)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    PrevFrame(bWarp);
  }

  void CAnimatedImage::StepFrame()
  {
    // Take the advantage that uint32_t(-1) is a very big uint32_t
    if(uint32_t(_un32CurFrame + _n8Step) >= _un32TotFrame) _n8Step = -_n8Step;
    _un32CurFrame += _n8Step;
    _pun8Buffer   += int32_t(_un32FrameSize * _n8Step);
  }

  void CAnimatedImage::StepFrame(uint32_t un32Delay)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    StepFrame();
  }

  void CAnimatedImage::SetFrame(uint32_t un32Frame)
  {
    _un32CurFrame = min(un32Frame, _un32TotFrame - 1);
    _pun8Buffer   = _pun8BufferArray + (_un32FrameSize * _un32CurFrame);
  }

} // namespace vge
#endif
