/** \file MiscellaneousUtilities.h */
// ############################################################################################
//  MiscellaneousUtilities.h
//  Definition of various utilities related to image processing
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___MISCELLANEOUS_UTILITIES_H_
#define _2D_MANIPULATION___MISCELLANEOUS_UTILITIES_H_

// VortexGE headers
#include "../Base/VBasics.h"


namespace vge {
  //!
  //! \addtogroup DCTRelatedFunctions
  //! @{

  // General functions
  //!
  //! Dump the contents of the given matrix.
  //! \param[in] pcInfoText Additional text to be printed when dumping the matrix.
  //! \param[in] pfMatrix   Matrix to be dumped.
  //! \param[in] un8Dim     Dimension of the matrix.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTDumpMatrix(const char *pcInfoText, const float *pfMatrix, uint8_t un8Dim);
  //!
  //! Transpose the given DMT matrix.
  //! \param[in,out] pfMatrix Matrix to be transposed.
  //! \param[in]     un8Dim   Dimension of the matrix.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTTransposeMatrix(float *pfMatrix, uint8_t un8Dim);
  //!
  //! Perform DMT matrix multiplication according to the equation:\n
  //! <span style="margin-left:2em;"><b>|I| = |K| x |I| x |K|</b></span>
  //! \param[in,out] pfIMatrix Image matrix to be multiplied.
  //! \param[in]     pfKMatrix Kernel matrix to be used.
  //! \param[in]     un8Dim    Dimension of the matrix.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTMultMatricesKIK(float *pfIMatrix, const float *pfKMatrix, uint8_t un8Dim);

  // Quantitization functions
  //!
  //! Remove the coefficients of which rows and columns satisfy the equation:\n
  //! <span style="margin-left:2em;"><b>Row x Col >= CRLimit</b></span>\n
  //! and quantitize the remaining coefficients.
  //! \param[in,out] pfMatrix        Matrix to be reduced & quantitized.
  //! \param[in]     un8Dim          Dimension of the matrix.
  //! \param[in]     un8CRStrength   Reduction strength.
  //! \param[out]    pun32CRLimit    Calculated reduction limit based on the equation:\n
  //!                                <span style="margin-left:2em;"><b>
  //!                                CRLimit&nbsp;=&nbsp;Dim<sup>2</sup>&nbsp;/&nbsp;2<sup>CRStrength</sup>
  //!                                </b></span>
  //! \param[out]    pun32CRRetained Total coefficients retained.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTQuantitizeIMatrix(float *pfIMatrix, uint8_t un8Dim, uint8_t un8CRStrength, uint32_t *pun32CRLimit = NULL, uint32_t *pun32CRRetained = NULL);
  //!
  //! Reverse the operations that have been done by the VDMTQuantitizeIMatrix() function.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTDequantitizeIMatrix(float *pfIMatrix, uint8_t un8Dim);

  // Discrete Cosine Transform (DCT) related functions
  //!
  //! Create a DCT kernel matrix.
  //! \param[out] pfKMatrix Matrix produced.
  //! \param[in]  un8Dim    Dimension of the matrix.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTCreateDCTKernelMatrix(float *pfKMatrix, uint8_t un8Dim);
  //!
  //! Create an inverse DCT kernel matrix.
  //! \param[out] pfKMatrix Matrix produced.
  //! \param[in]  un8Dim    Dimension of the matrix.
  //!
  //! \api beta
  extern void VGE_EXPORT VDMTCreateInvDCTKernelMatrix(float *pfKMatrix, uint8_t un8Dim);

  //! @}
  //!
} // namespace vge
#endif
