/** \file 3DEnvironment.h */
// ############################################################################################
//  3DEnvironment.h
//  Definition & implementation of C3DEnv class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___3D_ENVIRONMENT_H_
#define _3D_ENVIRONMENT___3D_ENVIRONMENT_H_

// VortexGE headers
#include "../2DManipulation/ImageClasses.h"
#include "../Base/LookUpTables.h"
#include "../HighLevelClass/FrameRateControl.h"
#include "../VideoEnvironment/VideoBase.h"
#include "CameraControl.h"


namespace vge {

  //!
  //! \brief Omni light structure
  //!
  //! \api pretty stable
  struct TOmniLight {
    float fX, fY, fZ;         // The omni light position
#ifdef _ALIGN_128_BITS_
    float _fdummy_to_make_the_XYZ_position_128_bits_long_;
#endif
    uint8_t un8R, un8G, un8B; // The omni light color
  };

  //!
  //! \brief Ambient light structure
  //!
  //! \api pretty stable
  struct TAmbientLight {
    uint8_t un8R, un8G, un8B; // The ambient light color
  };


  //!
  //! Used for calculating light intensity
  //!
  //! \api pretty stable
  extern float VGE_EXPORT VCalcLightCosValue(const float *pfPanelNormalXYZ4, const float *pfPanelPointXYZ4, const float *pfLightPosXYZ4);


  //!
  //! \brief The 3D environment class
  //!
  //! \api pretty stable
  class VGE_EXPORT C3DEnv : VCLASS_UNCOPYABLE {
    public:
      C3DEnv();
      ~C3DEnv() throw();

      // Initialization functions
      //!
      //! Initialize the 3D environment and attach it to the given
      //! video environment instance.
      void Initialize(CVideoEnv *pcsVVideoEnv);

      // Profiling functions (will be removed soon)
      //!
      //! \deprecated Will be removed soon; use the same function from CFPSControl class instead.
      inline static uint64_t RDTSC()
      { return CFPSControl::RDTSC(); }
      //!
      //! \deprecated Will be removed soon; use the same function from CFPSControl class instead.
      inline static uint64_t GetTicks()
      { return CFPSControl::GetTicks(); }
      //!
      //! \deprecated Will be removed soon; use the same function from CFPSControl class instead.
      inline static float Ticks2mS(uint64_t un64Ticks) 
      { return CFPSControl::Ticks2mS(un64Ticks); }
      //!
      //! \deprecated Will be removed soon; use the same function from CFPSControl class instead.
      inline static float CalcIntervalmS(uint64_t un64StartTicks, uint64_t un64EndTicks)
      { return CFPSControl::CalcIntervalmS(un64StartTicks, un64EndTicks); }
      //!
      //! \deprecated Will be removed soon; use the same function from CFPSControl class instead.
      inline static float CalcAvgFPS(uint64_t un64TotFrames, uint64_t un64StartTicks, uint64_t un64EndTicks)
      { return CFPSControl::CalcAvgFPS(un64TotFrames, un64StartTicks, un64EndTicks); }
      //!
      //! \deprecated Will be removed soon; use the CFPSControl class instead.
      inline float CalcFPS()
      { return(_csFPSCtrl.CalculateFPS()); }

      // Information functions
      //!
      //! Returns the width of the render buffer.
      inline uint32_t RenderBufferWidth() const { return(_tRB->un32Width); }
      //!
      //! Returns the height of the render buffer.
      inline uint32_t RenderBufferHeight() const { return(_tRB->un32Height); }
      //!
      //! Returns the minimum X coordinate of the render buffer.
      inline int32_t XMin() const { return(_n32MinX); }
      //!
      //! Returns the maximum X coordinate of the render buffer.
      inline int32_t XMax() const { return(_n32MaxX); }
      //!
      //! Returns the minimum Z coordinate of the render buffer.
      inline int32_t YMin() const { return(_n32MinY); }
      //!
      //! Returns the maximum Y coordinate of the render buffer.
      inline int32_t YMax() const { return(_n32MaxY); }
      //!
      //! Returns the minimum Z coordinate of the render buffer.
      inline uint32_t ZMin() const { return(_un32MinZ); }
      //!
      //! Returns the maximum Z coordinate of the render buffer.
      inline uint32_t ZMax() const { return(_un32MaxZ); }
      //!
      //! Returns the current Z-buffer translation factor.
      //! \n\n
      //! \sntc
      inline int32_t ZTrans()  const { return(_tRB->n32ZTrans); }
      //!
      //! Returns the X scalling factor.
      //! \n\n
      //! \sntc
      inline float XScale()  const { return(_fXScale); }
      //!
      //! Returns the Y scalling factor.
      //! \n\n
      //! \sntc
      inline float YScale() const { return(_fYScale); }
      //!
      //! Returns the X scalling factor divided by 4.
      //! \n\n
      //! \sntc
      inline float XScale4() const { return(_fXScale4); }
      //!
      //! Returns the Y scalling factor divided by 4.
      //! \n\n
      //! \sntc
      inline float YScale4() const { return(_fYScale4); }
      //!
      //! Returns the X shifting factor.
      //! \n\n
      //! \sntc
      inline float XShift() const { return(_fXShift); }
      //!
      //! Returns the Y shifting factor.
      //! \n\n
      //! \sntc
      inline float YShift() const { return(_fYShift); }
      //!
      //! Returns the fog shifting factor.
      //! \n\n
      //! \sntc
      inline uint32_t FogShift() const { return(_un32FogShift); }
      //!
      //! Returns the fog depth.
      inline uint8_t FogDepth() const { return(16 - _un32FogShift); }
      //!
      //! Returns the color (R component).
      inline uint8_t FogColorR() const { return(_un8FogR); }
      //!
      //! Returns the color (G component).
      inline uint8_t FogColorG() const { return(_un8FogG); }
      //!
      //! Returns the color (B component).
      inline uint8_t FogColorB() const { return(_un8FogB); }

      // Environment functions
      //!
      //! Access to the active 3D environment without needing an 
      //! instance of 3D environment.
      inline static C3DEnv *Active3DEnv()
      { return(C3DEnv::_active3DEnv); }
      //!
      //! Make this 3D environment instance as the active rendering context.
      void SetAsTheActiveEnv();
      //!
      //! Get the currently active rendering context (3D environment instance).
      inline C3DEnv *GetTheActiveEnv()
      { return(C3DEnv::Active3DEnv()); }
      //!
      //! Get the currently active rendering context (3D environment instance).
      inline const C3DEnv *GetTheActiveEnv() const
      { return(C3DEnv::Active3DEnv()); }
      //!
      //! Check if this 3D environment instance is the active rendering context.
      inline bool IsTheActiveEnv() const
      { return(C3DEnv::Active3DEnv() == this); }
      //!
      //! Set the default ambient light color.
      inline void SetDefaultAmbientLightColor(uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Get the default ambient light.
      inline TAmbientLight *GetDefaultAmbientLight()
      { return(&_tDefALight); }
      //!
      //! Get the default ambient light.
      inline const TAmbientLight *GetDefaultAmbientLight() const
      { return(&_tDefALight); }
      //!
      //! Set the default omni light color.
      inline void SetDefaultOmniLightColor(uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Set the default omni light position.
      inline void SetDefaultOmniLightPosition(float fX, float fY, float fZ);
      //!
      //! Get the default omni light.
      inline TOmniLight *GetDefaultOmniLight()
      { return(&_tDefOLight); }
      //!
      //! Get the default omni light.
      inline const TOmniLight *GetDefaultOmniLight() const
      { return(&_tDefOLight); }
      //!
      //! Set the fog state (color and depth).
      void SetFogState(uint8_t un8FogColorR, uint8_t un8FogColorG, uint8_t un8FogColorB, uint8_t un8FogDepth);
      //!
      //! Set the Z clipping value (the minimum Z and maximum Z coordinates
      //! of the render buffer).
      void SetZClipping(uint32_t un32MinZ = 50, uint32_t un32MaxZ = 5000);
      //!
      //! Set the viewport geometry (supplying -1 in the parameters means using
      //! the default values).
      void SetViewport(int32_t n32X1 = -1, int32_t n32Y1 = -1, int32_t n32X2 = -1, int32_t n32Y2 = -1);

      // Rendering functions
      //!
      //! Clear the render buffer with black fill.
      inline void ClearRenderBuffer()
      { ZeroMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), _tRB->un32Size48); }
      //!
      //! Clear the render buffer with the given color.
      inline void ClearRenderBuffer(XColor15 xColorVRGB)
      { FillMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), xColorVRGB, _tRB->un32Size48); }
      //!
      //! Clear the render buffer with the given image.
      inline void ClearRenderBuffer(const CImage &csImageBackground)
      { CopyMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), reinterpret_cast<const uint32_t*>(csImageBackground._pun8Buffer), _tRB->un32Size48); }
      //!
      //! Clear the render buffer with the given animated image.
      inline void ClearRenderBuffer(const CAnimatedImage &csImageBackground) 
      { ClearRenderBuffer(reinterpret_cast<const CImage&>(csImageBackground)); }
      //!
      //! Clear 75% of the render buffer (from top) with black fill.
      inline void ClearRenderBuffer75()
      { ZeroMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), _tRB->un32Size4875); }
      //!
      //! Clear 75% of the render buffer (from top) with the given color.
      inline void ClearRenderBuffer75(XColor15 xColorVRGB)
      { FillMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), xColorVRGB, _tRB->un32Size4875); }
      //!
      //! Clear 75% of the render buffer (from top) with the given image.
      inline void ClearRenderBuffer75(CImage &csImageBackground)
      { CopyMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), reinterpret_cast<const uint32_t*>(csImageBackground._pun8Buffer), _tRB->un32Size4875); }
      //!
      //! Clear 75% of the render buffer (from top) the given animated image.
      inline void ClearRenderBuffer75(CAnimatedImage &csImageBackground)
      { ClearRenderBuffer75(reinterpret_cast<CImage&>(csImageBackground)); }
      //!
      //! Clear 50% of the render buffer (from top) with black fill.
      inline void ClearRenderBuffer50()
      { ZeroMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), _tRB->un32Size4850); }
      //!
      //! Clear 50% of the render buffer (from top) with the given color.
      inline void ClearRenderBuffer50(XColor15 xColorVRGB)
      { FillMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), xColorVRGB, _tRB->un32Size4850); }
      //!
      //! Clear 50% of the render buffer (from top) with the given image.
      inline void ClearRenderBuffer50(CImage &csImageBackground)
      { CopyMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), reinterpret_cast<const uint32_t*>(csImageBackground._pun8Buffer), _tRB->un32Size4850); }
      //!
      //! Clear 50% of the render buffer (from top) the given animated image.
      inline void ClearRenderBuffer50(CAnimatedImage &csImageBackground)
      { ClearRenderBuffer50(reinterpret_cast<CImage&>(csImageBackground)); }
      //!
      //! Clear 25% of the render buffer (from top) with black fill.
      inline void ClearRenderBuffer25()
      { ZeroMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), _tRB->un32Size4825); }
      //!
      //! Clear 25% of the render buffer (from top) with the given color.
      inline void ClearRenderBuffer25(XColor15 xColorVRGB) 
      { FillMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), xColorVRGB, _tRB->un32Size4825); }
      //!
      //! Clear 25% of the render buffer (from top) with the given image.
      inline void ClearRenderBuffer25(CImage &csImageBackground)
      { CopyMem4X8(reinterpret_cast<uint32_t*>(_tRB->pun8Buffer), reinterpret_cast<uint32_t*>(csImageBackground._pun8Buffer), _tRB->un32Size4825); }
      //!
      //! Clear 25% of the render buffer (from top) the given animated image.
      inline void ClearRenderBuffer25(CAnimatedImage &csImageBackground)
      { ClearRenderBuffer25(reinterpret_cast<CImage&>(csImageBackground)); }
      //!
      //! Clear the Z-buffer if needed.
      inline void ClearZBuffer();
      //!
      //! Fore clear the Z-buffer even it may still not needed.
      inline void ForceClearZBuffer();

      // Drawing functions (with static image)
      //!
      //! Put the given image to the given coordinate.
      void PutImage(const CImage &csImagePut, uint32_t un32X, int32_t un32Y);
      //!
      //! Put the given image to the given coordinate (make the pixels
      //! with the given color as transparent pixel).
      void PutImage(const CImage &csImagePut, uint32_t un32X, int32_t un32Y, XColor15 xExcludedColorVRGB);
      //!
      //! Clip and put the given image to the given coordinate.
      void ClipPutImage(const CImage &csImagePut, int32_t n32X, int32_t n32Y);
      //!
      //! Clip and put the given image to the given coordinate (make the pixels
      //! with the given color as transparent pixel).
      void ClipPutImage(const CImage &csImagePut, int32_t n32X, int32_t n32Y, XColor15 xExcludedColorVRGB);

      // Drawing functions (with animated image)
      //!
      //! Put the given animated image to the given coordinate.
      inline void PutImage(const CAnimatedImage &csImagePut, uint32_t un32X, int32_t un32Y)
      { PutImage(reinterpret_cast<const CImage&>(csImagePut), un32X, un32Y); }
      //!
      //! Put the given animated image to the given coordinate (make
      //! the pixels with the given color as transparent pixel).
      inline void PutImage(const CAnimatedImage &csImagePut, uint32_t un32X, int32_t un32Y, XColor15 xExcludedColorVRGB)
      { PutImage(reinterpret_cast<const CImage&>(csImagePut), un32X, un32Y, xExcludedColorVRGB); }
      //!
      //! Clip and put the given animated image to the given coordinate.
      inline void ClipPutImage(const CAnimatedImage &csImagePut, int32_t n32X, int32_t n32Y)
      { ClipPutImage(reinterpret_cast<const CImage&>(csImagePut), n32X, n32Y); }
      //!
      //! Clip and put the given animated image to the given coordinate
      //! (make the pixels with the given color as transparent pixel).
      inline void ClipPutImage(const CAnimatedImage &csImagePut, int32_t n32X, int32_t n32Y, XColor15 xExcludedColorVRGB)
      { ClipPutImage(reinterpret_cast<const CImage&>(csImagePut), n32X, n32Y, xExcludedColorVRGB); }

      // Support for SSE
      DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()

      // Friend classes
      friend class CPanel3D;
      friend class CPanel3DBuffer;
      friend class _CUFCLineSeg;
      friend class _CUGCLineSeg;
      friend class _CTFCLineSeg;
      friend class _CTGCLineSeg;
      friend class _CFogUFCLineSeg;
      friend class _CFogUGCLineSeg;
      friend class _CFogTFCLineSeg;
      friend class _CFogTGCLineSeg;
      friend class CImageBuilder;
      friend class CSprite2D;
      friend class CSnow2D;
      friend class CBitmapFont;
      friend class CWalkingCameraControl::CSkybox;

      // Friend functions
      friend void _UFClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep);
      friend void _UGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                                int32_t &n32B, int32_t &n32BStep);
      friend void _TFClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep);
      friend void _TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                                int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                                int32_t &n32B, int32_t &n32BStep);
      friend void _FogUF_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                      int32_t &n32F, int32_t &n32FStep);
      friend void _FogUG_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                      int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                                      int32_t &n32B, int32_t &n32BStep, int32_t &n32F, int32_t &n32FStep);
      friend void _FogTF_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                      int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                                      int32_t &n32F, int32_t &n32FStep);
      friend void _FogTG_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                      int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                                      int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                                      int32_t &n32B, int32_t &n32BStep, int32_t &n32F, int32_t &n32FStep);

    protected:
      CFPSControl         _csFPSCtrl;    // Instance to CFPSControl class
      TAmbientLight       _tDefALight;   // The default ambient light
      TOmniLight     _sse _tDefOLight;   // The default omni light
      _TRenderBuffer     *_tRB;          // Pointer to the render buffer
      int32_t             _n32FCount;    // Cointer for clearing Z-buffer
      float               _fXScale;      // The X scale factor
      float               _fYScale;      // The Y scale factor
      float               _fXScale4;     // The X scale factor / 4
      float               _fYScale4;     // The Y scale factor / 4
      float               _fXShift;      // XMin + (XMax - XMin) / 2
      float               _fYShift;      // YMin + (YMax - YMin) / 2
      int32_t             _n32MinX;      // The minimum X clipping
      int32_t             _n32MaxX;      // The maximum X clipping
      int32_t             _n32MinY;      // The minimum Y clipping
      int32_t             _n32MaxY;      // The maximum Y clipping
      uint32_t            _un32MinZ;     // The minimum Z clipping
      uint32_t            _un32MaxZ;     // The maximum Z clipping
      uint32_t            _un32FogShift; // The fog shifting factor
      uint8_t             _un8FogR;      // The fog color (red component)
      uint8_t             _un8FogG;      // The fog color (green component)
      uint8_t             _un8FogB;      // The fog color (blue component)
      uint32_t           *_pun32RFLT;    // The R component of the fogged-lighting LUT
      uint32_t           *_pun32GFLT;    // The G component of the fogged-lighting LUT
      uint32_t           *_pun32BFLT;    // The B component of the fogged-lighting LUT
      uint32_t           *_pun32RFCT;    // The R component of the fogged-color LUT
      uint32_t           *_pun32GFCT;    // The G component of the fogged-color LUT
      uint32_t           *_pun32BFCT;    // The B component of the fogged-color LUT
      uint8_t            *_pun8TRFCT;    // The R component of the fogged-color LUT with transparency
      uint8_t            *_pun8TGFCT;    // The G component of the fogged-color LUT with transparency
      uint8_t            *_pun8TBFCT;    // The B component of the fogged-color LUT with transparency
      uint8_t            *_pun8TRFLT;    // The R component of the fogged-lighting LUT with transparency
      uint8_t            *_pun8TGFLT;    // The G component of the fogged-lighting LUT with transparency
      uint8_t            *_pun8TBFLT;    // The B component of the fogged-lighting LUT with transparency

      // The active 3D environment
      static C3DEnv *_active3DEnv;
  };

  // Environment functions
  void C3DEnv::SetDefaultAmbientLightColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
  {
    _tDefALight.un8R = un8R;
    _tDefALight.un8G = un8G;
    _tDefALight.un8B = un8B;
  }

  void C3DEnv::SetDefaultOmniLightColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
  {
    _tDefOLight.un8R = un8R;
    _tDefOLight.un8G = un8G;
    _tDefOLight.un8B = un8B;
  }

  void C3DEnv::SetDefaultOmniLightPosition(float fX, float fY, float fZ)
  {
    _tDefOLight.fX = fX;
    _tDefOLight.fY = fY;
    _tDefOLight.fZ = fZ;
  }

  void C3DEnv::ClearZBuffer()
  {
    // Clear Z buffer if has reached the translation limit
    if(_n32FCount >= _tRB->n32ZWait) {
      _n32FCount      = 0;
      _tRB->n32ZTrans = 0;
      ZeroMem4X8(reinterpret_cast<uint32_t*>(_tRB->pn32ZBuffer), _tRB->un32Size48);
    }
    // Otherwise, just translate it
    else {
      _n32FCount++;
      _tRB->n32ZTrans += _tRB->n32ZTransInc;
    }
  }

  void C3DEnv::ForceClearZBuffer()
  {
    _tRB->n32ZTrans  = 0;
    ZeroMem4X8(reinterpret_cast<uint32_t*>(_tRB->pn32ZBuffer), _tRB->un32Size48);
  }

} // namespace vge
#endif
