// ############################################################################################
//  FoggedRenderingFunctions.cc
//  Implementation of fogged rendering functions of CPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "FixedPointConsts.h"
#include "FoggedCeilingLineSegment.h"
#include "Panel3D.h"


namespace vge {

// From "lib/3DEnvironment/FoggedHLineClippingFunctions.cc"
extern void _FogUF_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32F, int32_t &n32FStep);
extern void _FogUG_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                                int32_t &n32B, int32_t &n32BStep, int32_t &n32F, int32_t &n32FStep);
extern void _FogTF_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                                int32_t &n32F, int32_t &n32FStep);
extern void _FogTG_TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                                int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                                int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                                int32_t &n32B, int32_t &n32BStep, int32_t &n32F, int32_t &n32FStep);


// Fogged, untextured, flat shaded
void CPanel3D::FogUFRender()
{
  // Ceiling line segments
  _CFogUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogUF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;

            *renderBuff = xprv::_pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | _tP.tTotal.un8R] |
                          xprv::_pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | _tP.tTotal.un8G] |
                          xprv::_pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | _tP.tTotal.un8B];
          }
          n32Z += n32ZStep;
          n32F += n32FStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, untextured, gouraud shaded
void CPanel3D::FogUGRender()
{
  // Ceiling line segments
  _CFogUGCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Ceiling line segments
  int32_t n32ZStep, n32Z;
  int32_t n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32R     = leftLineSeg.GetR();
        n32RStep = (rightLineSeg.GetR() - n32R) / xWidth;
        n32G     = leftLineSeg.GetG();
        n32GStep = (rightLineSeg.GetG() - n32G) / xWidth;
        n32B     = leftLineSeg.GetB();
        n32BStep = (rightLineSeg.GetB() - n32B) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogUG_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;

            *renderBuff = xprv::_pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | (n32R >> _ISTEP_PREC_)] |
                          xprv::_pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | (n32G >> _ISTEP_PREC_)] |
                          xprv::_pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | (n32B >> _ISTEP_PREC_)];
          }
          n32R += n32RStep;
          n32G += n32GStep;
          n32B += n32BStep;

          n32Z += n32ZStep;
          n32F += n32FStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, colored, not shaded
void CPanel3D::FogCNRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CFogUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogUF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;

            *renderBuff = xprv::_pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | un32R] |
                          xprv::_pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | un32G] |
                          xprv::_pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | un32B];
          }
          n32Z += n32ZStep;
          n32F += n32FStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, colored, flat shaded
void CPanel3D::FogCFRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CFogUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogUF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;

            *renderBuff = xprv::_pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32R << 5) | _tP.tTotal.un8R] |
                          xprv::_pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32G << 5) | _tP.tTotal.un8G] |
                          xprv::_pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32B << 5) | _tP.tTotal.un8B];
          }
          n32Z += n32ZStep;
          n32F += n32FStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, colored, gouraud shaded
void CPanel3D::FogCGRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CFogUGCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32R     = leftLineSeg.GetR();
        n32RStep = (rightLineSeg.GetR() - n32R) / xWidth;
        n32G     = leftLineSeg.GetG();
        n32GStep = (rightLineSeg.GetG() - n32G) / xWidth;
        n32B     = leftLineSeg.GetB();
        n32BStep = (rightLineSeg.GetB() - n32B) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogUG_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;
            *renderBuff = xprv::_pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32R >> _ISTEP_PREC_ << 5) | un32R] |
                          xprv::_pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32G >> _ISTEP_PREC_ << 5) | un32G] |
                          xprv::_pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (n32B >> _ISTEP_PREC_ << 5) | un32B];
          }
          n32R += n32RStep;
          n32G += n32GStep;
          n32B += n32BStep;

          n32Z += n32ZStep;
          n32F += n32FStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, textured, not shaded
void CPanel3D::FogTNRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CFogTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogTF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogColorTable_[(n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, textured, colored
void CPanel3D::FogTCRender(const CTexture &csTexture, uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CFogTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogTF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32R << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32G << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32R << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32G << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (un32B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, textured, flat shaded
void CPanel3D::FogTFRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CFogTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogTF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogLightTable_[(n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Fogged, textured, gouraud shaded
void CPanel3D::FogTGRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CFogTGCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        int32_t zIter = leftLineSeg.GetZ();
        int32_t uIter = leftLineSeg.GetU();
        int32_t vIter = leftLineSeg.GetV();
        int32_t zStep = (rightLineSeg.GetZ() - zIter) / xWidth;
        int32_t uStep = (rightLineSeg.GetU() - uIter) / xWidth;
        int32_t vStep = (rightLineSeg.GetV() - vIter) / xWidth;

        int32_t rIter = leftLineSeg.GetR();
        int32_t gIter = leftLineSeg.GetG();
        int32_t bIter = leftLineSeg.GetB();
        int32_t rStep = (rightLineSeg.GetR() - rIter) / xWidth;
        int32_t gStep = (rightLineSeg.GetG() - gIter) / xWidth;
        int32_t bStep = (rightLineSeg.GetB() - bIter) / xWidth;

        int32_t fIter = leftLineSeg.GetF();
        int32_t fStep = (rightLineSeg.GetF() - fIter) / xWidth;

        // Clip the scan-line
        _FogTG_TGClipHLSeg_(xStart, xEnd, zIter, zStep, uIter, uStep, vIter, vStep, rIter, rStep, gIter, gStep, bIter, bStep, fIter, fStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + size_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + size_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = zStep << _SDDIV_FACT_;
          subUStep = uStep << _SDDIV_FACT_;
          subVStep = vStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          int32_t rzIter  = zIter - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          int32_t tx1Iter = (uIter / rzIter) << _TSTEP_PREC_;
          int32_t ty1Iter = (vIter / rzIter) << _TSTEP_PREC_;

          rzIter += subZStep;
          uIter  += subUStep;
          vIter  += subVStep;

          int32_t tx2Iter = (uIter / rzIter) << _TSTEP_PREC_;
          int32_t ty2Iter = (vIter / rzIter) << _TSTEP_PREC_;

          int32_t txStep =  (tx2Iter - tx1Iter) >> _SDDIV_FACT_;
          int32_t tyStep = ((ty2Iter - ty1Iter) >> _SDDIV_FACT_) << preMultShift;

          int32_t txIter = tx1Iter;
          int32_t tyIter = ty1Iter << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < zIter) {
              *zBuff = zIter;

              const int tbPos = ((tyIter & texMask) + txIter) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogLightTable_[(fIter >> _ISTEP_PREC_ << 10) | (rIter >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogLightTable_[(fIter >> _ISTEP_PREC_ << 10) | (gIter >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogLightTable_[(fIter >> _ISTEP_PREC_ << 10) | (bIter >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            zIter  += zStep;
            txIter += txStep;
            tyIter += tyStep;

            rIter += rStep;
            gIter += gStep;
            bIter += bStep;

            fIter += fStep;

            ++zBuff;
            ++renderBuff;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          int32_t rzIter  = zIter - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          int32_t tx1Iter = (uIter / rzIter) << _TSTEP_PREC_;
          int32_t ty1Iter = (vIter / rzIter) << _TSTEP_PREC_;

          rzIter += (zStep * scanLen);
          uIter  += (uStep * scanLen);
          vIter  += (vStep * scanLen);

          int32_t tx2Iter = (uIter / rzIter) << _TSTEP_PREC_;
          int32_t ty2Iter = (vIter / rzIter) << _TSTEP_PREC_;

          int32_t txStep =  (tx2Iter - tx1Iter) / scanLen;
          int32_t tyStep = ((ty2Iter - ty1Iter) / scanLen) << preMultShift;

          int32_t txIter = tx1Iter;
          int32_t tyIter = ty1Iter << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < zIter) {
              *zBuff = zIter;

              const int tbPos = ((tyIter & texMask) + txIter) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_pun32RFogLightTable_[(fIter >> _ISTEP_PREC_ << 10) | (rIter >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]] |
                            xprv::_pun32GFogLightTable_[(fIter >> _ISTEP_PREC_ << 10) | (gIter >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]] |
                            xprv::_pun32BFogLightTable_[(fIter >> _ISTEP_PREC_ << 10) | (bIter >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            zIter  += zStep;
            txIter += txStep;
            tyIter += tyStep;

            rIter += rStep;
            gIter += gStep;
            bIter += bStep;

            fIter += fStep;

            ++zBuff;
            ++renderBuff;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Transparent, fogged, colored, flat shaded
void CPanel3D::FogCFTransRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CFogUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32F, n32FStep;

  // Calculate the index to the buffer
  rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogUF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t *zBuff    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint8_t *rndrBuff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2];

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;
            rndrBuff[2] = xprv::_pun8TRFogLightTable_[(rndrBuff[2] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (un32R << 5) | _tP.tTotal.un8R];
            rndrBuff[1] = xprv::_pun8TGFogLightTable_[(rndrBuff[1] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (un32G << 5) | _tP.tTotal.un8G];
            rndrBuff[0] = xprv::_pun8TBFogLightTable_[(rndrBuff[0] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (un32B << 5) | _tP.tTotal.un8B];
          }
          n32Z += n32ZStep;
          n32F += n32FStep;

          zBuff++;
          rndrBuff += 4;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Transparent, fogged, textured, flat shaded
void CPanel3D::FogTFTransRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CFogTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogTF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t *zBuff    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint8_t *rndrBuff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2];

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_pun8TRFogLightTable_[(rndrBuff[2] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_pun8TGFogLightTable_[(rndrBuff[1] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_pun8TBFogLightTable_[(rndrBuff[0] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            rndrBuff += 4;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_pun8TRFogLightTable_[(rndrBuff[2] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_pun8TGFogLightTable_[(rndrBuff[1] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_pun8TBFogLightTable_[(rndrBuff[0] << 12) | (n32F >> _ISTEP_PREC_ << 10) | (_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            rndrBuff += 4;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Transparent, fogged, textured, not shaded
void CPanel3D::FogTNTransRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CFogTFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  int32_t n32F, n32FStep;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        n32F     = leftLineSeg.GetF();
        n32FStep = (rightLineSeg.GetF() - n32F) / xWidth;

        // Clip the scan-line
        _FogTF_TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32F, n32FStep);
        xWidth = xEnd - xStart;

        int32_t *zBuff    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint8_t *rndrBuff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2];

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_pun8TRFogColorTable_[(rndrBuff[2] << 7) | (n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_pun8TGFogColorTable_[(rndrBuff[1] << 7) | (n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_pun8TBFogColorTable_[(rndrBuff[0] << 7) | (n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            rndrBuff += 4;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_pun8TRFogColorTable_[(rndrBuff[2] << 7) | (n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_pun8TGFogColorTable_[(rndrBuff[1] << 7) | (n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_pun8TBFogColorTable_[(rndrBuff[0] << 7) | (n32F >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32F += n32FStep;

            zBuff++;
            rndrBuff += 4;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

} // namespace vge
