// ############################################################################################
//  Panel3DBuffer.cc
//  Implementation of CPanel3DBuffer class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "../System/SystemUtilities.h"
#include "Panel3DBuffer.h"

// C++ headers
#include <fstream>


namespace vge {

// How often shall the dynamic lighting be updated
#define _LIGHTING_UPDATE_FRAME_LIMIT_ 10

// VPD header
#define _VPD_ID1_ 'V'
#define _VPD_ID2_ 'P'
#define _VPD_ID3_ 'D'
#define _VPD_VER_ 112
#define _VPD_OV1_ 100
#define _VPD_OV2_ 110
#define _VPD_OV3_ 111

struct _TVPDHeader {
  char     acID[3];         // File ID, must be "VPD"
  uint8_t  un8Version;      // File version
  uint32_t un32VertexCount; // Vertex count
  uint32_t un32PanelCount;  // Panel count
  uint16_t un16Reserved;    // Reserved, must be zero
} _attr_packed;


//
// The 3D panel buffer class
//
CPanel3DBuffer::CPanel3DBuffer()
: _pcsTexture (NULL), _pcsVList    (NULL), _pcsPList(NULL),
  _un32VCount (0),    _un32PCount  (0),
  _un32PCR    (MAX_COLOR_VALUE),
  _un32PCG    (MAX_COLOR_VALUE),
  _un32PCB    (MAX_COLOR_VALUE)
{
}

CPanel3DBuffer::~CPanel3DBuffer() throw()
{
  delete[] _pcsVList;
  delete[] _pcsPList;
}

void CPanel3DBuffer::_CalcAvgNormal_(uint32_t un32VCount, uint32_t un32PCount)
{
  // Temporary variables
  uint32_t un32V, un32P, un32C, un32Tot;
  float    fX, fY, fZ;

  // Walk trough all vertices
  for(un32V = 0; un32V < un32VCount; un32V++) {
    // Initialize variables for summing
    un32Tot = 0; fX = fY = fZ = 0;
    // Search in which panels the current vertex is included
    for(un32P = 0; un32P < un32PCount; un32P++) {
      // Search in the 4 vertices
      for(un32C = 0; un32C < 4; un32C++) {
        if(_pcsPList[un32P]._tP.apcsVPoint[un32C] == &_pcsVList[un32V]) {
          // Accumulate the normal
          fX += _pcsPList[un32P]._tP.tNormal.fX;
          fY += _pcsPList[un32P]._tP.tNormal.fY;
          fZ += _pcsPList[un32P]._tP.tNormal.fZ;
          un32Tot++;
          break;
        }
      }
    }
    // Calculate and store the average normal
    _pcsVList[un32V]._tAvgN.fX = fX / float(un32Tot);
    _pcsVList[un32V]._tAvgN.fY = fY / float(un32Tot);
    _pcsVList[un32V]._tAvgN.fZ = fZ / float(un32Tot);
  }
}

void CPanel3DBuffer::FSResetAmbientLight(TAmbientLight *ptVAmbientLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
  }
}

void CPanel3DBuffer::GSResetAmbientLight(TAmbientLight *ptVAmbientLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._GSCalcInten_();
  }
}

void CPanel3DBuffer::FSResetOmniLight(TOmniLight *ptVOmniLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
  }
}

void CPanel3DBuffer::GSResetOmniLight(TOmniLight *ptVOmniLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._GSCalcInten_();
  }
}

void CPanel3DBuffer::FSResetLights(TAmbientLight *ptVAmbientLight, TOmniLight *ptVOmniLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
  }
}

void CPanel3DBuffer::GSResetLights(TAmbientLight *ptVAmbientLight, TOmniLight *ptVOmniLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._GSCalcInten_();
  }
}

void CPanel3DBuffer::FSGSResetAmbientLight(TAmbientLight *ptVAmbientLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
    _pcsPList[un32Cnt]._GSCalcInten_();
  }
}

void CPanel3DBuffer::FSGSResetOmniLight(TOmniLight *ptVOmniLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
    _pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._GSCalcInten_();
  }
}

void CPanel3DBuffer::FSGSResetLights(TAmbientLight *ptVAmbientLight, TOmniLight *ptVOmniLight)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
    _pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._GSCalcInten_();
  }
}

void CPanel3DBuffer::CalcDefTexCoords()
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    _pcsPList[un32Cnt]._tP.atUV[0].fU = 0.0f;
    _pcsPList[un32Cnt]._tP.atUV[0].fV = 1.0f;
    _pcsPList[un32Cnt]._tP.atUV[1].fU = 0.0f;
    _pcsPList[un32Cnt]._tP.atUV[1].fV = 0.0f;
    _pcsPList[un32Cnt]._tP.atUV[2].fU = 1.0f;
    _pcsPList[un32Cnt]._tP.atUV[2].fV = 0.0f;
    _pcsPList[un32Cnt]._tP.atUV[3].fU = 1.0f;
    _pcsPList[un32Cnt]._tP.atUV[3].fV = 1.0f;
  }
}

void CPanel3DBuffer::Transform(const CMatrix3D &csTMatrix)
{
  // Transform the vertex list
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    csTMatrix.Transform(_pcsVList[un32Cnt]);

  // Update the panel's properties
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    csTMatrix.Transform(cp._tP.tBNormal);
    cp.UpdateFrame();
  }
}

void CPanel3DBuffer::DynTransform(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex list
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    csCameraMatrix.PstDynTransform(_pcsVList[un32Cnt]);

  // Update the panel's properties
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    csCameraMatrix.PstDynTransform(cp._tP.tBNormal);
    cp.UpdateFrame();
  }
}

void CPanel3DBuffer::PreDynNSRender(const CMatrix3D &csDynMatrix, const TOmniLight *ptVOmniLightUsed)
{
  UNUSED_PARAM(ptVOmniLightUsed);

  // Transform the vertex list
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    csDynMatrix.PreDynTransform(_pcsVList[un32Cnt]);

  // Transform the panel's normal
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++)
    csDynMatrix.PreDynTransform(_pcsPList[un32Cnt]._tP.tBNormal);
}

void CPanel3DBuffer::PreDynFSRender(const CMatrix3D &csDynMatrix, const TOmniLight *ptVOmniLightUsed)
{
  // Static variables
  static int8_t n8DFSCnt = 0;

  // Transform the vertex list
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    csDynMatrix.PreDynTransform(_pcsVList[un32Cnt]);

  // Transform the panel's normal
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++)
    csDynMatrix.PreDynTransform(_pcsPList[un32Cnt]._tP.tBNormal);

  // Update the lighting once each _LIGHTING_UPDATE_FRAME_LIMIT_ frames
  if(n8DFSCnt <= _LIGHTING_UPDATE_FRAME_LIMIT_) {
    n8DFSCnt++;
    return;
  }
  n8DFSCnt = 0;
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    cp._DynFSCalcIntenOmni_(ptVOmniLightUsed);
    cp._FSCalcInten_();
  }
}

void CPanel3DBuffer::PreDynGSRender(const CMatrix3D &csDynMatrix, const TOmniLight *ptVOmniLightUsed)
{
  // Static variables
  static int8_t n8DGSCnt = 0;

  // Transform the vertex list
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    csDynMatrix.PreDynTransform(_pcsVList[un32Cnt]);

  // Transform the panel's normal
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++)
    csDynMatrix.PreDynTransform(_pcsPList[un32Cnt]._tP.tBNormal);

  // Update the lighting once each _LIGHTING_UPDATE_FRAME_LIMIT_ frames
  if(n8DGSCnt <= _LIGHTING_UPDATE_FRAME_LIMIT_) {
    n8DGSCnt++;
    return;
  }
  n8DGSCnt = 0;
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    cp._DynGSCalcIntenOmni_(ptVOmniLightUsed);
    cp._GSCalcInten_();
  }
}

void CPanel3DBuffer::UFRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UFProject();
      if(cp.IsVisible2D()) cp.UFRender();
    }
  }
}

void CPanel3DBuffer::UGRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UGProject();
      if(cp.IsVisible2D()) cp.UGRender();
    }
  }
}

void CPanel3DBuffer::CNRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UFProject();
      if(cp.IsVisible2D()) cp.CNRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::CFRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UFProject();
      if(cp.IsVisible2D()) cp.CFRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::CGRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UGProject();
      if(cp.IsVisible2D()) cp.CGRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::TNRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TNRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::TCRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TCRender(*_pcsTexture, _un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::TFRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TFRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::TGRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TGProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TGRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::DynUFRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UFProject();
      if(cp.IsVisible2D()) cp.UFRender();
    }
  }
}

void CPanel3DBuffer::DynUGRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UGProject();
      if(cp.IsVisible2D()) cp.UGRender();
    }
  }
}

void CPanel3DBuffer::DynCFRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UFProject();
      if(cp.IsVisible2D()) cp.CFRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::DynCGRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UGProject();
      if(cp.IsVisible2D()) cp.CGRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::DynTNRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TNRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::DynTFRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TFRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::DynTGRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TGProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TGRender(*_pcsTexture);
    }
  }
}


void CPanel3DBuffer::FogUFRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUFProject();
      if(cp.IsVisible2D()) cp.FogUFRender();
    }
  }
}

void CPanel3DBuffer::FogUGRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUGProject();
      if(cp.IsVisible2D()) cp.FogUGRender();
    }
  }
}

void CPanel3DBuffer::FogCNRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUFProject();
      if(cp.IsVisible2D()) cp.FogCNRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::FogCFRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUFProject();
      if(cp.IsVisible2D()) cp.FogCFRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::FogCGRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUGProject();
      if(cp.IsVisible2D()) cp.FogCGRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::FogTNRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTNRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::FogTCRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTCRender(*_pcsTexture, _un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::FogTFRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTFRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::FogTGRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTGProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTGRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::DynFogUFRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUFProject();
      if(cp.IsVisible2D()) cp.FogUFRender();
    }
  }
}

void CPanel3DBuffer::DynFogUGRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUGProject();
      if(cp.IsVisible2D()) cp.FogUGRender();
    }
  }
}

void CPanel3DBuffer::DynFogCFRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUFProject();
      if(cp.IsVisible2D()) cp.FogCFRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::DynFogCGRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUGProject();
      if(cp.IsVisible2D()) cp.FogCGRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::DynFogTNRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTNRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::DynFogTFRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTFRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::DynFogTGRender(const CMatrix3D &csCameraMatrix)
{
  // Transform the vertex points
  DynTransform(csCameraMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTGProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTGRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::CFTransRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.UFProject();
      if(cp.IsVisible2D()) cp.CFTransRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::TFTransRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TFTransRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::TNTransRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.TFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.TNTransRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::FogCFTransRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogUFProject();
      if(cp.IsVisible2D()) cp.FogCFTransRender(_un32PCR, _un32PCG, _un32PCB);
    }
  }
}

void CPanel3DBuffer::FogTFTransRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTFTransRender(*_pcsTexture);
    }
  }
}

void CPanel3DBuffer::FogTNTransRender(const CMatrix3D &csTMatrix)
{
  // Transform the vertex points
  Transform(csTMatrix);

  // Render the panels
  for(uint32_t un32Cnt = 0; un32Cnt < _un32PCount; un32Cnt++) {
    CPanel3D &cp = _pcsPList[un32Cnt];
    if(cp.Invisible()) continue;
    if(cp.IsVisible3D()) {
      cp.FogTFProject(*_pcsTexture);
      if(cp.IsVisible2D()) cp.FogTNTransRender(*_pcsTexture);
    }
  }
}

ECollType CPanel3DBuffer::PlaneCollide(float afNewXPos, float afNewYPos, float afNewZPos, float fViewRadius) const
{
  // Check for collision with all panels
  for(uint32_t i = 0; i < _un32PCount; i++) {
    // Calculate the new position relative to the panel's center
    float rX = afNewXPos - _pcsPList[i]._tP.afCenter[0];
    float rY = afNewYPos - _pcsPList[i]._tP.afCenter[1];
    float rZ = afNewZPos - _pcsPList[i]._tP.afCenter[2];
    // Do rough test first
    float aD = min(min(fabs(rX), fabs(rY)), fabs(rZ));
    if(aD > (C3DEnv::Active3DEnv()->_un32MinZ + fViewRadius)) {
      //VPrint("SKIP   : %+03d %+9.2f %+9.2f %+9.2f %+9.2f\n", i, rX, rY, rZ, aD);
      continue;
    }
    // If rough test pass, perform more accurate distance calculation
    float mX = rX * _pcsPList[i]._tP.tNormal.fX;
    float mY = rY * _pcsPList[i]._tP.tNormal.fY;
    float mZ = rZ * _pcsPList[i]._tP.tNormal.fZ;
    float mD = sqrt(mX*mX + mY*mY + mZ*mZ);
    // Check for collision again
    if(mD <= (C3DEnv::Active3DEnv()->_un32MinZ + fViewRadius)) {
      // Perform dot product between the up vector (0, 1, 0) with the normal
      //   fS = 0.0f * _pcsPList[i]._tP.tNormal.fX +
      //        1.0f * _pcsPList[i]._tP.tNormal.fY +
      //        0.0f * _pcsPList[i]._tP.tNormal.fZ
      //      = _pcsPList[i]._tP.tNormal.fY
      if(_pcsPList[i]._tP.tNormal.fY > 0.707107f) {
        //VPrint("Floor  : %+03d %+9.2f %+9.2f %+9.2f %+9.2f\n", i, mX, mY, mZ, mD);
        return(ECT_Floor);
      }
      else if(_pcsPList[i]._tP.tNormal.fY < -0.707107f) {
        //VPrint("Ceiling: %+03d %+9.2f %+9.2f %+9.2f %+9.2f\n", i, mX, mY, mZ, mD);
        return(ECT_Ceiling);
      }
      else {
        //VPrint("Wall   : %+03d %+9.2f %+9.2f %+9.2f %+9.2f\n", i, mX, mY, mZ, mD);
        return(ECT_Wall);
      }
    }
  }
  return(ECT_None);
}

void CPanel3DBuffer::GetBoundingBox(TFloatXYZ &minCoord, TFloatXYZ &maxCoord) const
{
  minCoord.x = minCoord.y = minCoord.z =  numeric_limits<float>::max();
  maxCoord.x = maxCoord.y = maxCoord.z = -numeric_limits<float>::max();

  const CPoint3D *v = _pcsVList;
  for(uint32_t i = 0; i < _un32VCount; ++i) {
    const float x = v->GetLocalX();
    const float y = v->GetLocalY();
    const float z = v->GetLocalZ();
    v++;

    if(x < minCoord.x) minCoord.x = x; if(x > maxCoord.x) maxCoord.x = x;
    if(y < minCoord.y) minCoord.y = y; if(y > maxCoord.y) maxCoord.y = y;
    if(z < minCoord.z) minCoord.z = z; if(z > maxCoord.z) maxCoord.z = z;
  }
}

void CPanel3DBuffer::InitFromMemory(const float *pfVertices, uint32_t un32VertexCount,
                                    const uint32_t *pfPanelPoints, uint32_t un32PanelCount,
                                    bool bEnableSmoothShading,
                                    bool bUseAvgNormalForSmoothShading,
                                    const TAmbientLight *ptVAmbientLight,
                                    const TOmniLight *ptVOmniLight)
{
  // Temporary variables
  uint32_t un32Cnt, un32Index;

  // Destroy previous
  delete[] _pcsVList;
  delete[] _pcsPList;
  _pcsVList   = NULL;
  _pcsPList   = NULL;
  _un32VCount = 0;
  _un32PCount = 0;

  // Allocate memory
  _pcsVList = _nothrow_new CPoint3D[un32VertexCount];
  if(_pcsVList == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for vertex buffer");
  _pcsPList = _nothrow_new CPanel3D[un32PanelCount];
  if(_pcsPList == NULL) {
    delete[] _pcsVList;
    _pcsVList = NULL;
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for panel buffer");
  }

  // Copy vertex data
  un32Index = 0;
  for(un32Cnt = 0; un32Cnt < un32VertexCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX = pfVertices[un32Index++];
    _pcsVList[un32Cnt]._tLocal.fY = pfVertices[un32Index++];
    _pcsVList[un32Cnt]._tLocal.fZ = pfVertices[un32Index++];
    //if(fabs(_pcsVList[un32Cnt]._tLocal.fZ) < 0.00001f) _pcsVList[un32Cnt]._tLocal.fZ = 0.00001f;
  }

  // Generate panel data
  un32Index = 0;
  for(un32Cnt = 0; un32Cnt < un32PanelCount; un32Cnt++) {
    _pcsPList[un32Cnt]._tP.apcsVPoint[0] = &_pcsVList[pfPanelPoints[un32Index++]];
    _pcsPList[un32Cnt]._tP.apcsVPoint[1] = &_pcsVList[pfPanelPoints[un32Index++]];
    _pcsPList[un32Cnt]._tP.apcsVPoint[2] = &_pcsVList[pfPanelPoints[un32Index++]];
    _pcsPList[un32Cnt]._tP.apcsVPoint[3] = &_pcsVList[pfPanelPoints[un32Index++]];

    _pcsPList[un32Cnt]._CalcRadius_();
    _pcsPList[un32Cnt]._CalcNormal_();
    _pcsPList[un32Cnt]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[un32Cnt]._FSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[un32Cnt]._FSCalcInten_();
  }
  if(bEnableSmoothShading) {
    _CalcAvgNormal_(un32VertexCount, un32PanelCount);
    for(un32Cnt = 0; un32Cnt < un32PanelCount; un32Cnt++) {
      _pcsPList[un32Cnt]._GSCalcIntenOmni_(ptVOmniLight, bUseAvgNormalForSmoothShading);
      _pcsPList[un32Cnt]._GSCalcInten_();
    }
  }

  // OK !
  _un32VCount = un32VertexCount;
  _un32PCount = un32PanelCount;
}
void CPanel3DBuffer::ReinitShading(bool bEnableSmoothShading,
                                   bool bUseAvgNormalForSmoothShading,
                                   const TAmbientLight *ptVAmbientLight,
                                   const TOmniLight *ptVOmniLight)
{
  // Re-generate panel data
  for(uint32_t i = 0; i < _un32PCount; i++) {
    _pcsPList[i]._CalcRadius_();
    _pcsPList[i]._CalcNormal_();
    _pcsPList[i]._CalcIntenAmbient_(ptVAmbientLight);
    _pcsPList[i]._FSCalcIntenOmni_(ptVOmniLight);
    _pcsPList[i]._FSCalcInten_();
  }
  if(bEnableSmoothShading) {
    _CalcAvgNormal_(_un32VCount, _un32PCount);
    for(uint32_t i = 0; i < _un32PCount; i++) {
      _pcsPList[i]._GSCalcIntenOmni_(ptVOmniLight, bUseAvgNormalForSmoothShading);
      _pcsPList[i]._GSCalcInten_();
    }
  }
}

CPanel3D *CPanel3DBuffer::_GetPanel_(uint32_t un32Index) const
{
  if(un32Index >= _un32PCount) {
    VError(VERR_InvalidParameter, _NFL_, "The given panel index ('un32Index') exceeds the maximum panel index");
    return(NULL);
  }
  return(&_pcsPList[un32Index]);
}

void CPanel3DBuffer::LocalXScale(float fXAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX *= fXAmount;
  }
}

void CPanel3DBuffer::LocalYScale(float fYAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fY *= fYAmount;
  }
}

void CPanel3DBuffer::LocalZScale(float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fZ *= fZAmount;
  }
}

void CPanel3DBuffer::LocalXYScale(float fXAmount, float fYAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX *= fXAmount;
    _pcsVList[un32Cnt]._tLocal.fY *= fYAmount;
  }
}

void CPanel3DBuffer::LocalXZScale(float fXAmount, float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX *= fXAmount;
    _pcsVList[un32Cnt]._tLocal.fZ *= fZAmount;
  }
}

void CPanel3DBuffer::LocalYZScale(float fYAmount, float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fY *= fYAmount;
    _pcsVList[un32Cnt]._tLocal.fZ *= fZAmount;
  }
}

void CPanel3DBuffer::LocalXYZScale(float fXAmount, float fYAmount, float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX *= fXAmount;
    _pcsVList[un32Cnt]._tLocal.fY *= fYAmount;
    _pcsVList[un32Cnt]._tLocal.fZ *= fZAmount;
  }
}

void CPanel3DBuffer::LocalXTranslate(float fXAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    _pcsVList[un32Cnt]._tLocal.fX += fXAmount;
}

void CPanel3DBuffer::LocalYTranslate(float fYAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    _pcsVList[un32Cnt]._tLocal.fY += fYAmount;
}

void CPanel3DBuffer::LocalZTranslate(float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++)
    _pcsVList[un32Cnt]._tLocal.fZ += fZAmount;
}

void CPanel3DBuffer::LocalXYTranslate(float fXAmount, float fYAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX += fXAmount;
    _pcsVList[un32Cnt]._tLocal.fY += fYAmount;
  }
}

void CPanel3DBuffer::LocalXZTranslate(float fXAmount, float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX += fXAmount;
    _pcsVList[un32Cnt]._tLocal.fZ += fZAmount;
  }
}

void CPanel3DBuffer::LocalYZTranslate(float fYAmount, float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fY += fYAmount;
    _pcsVList[un32Cnt]._tLocal.fZ += fZAmount;
  }
}

void CPanel3DBuffer::LocalXYZTranslate(float fXAmount, float fYAmount, float fZAmount)
{
  for(uint32_t un32Cnt = 0; un32Cnt < _un32VCount; un32Cnt++) {
    _pcsVList[un32Cnt]._tLocal.fX += fXAmount;
    _pcsVList[un32Cnt]._tLocal.fY += fYAmount;
    _pcsVList[un32Cnt]._tLocal.fZ += fZAmount;
  }
}

void CPanel3DBuffer::_FindCoordMinMax_(float &xmin, float &xmax, float &ymin, float &ymax, float &zmin, float &zmax)
{
  xmin =  numeric_limits<float>::max();
  xmax = -xmin;
  ymin =  xmin;
  ymax = -xmin;
  zmin =  xmin;
  zmax = -xmin;

  for(uint32_t i = 0; i < _un32VCount; ++i) {
    const float &cx = _pcsVList[i]._tLocal.fX;
    const float &cy = _pcsVList[i]._tLocal.fY;
    const float &cz = _pcsVList[i]._tLocal.fZ;

    if(cx < xmin) xmin = cx; if(cx > xmax) xmax = cx;
    if(cy < ymin) ymin = cy; if(cy > ymax) ymax = cy;
    if(cz < zmin) zmin = cz; if(cz > zmax) zmax = cz;
  }
}

void CPanel3DBuffer::LocalRegionTranslate(ELRTMode eXMode, ELRTMode eYMode, ELRTMode eZMode)
{
  // Just return if no vertices to process
  if(_un32VCount == 0) return;

  // Find the minimum and maximum coordinates
  float xmin, xmax, ymin, ymax, zmin, zmax;
  _FindCoordMinMax_(xmin, xmax, ymin, ymax, zmin, zmax);

  // Calculate the delta
  const float xdlt = xmax - xmin;
  const float ydlt = ymax - ymin;
  const float zdlt = zmax - zmin;

  // Calculate the shift amount
  float xs, ys, zs;
  switch(eXMode) {
    case ELRTM_Balanced: xs = -xmin - xdlt/2; break;
    case ELRTM_Negative: xs = -xmax;          break;
    case ELRTM_Positive: xs = -xmin;          break;
    default            : xs = 0.0f;
  }
  switch(eYMode) {
    case ELRTM_Balanced: ys = -ymin - ydlt/2; break;
    case ELRTM_Negative: ys = -ymax;          break;
    case ELRTM_Positive: ys = -ymin;          break;
    default            : ys = 0.0f;
  }
  switch(eZMode) {
    case ELRTM_Balanced: zs = -zmin - zdlt/2; break;
    case ELRTM_Negative: zs = -zmax;          break;
    case ELRTM_Positive: zs = -zmin;          break;
    default            : zs = 0.0f;
  }

  // Translate the vertices
  LocalXYZTranslate(xs, ys, zs);
}

void CPanel3DBuffer::LocalRegionTranslateAllBalanced()
{
  // Just return if no vertices to process
  if(_un32VCount == 0) return;

  // Find the minimum and maximum coordinates
  float xmin, xmax, ymin, ymax, zmin, zmax;
  _FindCoordMinMax_(xmin, xmax, ymin, ymax, zmin, zmax);

  // Calculate the shift amount
  const float xs = -xmin - (xmax-xmin)/2;
  const float ys = -ymin - (ymax-ymin)/2;
  const float zs = -zmin - (zmax-zmin)/2;

  // Translate the vertices
  LocalXYZTranslate(xs, ys, zs);
}

void CPanel3DBuffer::LocalRegionTranslateCenterOfWeight()
{
  // Translate to positive region frist
  LocalRegionTranslate(ELRTM_Positive, ELRTM_Positive, ELRTM_Positive);

  // Calculate the cumulative coordinates
  float xcum = 0;
  float ycum = 0;
  float zcum = 0;
  for(uint32_t i = 0; i < _un32VCount; ++i) {
    xcum += _pcsVList[i]._tLocal.fX;
    ycum += _pcsVList[i]._tLocal.fY;
    zcum += _pcsVList[i]._tLocal.fZ;
  }

  // Translate the vertices
  LocalXYZTranslate(-xcum/_un32VCount, -ycum/_un32VCount, -zcum/_un32VCount);
}

void CPanel3DBuffer::SaveVPD(const CDataSink &sink) const
{
  // Check
  if((_un32VCount == 0) || (_un32PCount == 0))
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CPanel3DBuffer is not initialized)");

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Make header
  _TVPDHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VPD_ID1_;
  tHeader.acID[1]         = _VPD_ID2_;
  tHeader.acID[2]         = _VPD_ID3_;
  tHeader.un8Version      = _VPD_VER_;
  tHeader.un32VertexCount = _un32VCount;
  tHeader.un32PanelCount  = _un32PCount;

  // Write all the data
  try {
    // Write the header
    os.write(reinterpret_cast<const char*>(&tHeader), sizeof(tHeader));
  
    // Write vertex data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32VCount; un32Cnt1++) {
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fX),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fX));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fY),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fY));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fZ),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fZ));
    }
  
    // Write panel data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        for(uint32_t un32Cnt3 = 0; un32Cnt3 < _un32VCount; un32Cnt3++) {
          if(_pcsPList[un32Cnt1]._tP.apcsVPoint[un32Cnt2] == &_pcsVList[un32Cnt3]) {
            os.write(reinterpret_cast<const char*>(&un32Cnt3), sizeof(un32Cnt3));
            break;
          }
        }
      }
    }
  
    // Write panel's color
    os.write(reinterpret_cast<const char*>(&_un32PCR), sizeof(_un32PCR));
    os.write(reinterpret_cast<const char*>(&_un32PCG), sizeof(_un32PCG));
    os.write(reinterpret_cast<const char*>(&_un32PCB), sizeof(_un32PCB));
  
    // Write texture coordinates
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        os.write(reinterpret_cast<const char*>(&_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU),
                 sizeof(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU));
        os.write(reinterpret_cast<const char*>(&_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV),
                 sizeof(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV));
      }
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CPanel3DBuffer::LoadVPD(const CDataSource &source,
                             bool bEnableSmoothShading,
                             bool bUseAvgNormalForSmoothShading,
                             TAmbientLight *ptVAmbientLight,
                             TOmniLight *ptVOmniLight)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Temporary variables
  _TVPDHeader      tHeader;
  vector<float>    vertices;
  vector<uint32_t> panels;

  // Read all the data (except the texture coordinates)
  try {
    // Peek header
    is.peek(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  
    // Check header
    if((tHeader.acID[0] != _VPD_ID1_) ||
       (tHeader.acID[1] != _VPD_ID2_) || (tHeader.acID[2] != _VPD_ID3_))
      VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VPD stream", strmName);
    if(tHeader.un8Version == _VPD_OV1_) {
      __LoadVPD100(source);
      return;
    }
    if(tHeader.un8Version == _VPD_OV2_) {
      __LoadVPD110(source);
      return;
    }
    if(tHeader.un8Version == _VPD_OV3_) {
      __LoadVPD111(source);
      return;
    }
    if(tHeader.un8Version != _VPD_VER_)
      VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
    if(tHeader.un16Reserved)
      VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

    // Discard the already processed header data from the stream
    is.discard(sizeof(tHeader));
  
    // Allocate temporary memory
    vertices.resize(tHeader.un32VertexCount * 3);
    panels.resize(tHeader.un32PanelCount * 4);
    
    // Read vertex and panel data
    is.read(reinterpret_cast<char*>(&vertices[0]), tHeader.un32VertexCount * 3 * sizeof(float));
    is.read(reinterpret_cast<char*>(&panels[0]), tHeader.un32PanelCount * 4 * sizeof(uint32_t));
  
    // Read panel's color
    is.read(reinterpret_cast<char*>(&_un32PCR), sizeof(_un32PCR));
    is.read(reinterpret_cast<char*>(&_un32PCG), sizeof(_un32PCG));
    is.read(reinterpret_cast<char*>(&_un32PCB), sizeof(_un32PCB));
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }

  //  Generate panels
  InitFromMemory(&vertices[0], tHeader.un32VertexCount, &panels[0], tHeader.un32PanelCount,
                 bEnableSmoothShading, bUseAvgNormalForSmoothShading, ptVAmbientLight, ptVOmniLight);

  // Read texture coordinates
  try {
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        is.read(reinterpret_cast<char*>(&_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU),
                sizeof(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU));
        is.read(reinterpret_cast<char*>(&_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV),
                sizeof(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV));
      }
    }
  }
  catch(...) {
    CalcDefTexCoords();
    VWarning(_NFL_, "Couldn't read texture coordinates from stream \"%s\"\nUsing the default texture coordinates", strmName);
  }
}

void CPanel3DBuffer::__SaveVPD100(const CDataSink &sink) const
{
  // Check
  if((_un32VCount == 0) || (_un32PCount == 0))
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CPanel3DBuffer is not initialized)");

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Make header
  _TVPDHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VPD_ID1_;
  tHeader.acID[1]         = _VPD_ID2_;
  tHeader.acID[2]         = _VPD_ID3_;
  tHeader.un8Version      = _VPD_OV1_;
  tHeader.un32VertexCount = _un32VCount;
  tHeader.un32PanelCount  = _un32PCount;

  // Write all the data
  try {
    // Write the header
    os.write(reinterpret_cast<const char*>(&tHeader), sizeof(tHeader));
  
    // Write vertex data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32VCount; un32Cnt1++) {
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fX),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fX));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fY),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fY));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fZ),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fZ));
    }

    // Write panel data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        for(uint32_t un32Cnt3 = 0; un32Cnt3 < _un32VCount; un32Cnt3++) {
          if(_pcsPList[un32Cnt1]._tP.apcsVPoint[un32Cnt2] == &_pcsVList[un32Cnt3]) {
            os.write(reinterpret_cast<const char*>(&un32Cnt3), sizeof(un32Cnt3));
            break;
          }
        }
      }
    }
  
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CPanel3DBuffer::__LoadVPD100(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Temporary variables
  _TVPDHeader      tHeader;
  vector<float>    vertices;
  vector<uint32_t> panels;

  // Read all the data
  try {
    // Read header
    is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  
    // Check header
    if((tHeader.acID[0] != _VPD_ID1_) ||
       (tHeader.acID[1] != _VPD_ID2_) || (tHeader.acID[2] != _VPD_ID3_))
      VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VPD stream", strmName);
    if(tHeader.un8Version != _VPD_OV1_)
      VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
    if(tHeader.un16Reserved)
      VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

    // Allocate temporary memory
    vertices.resize(tHeader.un32VertexCount * 3);
    panels.resize(tHeader.un32PanelCount * 4);
    
    // Read vertex and panel data
    is.read(reinterpret_cast<char*>(&vertices[0]), tHeader.un32VertexCount * 3 * sizeof(float));
    is.read(reinterpret_cast<char*>(&panels[0]), tHeader.un32PanelCount * 4 * sizeof(uint32_t));
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }

  //  Generate panels
  InitFromMemory(&vertices[0], tHeader.un32VertexCount, &panels[0], tHeader.un32PanelCount);
  CalcDefTexCoords();
}

void CPanel3DBuffer::__SaveVPD110(const CDataSink &sink) const
{
  // Check
  if((_un32VCount == 0) || (_un32PCount == 0))
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CPanel3DBuffer is not initialized)");

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Make header
  _TVPDHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VPD_ID1_;
  tHeader.acID[1]         = _VPD_ID2_;
  tHeader.acID[2]         = _VPD_ID3_;
  tHeader.un8Version      = _VPD_OV2_;
  tHeader.un32VertexCount = _un32VCount;
  tHeader.un32PanelCount  = _un32PCount;

  // Write all the data
  try {
    // Write the header
    os.write(reinterpret_cast<const char*>(&tHeader), sizeof(tHeader));
  
    // Write vertex data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32VCount; un32Cnt1++) {
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fX),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fX));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fY),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fY));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fZ),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fZ));
    }
  
    // Write panel data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        for(uint32_t un32Cnt3 = 0; un32Cnt3 < _un32VCount; un32Cnt3++) {
          if(_pcsPList[un32Cnt1]._tP.apcsVPoint[un32Cnt2] == &_pcsVList[un32Cnt3]) {
            os.write(reinterpret_cast<const char*>(&un32Cnt3), sizeof(un32Cnt3));
            break;
          }
        }
      }
    }

    // Calculate the texture coordinates maximum values
    uint32_t w = (_pcsTexture == NULL) ? 256 : _pcsTexture->XMax();
    uint32_t h = (_pcsTexture == NULL) ? 256 : _pcsTexture->YMax();

    // Write texture coordinates
    os.write(reinterpret_cast<const char*>(&w), sizeof(w));
    os.write(reinterpret_cast<const char*>(&h), sizeof(h));
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        int32_t u = int32_t(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU * w);
        int32_t v = int32_t(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV * h);
        os.write(reinterpret_cast<const char*>(&u), sizeof(u));
        os.write(reinterpret_cast<const char*>(&v), sizeof(v));
      }
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CPanel3DBuffer::__LoadVPD110(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Temporary variables
  _TVPDHeader      tHeader;
  vector<float>    vertices;
  vector<uint32_t> panels;

  // Read all the data (except the texture coordinates)
  try {
    // Peek header
    is.peek(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));

    // Check header
    if((tHeader.acID[0] != _VPD_ID1_) ||
       (tHeader.acID[1] != _VPD_ID2_) || (tHeader.acID[2] != _VPD_ID3_))
      VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VPD stream", strmName);
    if(tHeader.un8Version == _VPD_OV1_) {
      __LoadVPD100(source);
      return;
    }
    if(tHeader.un8Version != _VPD_OV2_)
      VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
    if(tHeader.un16Reserved)
      VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

    // Discard the already processed header data from the stream
    is.discard(sizeof(tHeader));

    // Allocate temporary memory
    vertices.resize(tHeader.un32VertexCount * 3);
    panels.resize(tHeader.un32PanelCount * 4);

    // Read vertex and panel data
    is.read(reinterpret_cast<char*>(&vertices[0]), tHeader.un32VertexCount * 3 * sizeof(float));
    is.read(reinterpret_cast<char*>(&panels[0]), tHeader.un32PanelCount * 4 * sizeof(uint32_t));
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }

  //  Generate panels
  InitFromMemory(&vertices[0], tHeader.un32VertexCount, &panels[0], tHeader.un32PanelCount);

  // Read texture coordinates
  try {
    uint32_t w, h;
    is.read(reinterpret_cast<char*>(&w), sizeof(w));
    is.read(reinterpret_cast<char*>(&h), sizeof(h));
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        int32_t u, v;
        is.read(reinterpret_cast<char*>(&u), sizeof(u));
        is.read(reinterpret_cast<char*>(&v), sizeof(v));
        _pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU = float(u) / float(w);
        _pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV = float(v) / float(h);
      }
    }
  }
  catch(...) {
    CalcDefTexCoords();
    VWarning(_NFL_, "Couldn't read texture coordinates from stream \"%s\"\nUsing the default texture coordinates", strmName);
  }
}

void CPanel3DBuffer::__SaveVPD111(const CDataSink &sink) const
{
  // Check
  if((_un32VCount == 0) || (_un32PCount == 0))
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CPanel3DBuffer is not initialized)");

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Make header
  _TVPDHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VPD_ID1_;
  tHeader.acID[1]         = _VPD_ID2_;
  tHeader.acID[2]         = _VPD_ID3_;
  tHeader.un8Version      = _VPD_OV3_;
  tHeader.un32VertexCount = _un32VCount;
  tHeader.un32PanelCount  = _un32PCount;

  // Write all the data
  try {
    // Write the header
    os.write(reinterpret_cast<const char*>(&tHeader), sizeof(tHeader));
  
    // Write vertex data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32VCount; un32Cnt1++) {
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fX),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fX));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fY),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fY));
      os.write(reinterpret_cast<const char*>(&_pcsVList[un32Cnt1]._tLocal.fZ),
               sizeof(_pcsVList[un32Cnt1]._tLocal.fZ));
    }
  
    // Write panel data
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        for(uint32_t un32Cnt3 = 0; un32Cnt3 < _un32VCount; un32Cnt3++) {
          if(_pcsPList[un32Cnt1]._tP.apcsVPoint[un32Cnt2] == &_pcsVList[un32Cnt3]) {
            os.write(reinterpret_cast<const char*>(&un32Cnt3), sizeof(un32Cnt3));
            break;
          }
        }
      }
    }

    // Write panel's color
    os.write(reinterpret_cast<const char*>(&_un32PCR), sizeof(_un32PCR));
    os.write(reinterpret_cast<const char*>(&_un32PCG), sizeof(_un32PCG));
    os.write(reinterpret_cast<const char*>(&_un32PCB), sizeof(_un32PCB));
    
    // Calculate the texture coordinates maximum values
    uint32_t w = (_pcsTexture == NULL) ? 256 : _pcsTexture->XMax();
    uint32_t h = (_pcsTexture == NULL) ? 256 : _pcsTexture->YMax();

    // Write texture coordinates
    os.write(reinterpret_cast<const char*>(&w), sizeof(w));
    os.write(reinterpret_cast<const char*>(&h), sizeof(h));
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        int32_t u = int32_t(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU * w);
        int32_t v = int32_t(_pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV * h);
        os.write(reinterpret_cast<const char*>(&u), sizeof(u));
        os.write(reinterpret_cast<const char*>(&v), sizeof(v));
      }
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CPanel3DBuffer::__LoadVPD111(const CDataSource &source,
                                  bool bEnableSmoothShading,
                                  bool bUseAvgNormalForSmoothShading,
                                  TAmbientLight *ptVAmbientLight,
                                  TOmniLight *ptVOmniLight)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Temporary variables
  _TVPDHeader      tHeader;
  vector<float>    vertices;
  vector<uint32_t> panels;

  // Read all the data (except the texture coordinates)
  try {
    // Peek header
    is.peek(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));

    // Check header
    if((tHeader.acID[0] != _VPD_ID1_) ||
       (tHeader.acID[1] != _VPD_ID2_) || (tHeader.acID[2] != _VPD_ID3_))
      VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VPD stream", strmName);
    if(tHeader.un8Version == _VPD_OV1_) {
      __LoadVPD100(source);
      return;
    }
    if(tHeader.un8Version == _VPD_OV2_) {
      __LoadVPD110(source);
      return;
    }
    if(tHeader.un8Version != _VPD_OV3_)
      VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
    if(tHeader.un16Reserved)
      VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

    // Discard the already processed header data from the stream
    is.discard(sizeof(tHeader));

    // Allocate temporary memory
    vertices.resize(tHeader.un32VertexCount * 3);
    panels.resize(tHeader.un32PanelCount * 4);
    
    // Read vertex and panel data
    is.read(reinterpret_cast<char*>(&vertices[0]), tHeader.un32VertexCount * 3 * sizeof(float));
    is.read(reinterpret_cast<char*>(&panels[0]), tHeader.un32PanelCount * 4 * sizeof(uint32_t));

    // Read panel's color
    is.read(reinterpret_cast<char*>(&_un32PCR), sizeof(_un32PCR));
    is.read(reinterpret_cast<char*>(&_un32PCG), sizeof(_un32PCG));
    is.read(reinterpret_cast<char*>(&_un32PCB), sizeof(_un32PCB));
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }

  //  Generate panels
  InitFromMemory(&vertices[0], tHeader.un32VertexCount, &panels[0], tHeader.un32PanelCount,
                 bEnableSmoothShading, bUseAvgNormalForSmoothShading, ptVAmbientLight, ptVOmniLight);

  // Read texture coordinates
  try {
    uint32_t w, h;
    is.read(reinterpret_cast<char*>(&w), sizeof(w));
    is.read(reinterpret_cast<char*>(&h), sizeof(h));
    for(uint32_t un32Cnt1 = 0; un32Cnt1 < _un32PCount; un32Cnt1++) {
      for(uint32_t un32Cnt2 = 0; un32Cnt2 < 4; un32Cnt2++) {
        int32_t u, v;
        is.read(reinterpret_cast<char*>(&u), sizeof(u));
        is.read(reinterpret_cast<char*>(&v), sizeof(v));
        _pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fU = float(u) / float(w);
        _pcsPList[un32Cnt1]._tP.atUV[un32Cnt2].fV = float(v) / float(h);
      }
    }
  }
  catch(...) {
    CalcDefTexCoords();
    VWarning(_NFL_, "Couldn't read texture coordinates from stream \"%s\"\nUsing the default texture coordinates", strmName);
  }
}

} // namespace vge
