// ############################################################################################
//  UnfoggedProjectionFunctions.cc
//  Implementation of normal (unfogged) projection functions of CPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "FixedPointConsts.h"
#include "Panel3D.h"


namespace vge {

// Untextured/uncolored, flat shaded
void CPanel3D::UFProject()
{
  // For clipping points
  _TZClipXYZ tZClipXYZ;

  // Initialization
  _tP.un8SPCount    = 4;
  uint8_t un8Start  = 3;
  uint8_t un8OutCnt = 0;

  // Loop trough all edges of panel using Sutherland-Hodgman algorithm
  for(uint8_t un8End = 0; un8End < _tP.un8SPCount; un8End++) {
    if(_tP.apcsVPoint[un8Start]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entirely inside front view volume
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        un8OutCnt++;
      }
      else {
        // Leaving view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        un8OutCnt++;
      }
    }
    else {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entering view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        un8OutCnt++;

        // Add an extra edge to list
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        un8OutCnt++;
      }
      else {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }

  // Store the number of vertices from un8OutCnt
  _tP.un8SPCount = un8OutCnt;

  // Project panel's points
  for(uint8_t un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++) {
    const float fOneOverZ = 1.0f / tZClipXYZ.afZ[un8Cnt];

    _tP.atSPoint[un8Cnt].n32X =  int32_t(tZClipXYZ.afX[un8Cnt]
                                         * C3DEnv::Active3DEnv()->_fXScale
                                         * fOneOverZ
                                         + C3DEnv::Active3DEnv()->_fXShift);
    _tP.atSPoint[un8Cnt].n32Y = int32_t(tZClipXYZ.afY[un8Cnt]
                                        * C3DEnv::Active3DEnv()->_fYScale
                                        * fOneOverZ
                                        + C3DEnv::Active3DEnv()->_fYShift);
  
    _tP.atSPoint[un8Cnt].n32Z = int32_t(fOneOverZ * float(1 << _ZSTEP_PREC_));
  }
}

// Untextured/uncolored, gouraud shaded
void CPanel3D::UGProject()
{
  // For clipping points
  _TZClipXYZ tZClipXYZ;
  TUN8RGB    atClip[5];

  // Initialization
  _tP.un8SPCount    = 4;
  uint8_t un8Start  = 3;
  uint8_t un8OutCnt = 0;

  // Loop trough all edges of panel using Sutherland-Hodgman algorithm
  for(uint8_t un8End = 0; un8End < _tP.un8SPCount; un8End++) {
    if(_tP.apcsVPoint[un8Start]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entirely inside front view volume
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        xprv::_SHCopy(atClip[un8OutCnt], _tP.atTotal[un8End]);
        un8OutCnt++;
      }
      else {
        // Leaving view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        xprv::_SHClip(atClip[un8OutCnt], _tP.atTotal[un8Start], _tP.atTotal[un8End], fT);
        un8OutCnt++;
      }
    }
    else {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entering view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        xprv::_SHClip(atClip[un8OutCnt], _tP.atTotal[un8Start], _tP.atTotal[un8End], fT);
        un8OutCnt++;

        // Add an extra edge to list
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        xprv::_SHCopy(atClip[un8OutCnt], _tP.atTotal[un8End]);
        un8OutCnt++;
      }
      else {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }

  // Store the number of vertices from un8OutCnt
  _tP.un8SPCount = un8OutCnt;

  // Project panel's points
  for(uint8_t un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++) {
    const float fOneOverZ = 1.0f / tZClipXYZ.afZ[un8Cnt];

    _tP.atSPoint[un8Cnt].n32X =  int32_t(tZClipXYZ.afX[un8Cnt]
                                         * C3DEnv::Active3DEnv()->_fXScale
                                         * fOneOverZ
                                         + C3DEnv::Active3DEnv()->_fXShift);
    _tP.atSPoint[un8Cnt].n32Y = int32_t(tZClipXYZ.afY[un8Cnt]
                                        * C3DEnv::Active3DEnv()->_fYScale
                                        * fOneOverZ
                                        + C3DEnv::Active3DEnv()->_fYShift);
  
    _tP.atSPoint[un8Cnt].n32Z = int32_t(fOneOverZ * float(1 << _ZSTEP_PREC_));

    _tP.atSPoint[un8Cnt].n32R = int32_t(atClip[un8Cnt].un8R) << _ISTEP_PREC_;
    _tP.atSPoint[un8Cnt].n32G = int32_t(atClip[un8Cnt].un8G) << _ISTEP_PREC_;
    _tP.atSPoint[un8Cnt].n32B = int32_t(atClip[un8Cnt].un8B) << _ISTEP_PREC_;
  }
}

// Textured, not/flat shaded
void CPanel3D::TFProject(const CTexture &csTexture)
{
  // For clipping points
  _TZClipXYZ tZClipXYZ;
  TTexUV     atNewUV[5];

  // Initialization
  _tP.un8SPCount    = 4;
  uint8_t un8Start  = 3;
  uint8_t un8OutCnt = 0;

  // Loop trough all edges of panel using Sutherland-Hodgman algorithm
  for(uint8_t un8End = 0; un8End < _tP.un8SPCount; un8End++) {
    if(_tP.apcsVPoint[un8Start]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entirely inside front view volume
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        xprv::_SHCopy(atNewUV[un8OutCnt], _tP.atUV[un8End]);
        un8OutCnt++;
      }
      else {
        // Leaving view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        xprv::_SHClip(atNewUV[un8OutCnt], _tP.atUV[un8Start], _tP.atUV[un8End], fT);
        un8OutCnt++;
      }
    }
    else {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entering view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        xprv::_SHClip(atNewUV[un8OutCnt], _tP.atUV[un8Start], _tP.atUV[un8End], fT);
        un8OutCnt++;

        // Add an extra edge to list
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        xprv::_SHCopy(atNewUV[un8OutCnt], _tP.atUV[un8End]);
        un8OutCnt++;
      }
      else {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }

  // Store the number of vertices from un8OutCnt
  _tP.un8SPCount = un8OutCnt;

  // Project panel's points
  for(uint8_t un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++) {
    const float fOneOverZ = 1.0f / tZClipXYZ.afZ[un8Cnt];

    _tP.atSPoint[un8Cnt].n32X =  int32_t(tZClipXYZ.afX[un8Cnt]
                                         * C3DEnv::Active3DEnv()->_fXScale
                                         * fOneOverZ
                                         + C3DEnv::Active3DEnv()->_fXShift);
    _tP.atSPoint[un8Cnt].n32Y = int32_t(tZClipXYZ.afY[un8Cnt]
                                        * C3DEnv::Active3DEnv()->_fYScale
                                        * fOneOverZ
                                        + C3DEnv::Active3DEnv()->_fYShift);
  
    _tP.atSPoint[un8Cnt].n32Z = int32_t(fOneOverZ * float(1 << _ZSTEP_PREC_));

    _tP.atSPoint[un8Cnt].n32U = int32_t(atNewUV[un8Cnt].fU * csTexture.XMax()) * _tP.atSPoint[un8Cnt].n32Z;
    _tP.atSPoint[un8Cnt].n32V = int32_t(atNewUV[un8Cnt].fV * csTexture.YMax()) * _tP.atSPoint[un8Cnt].n32Z;
  }
}

// Textured, gouraud shaded
void CPanel3D::TGProject(const CTexture &csTexture)
{
  // For clipping points
  _TZClipXYZ tZClipXYZ;
  TUN8RGB    atClip[5];
  TTexUV     atNewUV[5];

  // Initialization
  _tP.un8SPCount    = 4;
  uint8_t un8Start  = 3;
  uint8_t un8OutCnt = 0;

  // Loop trough all edges of panel using Sutherland-Hodgman algorithm
  for(uint8_t un8End = 0; un8End < _tP.un8SPCount; un8End++) {
    if(_tP.apcsVPoint[un8Start]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entirely inside front view volume
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        xprv::_SHCopy(atClip[un8OutCnt], _tP.atTotal[un8End]);
        xprv::_SHCopy(atNewUV[un8OutCnt], _tP.atUV[un8End]);
        un8OutCnt++;
      }
      else {
        // Leaving view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        xprv::_SHClip(atClip[un8OutCnt], _tP.atTotal[un8Start], _tP.atTotal[un8End], fT);
        xprv::_SHClip(atNewUV[un8OutCnt], _tP.atUV[un8Start], _tP.atUV[un8End], fT);
        un8OutCnt++;
      }
    }
    else {
      if(_tP.apcsVPoint[un8End]->_tWorld.fZ >= C3DEnv::Active3DEnv()->_un32MinZ) {
        // Entering view volume
        const float fDeltaZ = _tP.apcsVPoint[un8End]->_tWorld.fZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ;
        const float fT      = float(C3DEnv::Active3DEnv()->_un32MinZ - _tP.apcsVPoint[un8Start]->_tWorld.fZ) / fDeltaZ;
        xprv::_SHClip(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8Start, un8End, fT, C3DEnv::Active3DEnv()->_un32MinZ);
        xprv::_SHClip(atClip[un8OutCnt], _tP.atTotal[un8Start], _tP.atTotal[un8End], fT);
        xprv::_SHClip(atNewUV[un8OutCnt], _tP.atUV[un8Start], _tP.atUV[un8End], fT);
        un8OutCnt++;

        // Add an extra edge to list
        xprv::_SHCopy(tZClipXYZ, un8OutCnt, _tP.apcsVPoint, un8End);
        xprv::_SHCopy(atClip[un8OutCnt], _tP.atTotal[un8End]);
        xprv::_SHCopy(atNewUV[un8OutCnt], _tP.atUV[un8End]);
        un8OutCnt++;
      }
      else {
        // Entirely outside clipping boundary
      }
    }
    // Advance to next vertex
    un8Start = un8End;
  }

  // Store the number of vertices from un8OutCnt
  _tP.un8SPCount = un8OutCnt;

  // Project panel's points
  for(uint8_t un8Cnt = 0; un8Cnt < un8OutCnt; un8Cnt++) {
    const float fOneOverZ = 1.0f / tZClipXYZ.afZ[un8Cnt];

    _tP.atSPoint[un8Cnt].n32X =  int32_t(tZClipXYZ.afX[un8Cnt]
                                         * C3DEnv::Active3DEnv()->_fXScale
                                         * fOneOverZ
                                         + C3DEnv::Active3DEnv()->_fXShift);
    _tP.atSPoint[un8Cnt].n32Y = int32_t(tZClipXYZ.afY[un8Cnt]
                                        * C3DEnv::Active3DEnv()->_fYScale
                                        * fOneOverZ
                                        + C3DEnv::Active3DEnv()->_fYShift);
  
    _tP.atSPoint[un8Cnt].n32Z = int32_t(fOneOverZ * float(1 << _ZSTEP_PREC_));

    _tP.atSPoint[un8Cnt].n32U = int32_t(atNewUV[un8Cnt].fU * csTexture.XMax()) * _tP.atSPoint[un8Cnt].n32Z;
    _tP.atSPoint[un8Cnt].n32V = int32_t(atNewUV[un8Cnt].fV * csTexture.YMax()) * _tP.atSPoint[un8Cnt].n32Z;

    _tP.atSPoint[un8Cnt].n32R = int32_t(atClip[un8Cnt].un8R) << _ISTEP_PREC_;
    _tP.atSPoint[un8Cnt].n32G = int32_t(atClip[un8Cnt].un8G) << _ISTEP_PREC_;
    _tP.atSPoint[un8Cnt].n32B = int32_t(atClip[un8Cnt].un8B) << _ISTEP_PREC_;
    ///**-+*/_tP.atSPoint[un8Cnt].n32R = atClip[un8Cnt].un8R * _tP.atSPoint[un8Cnt].n32Z;
    ///**-+*/_tP.atSPoint[un8Cnt].n32G = atClip[un8Cnt].un8G * _tP.atSPoint[un8Cnt].n32Z;
    ///**-+*/_tP.atSPoint[un8Cnt].n32B = atClip[un8Cnt].un8B * _tP.atSPoint[un8Cnt].n32Z;
  }
}

} // namespace vge
