/** \file Random.h */
// ############################################################################################
//  Random.h
//  Definition & implementation of random number generator classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  The "Additive Congruential Method" used for generating the random numbers was
//  taken from the book :
//    Algorithms, Second Edition
//    Robert Sedgewick, Princeton University
//    Addison-Wesley Publishing Company
//    USA & Canada, 1998
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___RANDOM_H_
#define _BASE___RANDOM_H_

// VortexGE headers
#include "../Base/VBasics.h"


namespace vge {

  //!
  //! \brief Core random number generator class that will produce
  //!        unsigned 32-bit random integers
  //!
  //! Features :
  //!   -# Fully controllable, the random sequences are fully reproducible by
  //!      selecting the seed value.
  //!   -# Save-able state, the current state of the generator can be retrieved
  //!      and then stored in non-volatile storage, later if needed, the state
  //!      can be fully restored, example : when saving and loading game.
  //!   -# Multiple instance, more than one instance of this class can exist at
  //!      the same time and all of them will have their own states.
  //!
  //! \api beta
  class VGE_EXPORT CRandomCoreUInt32 {
    public:
      //!
      //! Construct a CRandomCoreUInt32 class instance using the given seed.
      inline CRandomCoreUInt32(uint32_t seed = 0)
      : _shiftReg(SR_Size)
      { Reset(seed); }
  
      // Initialization functions
      //!
      //! Reinitialize the random number generator class using the given seed.
      void Reset(uint32_t seed = 0);
      //!
      //! Restore the current state of the random number generator using the
      //! given parameters.
      //! \n\n
      //! This function can be used to restore the state of the random number
      //! generator class (for example when loading game).
      void SetState(uint32_t newSeed, uint32_t newSRIndex, const vector<uint32_t> &newShiftReg);
      //!
      //! Store the current state of the random number generator to the given
      //! parameters.
      //! \n\n
      //! This function can be used to save the state of the random number
      //! generator class (for example when saving game).
      void GetState(uint32_t &dstSeed, uint32_t &dstSRIndex, vector<uint32_t> &dstShiftReg);
  
      // Random number generation functions
      //!
      //! Generate a random number in the range of MinFiniteNumber() to MaxFiniteNumber().
      inline uint32_t Generate()
      { return(_Generate_()); }
      //!
      //! Generate a random number in the range of MinFiniteNumber() to maxValue.
      inline uint32_t Generate(uint32_t maxValue)
      { return(_Generate_() % (maxValue + 1)); }
      //!
      //! Generate a random number in the range of minValue to maxValue.
      //! \n\n
      //! Note that if maxValue < minValue the result is not guaranteed.
      inline uint32_t Generate(uint32_t minValue, uint32_t maxValue)
      { return(Generate(maxValue - minValue) + minValue); }
  
      // Information functions
      //!
      //! Returns the minimum finite value that can be generated
      //! (will always returns zero).
      inline static uint32_t MinFiniteNumber()
      { return(numeric_limits<uint32_t>::min()); }
      //!
      //! Returns the maximum finite value that can be generated
      //! (will always returns positive number).
      inline static uint32_t MaxFiniteNumber()
      { return(numeric_limits<uint32_t>::max()); }
  
    protected:
      // Helper functions
      uint32_t _Generate_();
      static const uint32_t SR_MultFac = 94967821; // Multiplication factor
      static const uint32_t SR_Size    = 55;       // Size of the shift register
      static const uint32_t SR_TapPos  = 31;       // Tap position in the shift register

      // Data
      uint32_t         _rnSeed;   // Current seed
      uint32_t         _srIndex;  // Index to the current cell in the shift register
      vector<uint32_t> _shiftReg; // The shift register
  };


  //!
  //! \brief Random number generator class that will produce fractional
  //!        (floating points) random numbers
  //!
  //! \api beta
  template<typename NumberT_>
  class VGE_EXPORT CRandomFractional {
    public:
      typedef NumberT_ NumberT;
  
    public:
      //!
      //! Construct a CRandomFractional class instance using the given
      //! CRandomCoreUInt32 instance.
      inline CRandomFractional(CRandomCoreUInt32 &core)
      : _core(core),
        _div (1.0 / CRandomCoreUInt32::MaxFiniteNumber()),
        _div2(2.0 / CRandomCoreUInt32::MaxFiniteNumber())
      {}
  
      // Random number generation functions
      //!
      //! Generate a random number in the range of 0.0 to 1.0.
      inline NumberT GeneratePositive()
      { return(_core.Generate() * _div); }
      //!
      //! Generate a random number in the range of 0.0 to maxValue.
      //!
      //! Note that if maxValue < 0.0 the result is not guaranteed.
      inline NumberT GeneratePositive(NumberT maxValue)
      { return(_core.Generate() * _div * maxValue); }
      //!
      //! Generate a random number in the range of -1.0 to 0.0.
      inline NumberT GenerateNegative()
      { return(_core.Generate() * _div - 1.0); }
      //!
      //! Generate a random number in the range of minValue to 0.0.
      //!
      //! Note that if minValue > 0.0 the result is not guaranteed.
      inline NumberT GenerateNegative(NumberT minValue)
      { return(_core.Generate() * _div * minValue); }
      //!
      //! Generate a random number in the range of -1.0 to 1.0.
      inline NumberT GenerateBipolar()
      { return(_core.Generate() * _div2 - 1.0); }
      //!
      //! Generate a random number in the range of minValue to maxValue.
      //!
      //! Note that if maxValue < minValue the result is not guaranteed.
      inline NumberT GenerateBipolar(NumberT minValue, NumberT maxValue)
      { return(GeneratePositive() * (maxValue - minValue) + minValue); }
  
      // Information functions
      //!
      //! Returns the minimum finite value that can be generated
      //! (will always returns the smallest positive number).
      inline static NumberT MinFiniteNumber()
      { return(numeric_limits<NumberT>::min()); }
      //!
      //! Returns the maximum finite value that can be generated.
      //! (will always returns the largest positive number).
      inline static NumberT MaxFiniteNumber()
      { return(numeric_limits<NumberT>::max()); }
  
    protected:
      CRandomCoreUInt32 &_core;
      NumberT            _div;
      NumberT            _div2;
  };
  
  // For convenience
  typedef CRandomFractional<float>       CRandomFloat;
  typedef CRandomFractional<double>      CRandomDouble;
  typedef CRandomFractional<long double> CRandomLongDouble;

} // namespace vge

#endif
