/** \file MediaPool.h */
// ############################################################################################
//  MediaPool.h
//  Definition of CBasicMediaPool template class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _GAME_OBJECT___MEDIA_POOL_H_
#define _GAME_OBJECT___MEDIA_POOL_H_

// VortexGE headers
#include "../2DManipulation/ImageBuilder.h"
#include "../2DManipulation/ImageClasses.h"
#include "../AudioEnvironment/AudioEnvironment.h"
#include "../AudioEnvironment/SoundBuffer.h"


namespace vge {

  //!
  //! \internal
  //! \brief A class used for storing pointer to a media resource, its
  //!        last get time and its reference count
  //!
  //! \api beta
  template<typename MediaT_>
  class _CMediaInstE {
    public:
      typedef MediaT_ MediaT;

    public:
      inline _CMediaInstE(MediaT *mediaInst = NULL)
      : _mediaInst(mediaInst), _lastGetTime(0), _refCount(0)
      {}

      // Information functions
      inline bool    IsReferenced()   const { return(_refCount > 0); }
      inline ssize_t ReferenceCount() const { return(_refCount);     }
      inline time_t  LastGetTime()    const { return(_lastGetTime);  }

      // Direct media access functions
      inline       MediaT *MediaInst()       { return(_mediaInst); }
      inline const MediaT *MediaInst() const { return(_mediaInst); }

      // Reference counted media access functions
      inline MediaT *Get();
      inline void Release();

    private:
      MediaT *_mediaInst;   // The media instance
      time_t  _lastGetTime; // Last "media get" time
      ssize_t _refCount;    // Reference count of the media instance
  };


  //!
  //! \brief A class used for managing media resources.
  //!
  //! Because all game entity classes will always use string for the resource ID,
  //! a convenience typedef is provided:
  //! \code
  //!   typedef CBasicMediaPool<string> CMediaPool;
  //! \endcode
  //!
  //! \api experiment
  template<typename TResourceID>
  class VGE_EXPORT CBasicMediaPool : VCLASS_UNCOPYABLE {
    public:
      //!
      //! \brief Used to tell NewSoundBuffer() about the source audio data format.
      enum ESourceAudioDataFormat { ESADF_AUTO, ESADF_RAW, ESADF_VSW, ESADF_WAV };

    public:
      //!
      //! Standard ctor
      CBasicMediaPool(CAudioEnv *audioEnvironment, bool throwExceptionIfResourceNotFound = true);
      //!
      //! Standard dtor
      ~CBasicMediaPool() throw();

      // Initialization functions
      //!
      //! Sets the audio environment to be used with this class.
      void SetAudioEnvironment(CAudioEnv *audioEnvironment);
      //!
      //! Tell this CBasicMediaPool instance whether it should throw exception
      //! or just returns NULL if the requested resource is not found
      void ThrowExceptionIfResourceNotFound(bool throws = true);

      // Any media related functions
      //!
      //! Delete all unused (not referenced media resources) of which last
      //! get time is older than the specified time limit (in seconds).
      void Compact(time_t timeLimitInSeconds = 15 * 60);

      // CImage related functions
      //!
      //! Create a new CImage resource with the given ID using the given
      //! CImageBuilder instance.
      void NewImage(const TResourceID &id, const CImageBuilder &src);
      //!
      //! Create a new CImage resource with the given ID using the given
      //! CImageBuilder instance and get (return) the newly created resources
      inline CImage *GetNewImage(const TResourceID &id, const CImageBuilder &src)
      { NewImage(id, src); return(GetImage(id)); }
      //!
      //! Check if the given CImage resource ID is exist.
      bool IsRegisteredImage(const TResourceID &id) const;
      //!
      //! Get a CImage instance based on the ID.
      CImage *GetImage(const TResourceID &id);
      //!
      //! Get a CImage instance based on the given instance (increment
      //! the reference count of the media resource).
      CImage *GetImage(const CImage *inst);
      //!
      //! Release a CImage instance based on the ID.
      void ReleaseImage(const TResourceID &id);
      //!
      //! Release the given CImage instance.
      void ReleaseImage(const CImage *inst);
      //!
      //! Delete a CImage instance based on the ID.
      void DeleteImage(const TResourceID &id, bool forceEvenStillUsed = false);
      //!
      //! Delete all CImage resources.
      void DeleteAllImages(bool forceEvenStillUsed = false);

      // CAnimatedImage related functions
      //!
      //! Create a new CAnimatedImage resource with the given ID using the given
      //! CImageBuilder instance.
      void NewAnimatedImage(const TResourceID &id, const CImageBuilder &src, uint8_t rowCount, uint8_t colCount);
      //!
      //! Create a new CAnimatedImage resource with the given ID using the given
      //! CImageBuilder instance and get (return) the newly created resources
      inline CAnimatedImage *GetNewAnimatedImage(const TResourceID &id, const CImageBuilder &src, uint8_t rowCount, uint8_t colCount)
      { NewAnimatedImage(id, src); return(GetAnimatedImage(id, rowCount, colCount)); }
      //!
      //! Check if the given CAnimatedImage resource ID is exist.
      bool IsRegisteredAnimatedImage(const TResourceID &id) const;
      //!
      //! Get a CAnimatedImage instance based on the ID.
      CAnimatedImage *GetAnimatedImage(const TResourceID &id);
      //!
      //! Get a CAnimatedImage instance based on the given instance (increment
      //! the reference count of the media resource).
      CAnimatedImage *GetAnimatedImage(const CAnimatedImage *inst);
      //!
      //! Release a CAnimatedImage instance based on the ID.
      void ReleaseAnimatedImage(const TResourceID &id);
      //!
      //! Release the given CAnimatedImage instance.
      void ReleaseAnimatedImage(const CAnimatedImage *inst);
      //!
      //! Delete a CAnimatedImage instance based on the ID.
      void DeleteAnimatedImage(const TResourceID &id, bool forceEvenStillUsed = false);
      //!
      //! Delete all CImage resources.
      void DeleteAllAnimatedImages(bool forceEvenStillUsed = false);

      // CTexture related functions
      //!
      //! Create a new CTexture resource with the given ID using the given
      //! CImageBuilder instance.
      void NewTexture(const TResourceID &id, const CImageBuilder &src);
      //!
      //! Create a new CTexture resource with the given ID using the given
      //! CImageBuilder instance get (return) the newly created resources
      inline CTexture *GetNewTexture(const TResourceID &id, const CImageBuilder &src)
      { NewTexture(id, src); return(GetTexture(id)); }
      //!
      //! Check if the given CTexture resource ID is exist.
      bool IsRegisteredTexture(const TResourceID &id) const;
      //!
      //! Get a CTexture instance based on the ID.
      CTexture *GetTexture(const TResourceID &id);
      //!
      //! Get a CTexture instance based on the given instance (increment
      //! the reference count of the media resource).
      CTexture *GetTexture(const CTexture *inst);
      //!
      //! Release a CTexture instance based on the ID.
      void ReleaseTexture(const TResourceID &id);
      //!
      //! Release the given CTexture instance.
      void ReleaseTexture(const CTexture *inst);
      //!
      //! Delete a CTexture instance based on the ID.
      void DeleteTexture(const TResourceID &id, bool forceEvenStillUsed = false);
      //!
      //! Delete all CTexture resources.
      void DeleteAllTextures(bool forceEvenStillUsed = false);

      // CAnimatedTexture related functions
      //!
      //! Create a new CAnimatedTexture resource with the given ID using the given
      //! CImageBuilder instance.
      void NewAnimatedTexture(const TResourceID &id, const CImageBuilder &src, uint8_t rowCount, uint8_t colCount);
      //!
      //! Create a new CAnimatedTexture resource with the given ID using the given
      //! CImageBuilder instance and get (return) the newly created resources
      inline CAnimatedTexture *GetNewAnimatedTexture(const TResourceID &id, const CImageBuilder &src, uint8_t rowCount, uint8_t colCount)
      { NewAnimatedTexture(id, src); return(GetAnimatedTexture(id, rowCount, colCount)); }
      //!
      //! Check if the given CAnimatedTexture resource ID is exist.
      bool IsRegisteredAnimatedTexture(const TResourceID &id) const;
      //!
      //! Get a CAnimatedTexture instance based on the ID.
      CAnimatedTexture *GetAnimatedTexture(const TResourceID &id);
      //!
      //! Get a CAnimatedTexture instance based on the given instance (increment
      //! the reference count of the media resource).
      CAnimatedTexture *GetAnimatedTexture(const CAnimatedTexture *inst);
      //!
      //! Release a CAnimatedTexture instance based on the ID.
      void ReleaseAnimatedTexture(const TResourceID &id);
      //!
      //! Release the given CAnimatedTexture instance.
      void ReleaseAnimatedTexture(const CAnimatedTexture *inst);
      //!
      //! Delete a CAnimatedTexture instance based on the ID.
      void DeleteAnimatedTexture(const TResourceID &id, bool forceEvenStillUsed = false);
      //!
      //! Delete all CTexture resources.
      void DeleteAllAnimatedTextures(bool forceEvenStillUsed = false);

      // CSoundBuffer related functions
      //!
      //! Create a new CSoundBuffer resource with the given ID using the given
      //! CDataSource instance.
      //! \n\n
      //! The "rawDataFormat" and "rawDataFrequency" parameters will only be used
      //! if the audio data to be loaded is in raw format.
      void NewSoundBuffer(const TResourceID      &id,
                          const CDataSource      &srcAudioData,
                          ESourceAudioDataFormat  sourceAudioDataFormat = ESADF_AUTO,
                          ERawFormat              rawDataFormat         = ERF_Invalid,
                          int32_t                 rawDataFrequency      = 0);
      //!
      //! Create a new CSoundBuffer resource with the given ID using the given
      //! and get (return) the newly created CSoundBuffer resource.
      //! \n\n
      //! See NewSoundBuffer() for parameters details.
      inline CSoundBuffer *GetNewSoundBuffer(const TResourceID      &id,
                                             const CDataSource      &srcAudioData,
                                             ESourceAudioDataFormat  sourceAudioDataFormat = ESADF_AUTO,
                                             ERawFormat              rawDataFormat         = ERF_Invalid,
                                             int32_t                 rawDataFrequency      = 0)
      { NewSoundBuffer(id, srcAudioData, sourceAudioDataFormat, rawDataFormat, rawDataFrequency); return(GetSoundBuffer(id)); }
      //!
      //! Check if the given CSoundBuffer resource ID is exist.
      bool IsRegisteredSoundBuffer(const TResourceID &id) const;
      //!
      //! Get a CSoundBuffer instance based on the ID.
      CSoundBuffer *GetSoundBuffer(const TResourceID &id);
      //!
      //! Get a CSoundBuffer instance based on the given instance (increment
      //! the reference count of the media resource).
      CSoundBuffer *GetSoundBuffer(const CSoundBuffer *inst);
      //!
      //! Release a CSoundBuffer instance based on the ID.
      void ReleaseSoundBuffer(const TResourceID &id);
      //!
      //! Release the given CSoundBuffer instance.
      void ReleaseSoundBuffer(const CSoundBuffer *inst);
      //!
      //! Delete a CSoundBuffer instance based on the ID.
      void DeleteSoundBuffer(const TResourceID &id, bool forceEvenStillUsed = false);
      //!
      //! Delete all CSoundBuffer resources.
      void DeleteAllSoundBuffers(bool forceEvenStillUsed = false);

  protected:
      // Helper functions
      template<typename MapT> static
        void _Compact(MapT &map, time_t timeLimitInSeconds);
      template<typename MapT> inline static
        void _ThrowIfMediaAlreadyExist(const MapT &map, const TResourceID &id, const char* mname);
      template<typename EntryT, typename MapT> inline
        void _AddMediaToList(MapT &map, const TResourceID &id, typename MapT::mapped_type::MediaT *media);
      template<typename MapT> inline static
        bool _IsRegisteredMedia(const MapT &map, const TResourceID &id);
      template<typename MapT> static
        typename MapT::mapped_type::MediaT *_GetMedia(MapT &map, const TResourceID &id, const char* mname, bool throwIfNotExist);
      template<typename MapT> static
        typename MapT::mapped_type::MediaT *_GetMediaInst(MapT &map, const typename MapT::mapped_type::MediaT *inst, const char* mname, bool throwIfNotExist);
      template<typename MapT> static
        void _ReleaseMedia(MapT &map, const TResourceID &id, const char* mname, bool throwIfNotExist);
      template<typename MapT> static
        void _ReleaseMediaInst(MapT &map, const typename MapT::mapped_type::MediaT *inst, const char* mname, bool throwIfNotExist);
      template<typename MapT> static
        void _DeleteMedia(MapT &map, const TResourceID &id, const char* mname, bool forceEvenStillUsed);
      template<typename MapT> static
        void _DeleteAllMedias(MapT &map, const char* mname, bool forceEvenStillUsed);

      // Media resources' entry classes
      typedef _CMediaInstE<CImage>           _CImageE;
      typedef _CMediaInstE<CAnimatedImage>   _CAnimatedImageE;
      typedef _CMediaInstE<CTexture>         _CTextureE;
      typedef _CMediaInstE<CAnimatedTexture> _CAnimatedTextureE;
      typedef _CMediaInstE<CSoundBuffer>     _CSoundBufferE;

      // Media resources' map classes
      typedef map<TResourceID, _CImageE>           _CImageMap;
      typedef map<TResourceID, _CAnimatedImageE>   _CAnimatedImageMap;
      typedef map<TResourceID, _CTextureE>         _CTextureMap;
      typedef map<TResourceID, _CAnimatedTextureE> _CAnimatedTextureMap;
      typedef map<TResourceID, _CSoundBufferE>     _CSoundBufferMap;

      // Data
      CAudioEnv *_audEnv;             // CAudioEnv instance to be used to create CSoundBuffer instances
      bool       _throwIfResNotFound; // Should this class throw if the requested resource is not found ?

      _CImageMap           _imageMap;       // Map of CImage instanecs
      _CAnimatedImageMap   _aimageMap;      // Map of CAnimatedImage instanecs
      _CTextureMap         _textureMap;     // Map of CTexture instanecs
      _CAnimatedTextureMap _atextureMap;    // Map of CAnimatedTexture instanecs
      _CSoundBufferMap     _soundbufferMap; // Map of CSoundBuffer instanecs
  };

  // For convenience
  typedef CBasicMediaPool<string> CMediaPool;


  // Helper function prototypes for CBasicMedia_MP<TResourceID, MediaT>
  // template class
  //!
  //! \internal Used by CBasicMedia_MP<TResourceID, MediaT> template class.
  template<typename TResourceID, typename MediaT> inline
  void _GetMedia_MP(MediaT *&inst, CBasicMediaPool<TResourceID> &mediaPool, const TResourceID &resourceID);
  //!
  //! \internal Used by CBasicMedia_MP<TResourceID, MediaT> template class.
  template<typename TResourceID, typename MediaT> inline
  void _GetMedia_MP(MediaT *&inst, CBasicMediaPool<TResourceID> *mediaPool, const MediaT *mediaInst);
  //!
  //! \internal Used by CBasicMedia_MP<TResourceID, MediaT> template class.
  template<typename TResourceID, typename MediaT> inline
  void _ReleaseMedia_MP(const MediaT *inst, CBasicMediaPool<TResourceID> *mediaPool);

  //!
  //! \brief Convenience class to pass pooled-media as function parameter.
  //!
  //! \api preliminary experiment
  template<typename TResourceID_, typename MediaT_>
  class CBasicMedia_MP {
    public:
      typedef TResourceID_ TResourceID;
      typedef MediaT_      MediaT;

      typedef CBasicMediaPool<TResourceID> MediaPoolT;

    public:
      //!
      //! Construct an empty pooled-media.
      inline CBasicMedia_MP()
      : _mediaPool(NULL), _mediaInst(NULL)
      {}
      //!
      //! Copy ctor.
      inline CBasicMedia_MP(const CBasicMedia_MP &inst)
      : _mediaPool(NULL), _mediaInst(NULL)
      { *this = inst; }
      //!
      //! Construct a pooled-media using the given CMediaPool instance
      //! and resource ID.
      inline CBasicMedia_MP(MediaPoolT &mediaPool, const TResourceID &resourceID)
      : _mediaPool(NULL), _mediaInst(NULL)
      {
        _GetMedia_MP(_mediaInst, mediaPool, resourceID);
        _mediaPool = &mediaPool;
      }
      //!
      //! Standard dtor.
      inline ~CBasicMedia_MP() throw()
      { _releaseMedia(); }

      // Overloaded operators
      //!
      //! Assignment operator.
      inline CBasicMedia_MP &operator=(const CBasicMedia_MP &inst)
      {
        if(this == &inst) return(*this);
        _releaseMedia();
        if(inst._mediaPool) _GetMedia_MP(_mediaInst, inst._mediaPool, inst._mediaInst);
        _mediaPool = inst._mediaPool;
        return(*this);
      }

      // Access to the media instance.
      //!
      //! Access to the media instance.
      inline MediaT *MediaInst() { return(_mediaInst); }
      //!
      //! Access to the media instance.
      inline const MediaT *MediaInst() const { return(_mediaInst); }

    private:
      // Helper functions
      inline void _releaseMedia()
      {
        if(_mediaPool && _mediaInst) _ReleaseMedia_MP(_mediaInst, _mediaPool);
        _mediaPool = NULL;
        _mediaInst = NULL;
      }

      // Data
      MediaPoolT *_mediaPool;
      MediaT     *_mediaInst;
  };

  // For convenience
  typedef CBasicMedia_MP<string, CImage>           CImage_MP;
  typedef CBasicMedia_MP<string, CAnimatedImage>   CAnimatedImage_MP;
  typedef CBasicMedia_MP<string, CTexture>         CTexture_MP;
  typedef CBasicMedia_MP<string, CAnimatedTexture> CAnimatedTexture_MP;
  typedef CBasicMedia_MP<string, CSoundBuffer>     CSoundBuffer_MP;

  // Implement the helper functions for the CBasicMedia_MP<TResourceID, MediaT>
  // template class prototyped above
#ifndef _DOXYGEN_SHOULD_IGNORE_THIS_
#define _IMPL_FUNC___GetReleaseMedia_MP_(MediaN)                                                                \
  template<typename TResourceID> inline                                                                         \
    void _GetMedia_MP(C##MediaN *&inst, CBasicMediaPool<TResourceID> &mediaPool, const TResourceID &resourceID) \
    { inst = mediaPool.Get##MediaN(resourceID); }                                                               \
  template<typename TResourceID> inline                                                                         \
    void _GetMedia_MP(C##MediaN *&inst, CBasicMediaPool<TResourceID> *mediaPool, const C##MediaN *mediaInst)    \
    { inst = mediaPool->Get##MediaN(mediaInst); }                                                               \
  template<typename TResourceID> inline                                                                         \
    void _ReleaseMedia_MP(const C##MediaN *inst, CBasicMediaPool<TResourceID> *mediaPool)                       \
    { mediaPool->Release##MediaN(inst); }
  _IMPL_FUNC___GetReleaseMedia_MP_(Image)
  _IMPL_FUNC___GetReleaseMedia_MP_(AnimatedImage)
  _IMPL_FUNC___GetReleaseMedia_MP_(Texture)
  _IMPL_FUNC___GetReleaseMedia_MP_(AnimatedTexture)
  _IMPL_FUNC___GetReleaseMedia_MP_(SoundBuffer)
#undef _IMPL_FUNC___GetReleaseMedia_MP_
#endif

} // namespace vge
#endif
