/** \file DataSinkSource.h */
// ############################################################################################
//  DataSinkSource.h
//  Definition & implementation of CDataSink and CDataSource classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _SYSTEM___DATA_SINK_SOURCE_H_
#define _SYSTEM___DATA_SINK_SOURCE_H_

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../Base/SpecialClasses.h"


namespace vge {

  //!
  //! \brief Wrapper class to make an istream become a peekable istream
  //!
  //! \api beta
  //!
  //! This class is made so that one can peek more than one bytes in all
  //! std::istream (and it's derivatives) class instances because normal
  //! std::istream can only peek one byte and not all istream implementation
  //! support seeking.
  //! \n\n
  //! The main purpose of this class is so that all existing code which use
  //! std::istream (and its derivatives) classes can be easily converted to
  //! support data peeking (for example peeking some bytes to determine the
  //! stored image/audio format and to call the appropriate function to load
  //! the image/audio data).
  //! \n\n
  //! Note that user shouldn't need to use this class directly since this
  //! class is mean to be used by VortexGE I/O functions. If user wants to use
  //! this class, please use it with caution.
  //! \n\n
  //! Calling other member functions of std::istream than the delegated member
  //! functions in this class may cause undefined result.
  class VGE_EXPORT pistream : VCLASS_UNCOPYABLE {
    public:
      // Redefine the data types for convenience
      typedef istream::char_type char_type;
      typedef istream::pos_type  pos_type;
      typedef istream::off_type  off_type;
      typedef istream::iostate   iostate;

    public:
      //!
      //! Construct a new pistream instance using the given istream instance
      inline pistream(istream &is)
      : _is(is), _pbuff(_ibuff), _gcount(0)
      {}
      //!
      //! Construct a new pistream instance using the given istream and peek
      //! buffer instances
      inline pistream(istream &is, vector<char> &pb)
      : _is(is), _pbuff(pb), _gcount(0)
      {}

      // Information functions
      //!
      //! Check if the previous operation has failed
      inline bool fail()
      { return(_is.fail()); }
      //!
      //! Returns the number of bytes extracted by previous peek/read.
      inline streamsize gcount() const
      { return(_gcount); }
      //!
      //! Returns the current peek/read position.
      inline pos_type tellg() const
      { return(_is.tellg() - static_cast<pos_type>(_pbuff.size())); }

      // Exception functions
      //!
      //! Sets the current exceptions mask.
      inline iostate exceptions() const
      { return(_is.exceptions()); }
      //!
      //! Set a new exceptions mask.
      inline void exceptions(iostate e) const
      { _is.exceptions(e); }

      // Data reading functions
      //!
      //! Peek some bytes from the istream.
      pistream &peek(char_type *s, streamsize n);
      //!
      //! Read some bytes from the istream.
      pistream &read(char_type *s, streamsize n);
      //!
      //! Read one byte from the istream.
      pistream &get(char_type &c);
      //!
      //! Discard some bytes
      pistream &discard(streamsize n);
      //!
      //! Discard one byte
      pistream &discard();

      // Data seeking functions
      //!
      //! Seek to a new peek/read position.
      pistream &seekg(off_type off, ios_base::seekdir dir);
      //!
      //! Seek to a new peek/read position.
      pistream &seekg(pos_type pos);

      // Overloaded operators
      inline      operator void * () const { return(_is.operator void * ()); }
      inline bool operator      ! () const { return(_is.operator      ! ()); }


    protected:
      istream      &_is;     // The istream instance
      vector<char> &_pbuff;  // Peek buffer for the istream instance above
      vector<char>  _ibuff;  // Internal peek buffer (used if the user does not specify a custom peek buffer)
      streamsize    _gcount; // Numebr of last bytes read
  };


  //!
  //! \brief A class used to wrap an instance of istream and its name
  //!
  //! \api beta
  //!
  //! <b>Usage examples:</b>
  //! \code
  //! // Make sure that CDataSource is declared as constant reference
  //! void some_function(const CDataSource &id)
  //! {
  //!   const string &nm = id.Name();
  //!   istream      &is = id.Stream();
  //!
  //!   //
  //!   // Use "is" and "nm" as needed
  //!   //
  //!
  //! }
  //!
  //! // Code which calls the above function
  //! int main()
  //! {
  //!   // Using file name
  //!   some_function("/proc/cpuinfo");
  //!
  //!   // Using ifstream without specifying stream name
  //!   ifstream ifs1;
  //!   ifs1.open("/proc/cpuinfo");
  //!   some_function(ifs1);
  //!   ifs1.close();
  //!
  //!   // Using ifstream with specifying stream name
  //!   ifstream ifs2;
  //!   ifs2.open("/proc/cpuinfo");
  //!   some_function(CDataSource(ifs2, "/proc/cpuinfo"));
  //!   ifs2.close();
  //!
  //!   // Using ifstream with specifying stream name and stream size
  //!   ifstream ifs3;
  //!   ifs3.open("/proc/cpuinfo");
  //!   some_function(CDataSource(ifs3, "/proc/cpuinfo", 128));
  //!   ifs3.close();
  //!
  //!   return(0);
  //! }
  //! \endcode
  //!
  //! Some notes when using this CDataSource class :
  //!   -# Mixing the use of the istream instance returned by Stream() and the
  //!      pistream instance returned by PStream() may cause undefined behavior.
  //!   -# A custom peek buffer can be used when planning to use the pistream
  //!      instance returned by PStream() for a long time. The life span of the
  //!      custom peek buffer can be made as long as the life span of the istream
  //!      instance, this to make sure that peeked data can survive function call
  //!      boundaries. So, rather than writing code like:
  //!        \code
  //!        ifstream ifs;
  //!        ifs.open("/proc/cpuinfo");
  //!        my_function_call(CDataSource(ifs, "/proc/cpuinfo"));
  //!        \endcode
  //!      write it like:
  //!        \code
  //!        vector<char> pb;
  //!        ifstream     ifs;
  //!        ifs.open("/proc/cpuinfo");
  //!        my_function_call(CDataSource(ifs, pb, "/proc/cpuinfo"));
  //!        \endcode
  class VGE_EXPORT CDataSource : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Open the given file name in read only mode and use it as the input source.
      CDataSource(const char *fileName);
      //!
      //! Open the given file name in read only mode and use it as the input source.
      CDataSource(const string &fileName);
      //!
      //! Use the given istream instance as the input source (and optionally names the istream).
      //! \param[in] is                An instance to an ostream derived class to be used as the input source.
      //! \param[in] streamName        Optional name for the stream.
      //! \param[in] streamSize        Optional size for the stream (some I/O functions needs to know how many bytes the stream can give).
      //! \param[in] startOffsetOfData Optional start offset (in bytes) of the data in the stream.
      CDataSource(istream &is, const char *streamName = "[unnamed istream]", istream::pos_type streamSize = 0, istream::off_type startOffsetOfData = 0);
      //!
      //! Use the given istream instance as the input source (and optionally names the istream),
      //! use the given vector<char> instance as the peek buffer for pistream.
      //! \param[in] is                An instance to an ostream derived class to be used as the input source.
      //! \param[in] peekBuffer        Custom peek buffer to be used.
      //! \param[in] streamName        Optional name for the stream.
      //! \param[in] streamSize        Optional size for the stream (some I/O functions needs to know how many bytes the stream can give).
      //! \param[in] startOffsetOfData Optional start offset (in bytes) of the data in the stream.
      CDataSource(istream &is, vector<char> &peekBuffer, const char *streamName = "[unnamed istream]", istream::pos_type streamSize = 0, istream::off_type startOffsetOfData = 0);
      //!
      //! Standard dtor
      ~CDataSource() throw();

      // Getters
      //!
      //! Returns the istream instance.
      //! \n\n
      //! Mixing the use of the istream instance returned by Stream() and the
      //! pistream instance returned by PStream() may cause undefined behavior.
      inline istream &Stream() const
      { return(*_inpStream); }
      //!
      //! Returns the pistream instance.
      //! \n\n
      //! Mixing the use of the istream instance returned by Stream() and the
      //! pistream instance returned by PStream() may cause undefined behavior.
      inline pistream &PStream() const
      { return(*_piStream); }
      //!
      //! Returns the file name or the stream name.
      inline const string &Name() const
      { return(_strName); }
      //!
      //! Returns the size of the file or the stream.
      inline istream::pos_type Size() const
      { return(_streamSize); }
      //!
      //! Returns the start offset of the real data in the stream (in bytes),
      //! can be used when seeking from something like CArchiveDataSource.
      inline istream::off_type GetStartOffsetOfData() const
      { return(_dataStartOfs); }

    protected:
      // Make sure that this class is uncopyable unless the user has derived it
      CDataSource(const CDataSource&);
      CDataSource &operator=(const CDataSource&);

      // Helper functions
      void _InitFromFile_();
      void _InitFromIStream_();

      // Data
      bool              _ownStream;     // Flag to indice if this class own the istream instance or not
      string            _strName;       // Name of the file or the stream
      istream          *_inpStream;     // The istream instance
      pistream         *_piStream;      // The pistream instance
      ios::iostate      _orgExceptions; // The original exception settings of the istream instance
      istream::pos_type _streamSize;    // Size of the file or the stream
      istream::off_type _dataStartOfs;  // Start offset of the data in the stream
  };


  //!
  //! \brief A class used to wrap an instance of ostream and its name
  //!
  //! \api beta
  //!
  //! <b>Usage examples:</b>
  //! \code
  //! // Make sure that CDataSink is declared as constant reference
  //! void some_function(const CDataSink &id)
  //! {
  //!   const string &nm = id.Name();
  //!   ostream      &os = id.Stream();
  //!
  //!   //
  //!   // Use "os" and "nm" as needed
  //!   //
  //!
  //! }
  //!
  //! // Code which calls the above function
  //! int main()
  //! {
  //!   // Using file name
  //!   some_function("/proc/cpuinfo");
  //!
  //!   // Using ofstream without specifying stream name
  //!   ofstream ofs1;
  //!   ofs1.open("/proc/cpuinfo");
  //!   some_function(ofs1);
  //!   ofs1.close();
  //!
  //!   // Using ofstream with specifying stream name
  //!   ofstream ofs2;
  //!   ofs2.open("/proc/cpuinfo");
  //!   some_function(CDataSink(ofs2, "/proc/cpuinfo"));
  //!   ofs2.close();
  //!
  //!   return(0);
  //! }
  //! \endcode
  class VGE_EXPORT CDataSink : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Open the given file name in write only mode and use it as the output sink.
      CDataSink(const char *fileName);
      //!
      //! Open the given file name in write only mode and use it as the output sink.
      CDataSink(const string &fileName);
      //!
      //! Use the given ostream instance as the output sink (and optionally names the ostream).
      //! \param[in] is         An instance to an ostream derived class to be used as the output sink.
      //! \param[in] streamName Optional name for the stream.
      CDataSink(ostream &os, const char *streamName = "[unnamed ostream]");
      //!
      //! Standard dtor
      ~CDataSink() throw();

      // Getters
      //!
      //! Returns the ostream instance.
      inline ostream &Stream() const
      { return(*_outStream); }
      //!
      //! Returns the file name or the stream name.
      inline const string &Name() const
      { return(_strName); }

    protected:
      // Make sure that this class is uncopyable unless the user has derived it
      CDataSink(const CDataSink&);
      CDataSink &operator=(const CDataSink&);

      // Helper functions
      void _InitFromFile_();

      // Data
      bool         _ownStream;     // Flag to indice if this class own the ostream instance or not
      string       _strName;       // Name of the file or the stream
      ostream     *_outStream;     // The ostream instance
      ios::iostate _orgExceptions; // The original exception settings of the ostream instance
  };

} // namespace vge
#endif
