// ############################################################################################
//  MemoryOperations.cc
//  Implementation of memory manipulation functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "MemoryOperations.h"


namespace vge {

//
// Aligned memory allocation functions
//
void *amalloc(size_t szSize, size_t szAlignment)
{
#ifdef _XOPEN_SOURCE
  // Make sure that the alignment >= sizeof(void*)
  szAlignment = max(szAlignment, sizeof(void*));

  // Allocate aligned memory
  void *pvTmp;
  int r = posix_memalign(&pvTmp, szAlignment, szSize);
  if(r) {
    if(r == EINVAL)
      VError(VERR_InvalidParameter, _NFL_, "The alignment must be a power of two and a multiple of %zd (size of void *)\n", sizeof(void *));
    if(r == ENOMEM)
      VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory\n");
    return(NULL);
  }
  memset(pvTmp, 0, szSize);
  return(pvTmp);
#else
    // Show information
    static bool displayed = false;
    if(!displayed) {
      displayed = true;
      VInfo(_NFL_, "posix_memalign(void **, size_t, size_t) is not supported\nUsing address alignment workaround (may cause performance degradation)");
    }

    // Allocate memory
    const size_t vv = sizeof(void*);
    const size_t sz = szSize + (szAlignment + vv) * 2;
    void*        pt = malloc(sz);
    if(!pt) {
      VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory\n");
      return(NULL);
    }
    memset(pt, 0, sz);

    // Simulate alignment
    uintptr_t pn = reinterpret_cast<uintptr_t>(pt);
    uintptr_t bn = (pn + szAlignment + vv) & (~static_cast<uintptr_t>(szAlignment - 1));
    *reinterpret_cast<uintptr_t*>(bn - vv) = pn;
    return(reinterpret_cast<void*>(bn));
    //VPrint("M: Real=%p Simulated=%p Secret=%p (%zu)\n", pt, reinterpret_cast<void*>(bn), reinterpret_cast<uintptr_t*>(bn - vv), pn);
#endif
}

void *agetrealaddress(void *pvBuffer)
{
#ifdef _XOPEN_SOURCE
  return(pvBuffer);
#else
  uintptr_t bn = reinterpret_cast<uintptr_t>(pvBuffer);
  uintptr_t pn = *reinterpret_cast<uintptr_t*>(bn -sizeof(void*));
  return(reinterpret_cast<void*>(pn));
  //VPrint("F: Real=%p Simulated=%p Secret=%p (%zu)\n", reinterpret_cast<void*>(pn), pvBuffer, reinterpret_cast<uintptr_t*>(bn -sizeof(void*)), pn);
#endif
}

void afree(void *pvBuffer)
{
  free(agetrealaddress(pvBuffer));
}

} // namespace vge
