/** \file DialogBase.h */
// ############################################################################################
//  DialogBase.h
//  Definition  & implementation of VortexGE base dialog controls
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___DIALOG_BASE_H_
#define _VIDEO_ENVIRONMENT___DIALOG_BASE_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"
#include "DialogTypeDefs.h"


namespace vge {

  //!
  //! \internal Default text to be used for measuring UTF-8 font height
  extern const char *_DTM;


  //!
  //! \internal
  //! \brief The base of all control classes
  //!
  //! \api beta
  class VGE_EXPORT _CControl : VCLASS_UNCOPYABLE {
    public:
      virtual ~_CControl() throw();

      // Initialization functions
      //!
      //! Initialize the _tBD member of this control, meant to be called by derivative classes.
      void InitBaseData(CVideoEnv &vidEnv, const _CControl *pcsParent, ECtrlID eID, int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height);

      // User ID functions
      //!
      //! Set a custom user ID to be used.
      inline void SetUserID(const char *pcNewUserID)
      { _tBD.sUserID = pcNewUserID ? pcNewUserID : ""; }
      //!
      //! Returns the custom user ID.
      inline const char *GetUserID() const 
      { return(_tBD.sUserID.c_str()); }

      // Tooltip functions
      //!
      //! Set a tooltip to be used with this control.
      inline void SetTooltip(const char *tooltip)
      { _tBD.sToolTip = tooltip ? tooltip : ""; }
      //!
      //! Returns the tooltip of this control.
      inline const char *GetTooltip() const 
      { return(_tBD.sToolTip.c_str()); }

      // Information functions
      //!
      //! Get the control's ID.
      inline ECtrlID GetControlID() const
      { return(_tBD.eID); }
      //!
      //! Get the control's relative top-left coordinate.
      inline void GetRelativeOrigin(int32_t &x, int32_t &y) const;
      //!
      //! Get the control's absolute top-left coordinate.
      inline void GetAbsoluteOrigin(int32_t &x, int32_t &y) const
      { x = _tBD.n32X1; y = _tBD.n32Y1; }
      //!
      //! Get the control's relative bottom-right coordinate.
      inline void GetRelativeEnd(int32_t &x, int32_t &y) const;
      //!
      //! Get the control's absolute bottom-right coordinate.
      inline void GetAbsoluteEnd(int32_t &x, int32_t &y) const
      { x = _tBD.n32X2; y = _tBD.n32Y2; }
      //!
      //! Get the control's size
      inline void GetSize(uint32_t &w, uint32_t &h) const
      { w = _tBD.un32W; h = _tBD.un32H; }
      //!
      //! Get the control's width
      inline uint32_t GetWidth() const
      { return(_tBD.un32W); }
      //!
      //! Get the control's height
      inline uint32_t GetHeight() const
      { return(_tBD.un32H); }

      // Control functions
      //!
      //! Enable / disable this control.
      inline void SetEnable(bool bEnabled = true)
      { _tBD.bEnabled = bEnabled; }
      //!
      //! Check if this control is enabled or disabled.
      inline bool GetEnable() const 
      { return(_tBD.bEnabled); }

      // Mouse functions
      //!
      //! Check if the mouse pointer is over the control's region.
      bool MouseOver(int32_t n32MouseX, int32_t n32MouseY) const;
      //!
      //! Check if the mouse pointer is over the given region.
      bool MouseOver(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2,
                     int32_t n32MouseX, int32_t n32MouseY) const;

      // Drawing functions
      //!
      //! Draw the control in its normal appearance.
      virtual void DrawControl() const = 0;
      //!
      //! Draw the control in its normal appearance if the mouse pointer is not
      //! in the control's area or no button pressed, otherwise draw the control
      //! in its pressed appearance (if a control does not have pressed appearance,
      //! it will just draw the normal appearance).
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const = 0;

      // Event awareness information functions
      //!
      //! Controls that needn't process mouse events (button press, button release,
      //! and pointer movements) should returns 'false' by overriding this function.
      virtual bool CanProcess_MouseEvents() const { return(true); }

    protected:
      // This structure is meant to be accessed directly by all derivative control classes.
      struct _TBASEDATA {
        ECtrlID          eID;          // Control ID
        string           sUserID;      // User specific ID (control's name)
        string           sToolTip;     // Tooltip
        CVideoEnv       *pcsVidEnv;    // Pointer to video environment class
        const _CControl *pcsParent;    // Pointer to parent control
        bool             bEnabled;     // This control's enabled flag
        uint32_t         un32W, un32H; // This control's width and height
        int32_t          n32X1, n32Y1; // Top-left coordinate of this control
        int32_t          n32X2, n32Y2; // Bottom-right coordinate of this control

        inline _TBASEDATA()
        : eID(ECID_Control), pcsVidEnv(NULL), pcsParent(NULL), bEnabled(false),
          un32W(0), un32H(0), n32X1(0), n32Y1(0), n32X2(0), n32Y2(0)
        {}
      } _tBD;

      // Friend classes
      friend class CVideoEnv;
  };

  void _CControl::GetRelativeOrigin(int32_t &x, int32_t &y) const
  {
    if(!_tBD.pcsParent) 
      GetAbsoluteOrigin(x, y);
    else {
      int32_t px, py;
      _tBD.pcsParent->GetAbsoluteOrigin(px, py);
      x = _tBD.n32X1 - px; y = _tBD.n32Y1 - py;
    }
  }

  void _CControl::GetRelativeEnd(int32_t &x, int32_t &y) const
  {
    if(!_tBD.pcsParent) 
      GetAbsoluteEnd(x, y);
    else {
      int32_t px, py;
      _tBD.pcsParent->GetAbsoluteOrigin(px, py);
      x = _tBD.n32X2 - px; y = _tBD.n32Y2 - py;
    }
  }


  //!
  //! \internal
  //! \brief The base for all control classes which store string
  //!
  //! \api beta
  class VGE_EXPORT _CStringControl : public _CControl {
    public:
      _CStringControl();
      ~_CStringControl() throw();

      // String functions
      //!
      //! Clear the stored string.
      inline void ClearString()
      { _tSD._str.clear(); _tSD._len = 0; }
      //!
      //! Set a new string as the stored string.
      inline void SetString(const char *pcNewString)
      { _tSD._str = pcNewString ? pcNewString : ""; _tSD._len = _tSD._str.length(); }
      //!
      //! Get the stored string.
      inline const char *GetString() const
      { return(_tSD._str.c_str()); }
      //!
      //! Return the length of the stored string.
      inline uint32_t StringLen() const
      { return(_tSD._len); }

    protected:
      // This structure is meant to be accessed directly by 
      // all derivative control classes.
      struct _TSTRINGDATA {
        string _str;
        size_t _len;
      } _tSD;

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief A class used for storing controls in a list
  //!
  //! \api pretty stable
  class VGE_EXPORT CControlList {
    protected:
      class _It;
      typedef list<_CControl*> _CControlList;

    public:
      typedef       _It iterator;
      typedef const _It const_iterator;

    public:
      // Registration functions
      //!
      //! Add a control to this list.
      void add(_CControl *control);
      //!
      //! Add a control to this list.
      inline void add(_CControl &control)
      { add(&control); }
      //!
      //! Add all the controls listed in the given list to this list.
      void add(CControlList &clist);
      //!
      //! Remove a control from this list.
      inline void remove(_CControl *control)
      { _clist.remove(control); }
      //!
      //! Remove a control from this list.
      inline void remove(_CControl &control)
      { remove(&control); }
      //!
      //! Remove all the controls listed in the given list from this list.
      void remove(CControlList &clist);
      //!
      //! Remove all controls from this list.
      void clear()
      { _clist.clear(); }

      // Iterator support
      //!
      //! Returns an iterator pointing to the first element.
      iterator begin();
      //!
      //! Returns a constant iterator pointing to the first element.
      const_iterator begin() const;
      //!
      //! Returns an iterator pointing to position after the last element.
      iterator end();
      //!
      //! Returns a constant iterator pointing to position after the last element.
      const_iterator end() const;

    protected:
      _CControlList _clist;
  };

  //!
  //! \internal
  //! \brief An iterator class used by the CControlList class
  //!
  //! \api pretty stable
  class VGE_EXPORT CControlList::_It {
    public:
      inline _It(_CControlList::iterator it) : _it(it) {}

      inline       _CControl &operator* ()                    { return(**_it); }
      inline const _CControl &operator* ()              const { return(**_it); }
      inline       _CControl *operator->()                    { return(*_it); }
      inline const _CControl *operator->()              const { return(*_it); }
      inline void             operator++()              const { _it++; }
      inline bool             operator!=(const _It &it) const { return(_it != it._it); }

    private:
      mutable _CControlList::iterator _it;
  };

} // namespace vge
#endif
