// ############################################################################################
//  DialogPage.cc
//  Implementation of VortexGE dialog pages
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../HighLevelClass/ApplicationObject.h"
#include "DialogPage.h"


namespace vge {

//
// A structure used for storing keyboard shortcut's key combination
//
bool CDialogPage::_TKsc::operator==(const _TKsc &val) const
{
  if(keySymbol == val.keySymbol &&
     modCtrl   == val.modCtrl   &&
     modAlt    == val.modAlt    &&
     modShift  == val.modShift) {
    if(keySymbol == VKS_UNKNOWN) return(keyString == val.keyString);
    else                         return(keySymbol == val.keySymbol);
  }
  return(false);
}

bool CDialogPage::_TKsc::operator<(const _TKsc &val) const
{
  if(keySymbol != val.keySymbol) 
    return(keySymbol < val.keySymbol);
  else if(keySymbol == VKS_UNKNOWN) {
    if(keyString != val.keyString) return(keyString < val.keyString);
  }

       if(modCtrl != val.modCtrl) return(modCtrl   < val.modCtrl);
  else if(modAlt  != val.modAlt)  return(modAlt    < val.modAlt);
  else                            return(modShift  < val.modShift);
}


//
//  A class used for managing a dialog page (a set of controls)
//
CDialogPage::CDialogPage(CBasicAppObjects  &appObjs,
                         CEventHandlerBase *eventHandler)
: _enabled      (true),
  _iEventHandler(eventHandler),
  _iVideoEnv    (&appObjs.iVideoEnv()),
  _i3DEnv       (&appObjs.i3DEnv())
{
}

CDialogPage::CDialogPage(CVideoEnv         &iVideoEnv,
                         C3DEnv            *i3DEnv,
                         CEventHandlerBase *eventHandler)
: _enabled      (true),
  _iEventHandler(eventHandler),
  _iVideoEnv    (&iVideoEnv),
  _i3DEnv       (i3DEnv)
{
}

CDialogPage::~CDialogPage() throw()
{
  for(size_t i = 0; i < _rbgList.size(); ++i) delete _rbgList[i];
  _rbgList.clear();
}

void CDialogPage::RegisterControl(_CControl &ctrl, const char *tooltip)
{
  _ctlList.add(&ctrl);
  if(tooltip) ctrl.SetTooltip(tooltip);
}

void CDialogPage::GroupRadioButtons(CRadioButton *ctrl, ...)
{
  // Create a temporary group and fill it with the CRadioButton
  // instances to be grouped
  _CRBGroup tmpGroup;
  tmpGroup.push_back(ctrl);
  va_list vl;
  va_start(vl, ctrl);
    for(;;) {
      CRadioButton *rb = va_arg(vl, CRadioButton*);
      if(!rb) break;
      tmpGroup.push_back(rb);
    }
  va_end(vl);

  // Create a new group
  _CRBGroup *grp = new _CRBGroup;
  grp->swap(tmpGroup);

  // Adjust the group map
  try {
    _CRBGMap tmpMap = _rbgMap;
    for(size_t i = 0; i < grp->size(); ++i) tmpMap[(*grp)[i]] = grp;
    _rbgMap.swap(tmpMap);
  }
  catch(...) {
    delete grp;
    throw;
  }
}

void CDialogPage::InitControl(CLabel &ctrl, const CGLayout &parent, size_t blockIndex, const string& caption, EHAlignment captionHOrigin, EVAlignment captionVOrigin)
{
  int32_t x;
  if(captionHOrigin == EHTA_Center)
    x = (parent[blockIndex].x1 + parent[blockIndex].x2) / 2;
  else if(captionHOrigin == EHTA_Right)
    x = parent[blockIndex].x2;
  else // EHTA_Left
    x = parent[blockIndex].x1;

  int32_t y;
  if(captionVOrigin == EVTA_Center)
    y = (parent[blockIndex].y1 + parent[blockIndex].y2) / 2;
  else if(captionVOrigin == EVTA_Bottom)
    y = parent[blockIndex].y2;
  else // EVTA_Top
    y = parent[blockIndex].y1;

  InitControl(ctrl, &parent, x, y, caption, captionHOrigin, captionVOrigin); 
}

void CDialogPage::RedrawAllControls()
{
  if(!_enabled) return;

  for(CControlList::const_iterator it  = _ctlList.begin();
                                   it != _ctlList.end(); 
                                 ++it) {
    it->DrawControl();
  }
  CustomDraw_OnRedrawAllControls();
}

bool CDialogPage::HandleKeyboardShortcut(EKeySymbol keySymbol, const char *str, bool modCtrl, bool modAlt, bool modShift)
{
  _CKeySCutEHMap::const_iterator it = _mapKeySCutEH.find(_TKsc(keySymbol,   modCtrl, modAlt, modShift));
  if(it == _mapKeySCutEH.end())  it = _mapKeySCutEH.find(_TKsc(string(str), modCtrl, modAlt, modShift));
  if(it == _mapKeySCutEH.end()) return(false);

  if(!it->second) return(false);

  _PFNKsc pfn = it->second;
  (this->*pfn)();

  return(true);
}

void CDialogPage::HandleEvent_OnButtonPress(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY)
{
  // Check if the dialog page is currently disabled or the
  // pressed mouse button is not the "left click button"
  if(!_enabled || buttonSymbol!=TMouseSettings::GLOBAL_SET.leftClickButton) return;

  // Display no tooltip
  DisplayTooltip("");

  // Walk trough all the controls
  for(CControlList::const_iterator it  = _ctlList.begin();
                                   it != _ctlList.end(); 
                                 ++it) {
    // Skip the control if the mouse pointer is not in the control's area
    if(!it->MouseOver(posX, posY)) continue;
    // Display tooltip from the control
    DisplayTooltip(it->GetTooltip());
    // Process further if the control can process mouse events
    if(it->CanProcess_MouseEvents()) {
      it->DrawControl(posX, posY, true);
      CustomDraw_OnButtonPress(buttonSymbol);
      _iVideoEnv->UpdateDisplay();
      break;
    }
  }
}

void CDialogPage::HandleEvent_OnButtonRelease(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY)
{
  // Check if the dialog page is currently disabled or the
  // pressed mouse button is not the "left click button"
  if(!_enabled || buttonSymbol!=TMouseSettings::GLOBAL_SET.leftClickButton) return;

  // Trap double click
  static uint64_t oldDTM = 0;
  const  uint64_t curDTM = _iVideoEnv->GetButtonTimeMs();
  const  bool     gotDBL = (curDTM - oldDTM) < uint64_t(TMouseSettings::GLOBAL_SET.doubleClickDelay);
                  oldDTM = curDTM;

  // Process all the controls
  bool processed = false;
  for(CControlList::iterator it  = _ctlList.begin();
                             it != _ctlList.end(); 
                             ++it) {
    if(!it->CanProcess_MouseEvents()) continue;
    switch(it->GetControlID()) {
      // Process all CButton instances
      case ECID_Button: {
        CButton &ctrl = dynamic_cast<CButton&>(*it);
        if(ctrl.MouseOver(posX, posY)) {
          ctrl.DrawControl(false);
          OnCommandButton(ctrl);
          processed = true;
        }
        break;
      }
      // Process all CCheckBox instances
      case ECID_CheckBox: {
        CCheckBox &ctrl = dynamic_cast<CCheckBox&>(*it);
        if(ctrl.ToggleValue(posX, posY)) {
          ctrl.DrawControl(false);
          OnCheckBox(ctrl);
          processed = true;
        }
        break;
      }
      // Process all CRadioButton instances
      case ECID_RadioButton: {
        CRadioButton             &ctrl = dynamic_cast<CRadioButton&>(*it);
        _CRBGMap::const_iterator  rbgi = _rbgMap.find(&ctrl);
        if(rbgi == _rbgMap.end())
          VError(VERR_NotDefined, _NFL_, "The CRadioButton instance %p has no group associated with it", &ctrl);
        _CRBGroup &rbg = *rbgi->second;
        if(ctrl.ToggleValue(posX, posY, rbg)) {
          for(size_t i = 0; i < rbg.size(); ++i) {
            rbg[i]->DrawControl(false);
          }
          OnRadioButton(ctrl);
          processed = true;
        }
        break;
      }
      // Process all CSpinButton instances
      case ECID_SpinButton: {
        CSpinButton &ctrl = dynamic_cast<CSpinButton&>(*it);
        if(ctrl.ToggleValue(posX, posY)) {
          ctrl.UpdateValue();
          ctrl.DrawControl(false, false);
          OnSpinButton(ctrl);
          processed = true;
        }
        break;
      }
      // Process all CListBox instances
      case ECID_ListBox: {
        CListBox &ctrl = dynamic_cast<CListBox&>(*it);
        if(ctrl.ToggleValue(posX, posY)) {
          ctrl.DrawControl(false, false, false, 0);
          OnListBox(ctrl, gotDBL);
          processed = true;
        }
        break;
      }
      // Process all CSLTextBox instances
      case ECID_SLTextBox: {
        CSLTextBox &ctrl = dynamic_cast<CSLTextBox&>(*it);
        if(ctrl.MouseOver(posX, posY)) {
          const bool accepted = ctrl.DoEdit(_iEventHandler);
          ctrl.DrawControl();
          OnSLTextBox(ctrl, accepted);
          processed = true;
        }
        break;
      }
      // Process all CMLTextBox instances
      case ECID_MLTextBox: {
        CMLTextBox &ctrl = dynamic_cast<CMLTextBox&>(*it);
        if(ctrl.MouseOver(posX, posY)) {
          const bool accepted = ctrl.DoEdit(_iEventHandler);
          ctrl.DrawControl();
          OnMLTextBox(ctrl, accepted);
          processed = true;
        }
        break;
      }
      // Do nothing for other controls
      default:
        break;
    }
    // If already processed, exit the loop
    if(processed) break;
  }

  // Process further if processed
  if(processed) {
    // Invoke the custom drawing function if the user controls' event
    // handlers invoked by the above code didn't disable the page
    if(_enabled) CustomDraw_OnButtonRelease(buttonSymbol);
    // Update display
    _iVideoEnv->UpdateDisplay();
  }
}

void CDialogPage::HandleEvent_OnMouseMove(int32_t pointerXPos, int32_t pointerYPos, bool buttonState[5])
{
  // Check if the dialog page is currently disabled
  if(!_enabled) return;

  // Check the event time so that we won't need to handle
  // mouse pointer move event to often
  static uint64_t oldMTM = 0;
  const  uint64_t curMTM = _iVideoEnv->GetButtonTimeMs();
  if(curMTM-oldMTM < 50) return ;
  oldMTM = curMTM;

  // Display no tooltip
  DisplayTooltip("");

  // Get the "left click button" state
  const bool &pressed = buttonState[TMouseSettings::GLOBAL_SET.leftClickButton - 1];

  //! \todo Possible optimization : remember the last control which previously
  //!                               owns the mouse pointer so there is no need
  //!                               to redraw all the controls

  // Walk trough all the controls
  for(CControlList::const_iterator it  = _ctlList.begin();
                                   it != _ctlList.end(); 
                                 ++it) {
    // Display tooltip from the control
    if(it->MouseOver(pointerXPos, pointerYPos)) DisplayTooltip(it->GetTooltip());
    // Redraw if the control can process mouse events
    if(it->CanProcess_MouseEvents()) it->DrawControl(pointerXPos, pointerYPos, pressed);
  }
  CustomDraw_OnMouseMove(pointerXPos, pointerYPos);
  _iVideoEnv->UpdateDisplay();
}

void CDialogPage::HandleEvent_OnWindowActivated()
{
  if(!_enabled) return;

  DisplayTooltip("");
}

void CDialogPage::HandleEvent_OnWindowDeactivated()
{
  if(!_enabled) return;

  DisplayTooltip("");
}


//
// Controls' event handlers
//
#define _DELEGATE___ControlEH_(NameC, CtrlI)                     \
  if(!_enabled) return;                                          \
  _C##NameC##EHMap::iterator it = _mapC##NameC##EH.find(&CtrlI); \
  if(it != _mapC##NameC##EH.end()) (this->*(it->second))(&CtrlI);

#define _DELEGATE___ControlEH_B_(NameC, CtrlI, BoolI)                   \
  if(!_enabled) return;                                                 \
  _C##NameC##EHMap::iterator it = _mapC##NameC##EH.find(&CtrlI);        \
  if(it != _mapC##NameC##EH.end()) (this->*(it->second))(&CtrlI, BoolI);

void CDialogPage::OnCommandButton(CButton &ctrl)
{ _DELEGATE___ControlEH_(Button, ctrl); }

void CDialogPage::OnCheckBox(CCheckBox &ctrl)
{ _DELEGATE___ControlEH_(CheckBox, ctrl); }

void CDialogPage::OnRadioButton(CRadioButton &ctrl)
{ _DELEGATE___ControlEH_(RadioButton, ctrl); }

void CDialogPage::OnSpinButton(CSpinButton &ctrl)
{ _DELEGATE___ControlEH_(SpinButton, ctrl); }

void CDialogPage::OnListBox(CListBox &ctrl, bool doubleClick)
{ _DELEGATE___ControlEH_B_(ListBox, ctrl, doubleClick); }

void CDialogPage::OnSLTextBox(CSLTextBox &ctrl, bool accepted)
{ _DELEGATE___ControlEH_B_(SLTextBox, ctrl, accepted); }

void CDialogPage::OnMLTextBox(CMLTextBox &ctrl, bool accepted)
{ _DELEGATE___ControlEH_B_(MLTextBox, ctrl, accepted); }


//
//  A class used for managing a dialog pages
//
CDialogPageManager::CDialogPageManager()
{
}

CDialogPageManager::CDialogPageManager(CDialogPage &mainPageInst, const string &mainPageName)
{
  RegisterPage(mainPageInst, mainPageName);
}

CDialogPageManager::~CDialogPageManager() throw()
{
}

void CDialogPageManager::RegisterPage(CDialogPage &pageInst, const string &pageName)
{
  _CPageMap::const_iterator it = _pages.find(pageName);
  if(it!=_pages.end() && it->second!=&pageInst) 
    VError(VERR_Exist, _NFL_, "A dialog page with the same name is already exist");

  // If no main page is set yet, set it first
  if(_mainPage.empty()) {
    _mainPage   = pageName;
    _activePage = pageName;
    _pages[pageName] = &pageInst;
    _pages[pageName]->SetEnable();
    _pages[pageName]->OnPageSelected();
    return;
  }

  // Other pages
  _pages[pageName] = &pageInst;
  _pages[pageName]->SetEnable(false);
  _pages[pageName]->OnPageRegistered();
}

void CDialogPageManager::SelectPage(const string &pageName)
{
  if(_pages.find(pageName) == _pages.end())
    VError(VERR_Exist, _NFL_, "No dialog page with the given name is exist");

  _pages[_activePage]->DisplayTooltip("");
  _pages[_activePage]->OnPageDeselected();

  for(_CPageMap::iterator it = _pages.begin(); it != _pages.end(); ++it)
    it->second->SetEnable(false);

  _activePage = pageName;
  _pages[_activePage]->SetEnable();
  _pages[_activePage]->OnPageSelected();
}

void CDialogPageManager::RedrawPage()
{
  if(_pages.empty()) return;
  _pages[_activePage]->RedrawAllControls();
}

void CDialogPageManager::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnKeyPress(keySymbol, str);
}

void CDialogPageManager::HandleEvent_OnKeyRelease(EKeySymbol keySymbol, const char *str)
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnKeyRelease(keySymbol, str);
}

void CDialogPageManager::HandleEvent_OnButtonPress(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY)
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnButtonPress(buttonSymbol, posX, posY);
}

void CDialogPageManager::HandleEvent_OnButtonRelease(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY)
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnButtonRelease(buttonSymbol, posX, posY);
}

void CDialogPageManager::HandleEvent_OnMouseMove(int32_t posX, int32_t posY, bool buttonState[5])
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnMouseMove(posX, posY, buttonState);
}

void CDialogPageManager::HandleEvent_OnUpdateFrame(bool windowIsActive)
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnUpdateFrame(windowIsActive);
}

void CDialogPageManager::HandleEvent_OnWindowActivated()
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnWindowDeactivated();
  _pages[_activePage]->RedrawAllControls();
}

void CDialogPageManager::HandleEvent_OnWindowDeactivated()
{
  if(_pages.empty()) return;
  _pages[_activePage]->HandleEvent_OnWindowDeactivated();
  _pages[_activePage]->RedrawAllControls();
}

} // namespace vge
