/** \file DialogPage.h */
// ############################################################################################
//  DialogPage.h
//  Definition & implementation of VortexGE dialog pages
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___DIALOG_PAGE_H_
#define _VIDEO_ENVIRONMENT___DIALOG_PAGE_H_

// VortexGE headers
#include "DialogLayout.h"
#include "../VideoEnvironment/VideoEnvironment.h"


namespace vge {

  // Foward reference from "lib/3DEnvironment/3DEnvironment.h"
  class C3DEnv;

  // Foward reference from "lib/HighLevelClass/ApplicationObject.h"
  class CBasicAppObjects;


  //!
  //! \brief A class used for managing a dialog page (a set of controls)
  //!
  //! \see CDialogPageManager, for easier usage (management) of this class.
  //!
  //! \api experiment
  class VGE_EXPORT CDialogPage : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct the CDialogPage instance using the given parameters.
      CDialogPage(CBasicAppObjects  &appObjs,
                  CEventHandlerBase *eventHandler = NULL);
      //!
      //! Construct the CDialogPage instance using the given parameters.
      CDialogPage(CVideoEnv         &iVideoEnv,
                  C3DEnv            *i3DEnv       = NULL,
                  CEventHandlerBase *eventHandler = NULL);
      //!
      //! Standard dtor.
      virtual ~CDialogPage() throw();

      // Generic initialization functions
      //!
      //! Must be overriden by the derivative class to initialize all its
      //! control instances.
      virtual void InitializeAllControls() = 0;
      //!
      //! Must be called by the derivative class to register all its
      //! control instances if they are not initialized using one of
      //! CDialogPage::InitControl()s.
      void RegisterControl(_CControl &ctrl, const char *tooltip = NULL);
      //!
      //! Must be called by the derivative class to group all its
      //! CRadioButton instance.
      void GroupRadioButtons(CRadioButton *ctrl, ...);
      //!
      //! Use this to enable (or disable) the dialog page.
      //!
      //! If not using CDialogPageManager class, all not selected (displayed)
      //! dialog pages should be disabled manually.
      inline void SetEnable(bool bEnabled = true)
      { _enabled = bEnabled; }

      // Keyboard shortcuts registration functions
      //!
      //! Used for registering a keyboard shortcut to an event handler.
      template <typename DialogPageT> inline
      void RegisterKeyboardShortcut(void(DialogPageT::*kscEH)(), char keyChar, bool modCtrl = false, bool modAlt = false, bool modShift = false)
      { _mapKeySCutEH[_TKsc(keyChar, modCtrl, modAlt, modShift)] = _PFNKsc(kscEH); }
      //!
      //! Used for registering a keyboard shortcut to an event handler.
      template <typename DialogPageT> inline
      void RegisterKeyboardShortcut(void(DialogPageT::*kscEH)(), EKeySymbol keySymbol, bool modCtrl = false, bool modAlt = false, bool modShift = false)
      {_mapKeySCutEH[_TKsc(keySymbol, modCtrl, modAlt, modShift)] = _PFNKsc(kscEH); }
      //!
      //! Used for registering a keyboard shortcut to an event handler.
      template <typename DialogPageT> inline
      void RegisterKeyboardShortcut(void(DialogPageT::*kscEH)(), const string &keyString, bool modCtrl = false, bool modAlt = false, bool modShift = false)
      { _mapKeySCutEH[_TKsc(keyString, modCtrl, modAlt, modShift)] = _PFNKsc(kscEH); }

      // Controls' event handlers setters
      // Functions to connect a control to an event handler
      // 
      //!
      //! Event handler setter for CButton.
      template<typename DialogPageT> inline
      void Connect(CButton &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl))
      { _mapCButtonEH[&ctrl] = _PFN_CButtonEH(ctrlEH); }
      //!
      //! Event handler setter for CCheckBox.
      template<typename DialogPageT> inline
      void Connect(CCheckBox &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl))
      { _mapCCheckBoxEH[&ctrl] = _PFN_CCheckBoxEH(ctrlEH); }
      //!
      //! Event handler setter for CRadioButton.
      template<typename DialogPageT> inline
      void Connect(CRadioButton &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl))
      { _mapCRadioButtonEH[&ctrl] = _PFN_CRadioButtonEH(ctrlEH); }
      //!
      //! Event handler setter for CSpinButton.
      template<typename DialogPageT> inline
      void Connect(CSpinButton &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl))
      { _mapCSpinButtonEH[&ctrl] = _PFN_CSpinButtonEH(ctrlEH); }
      //!
      //! Event handler setter for CListBox.
      template<typename DialogPageT> inline
      void Connect(CListBox &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl, bool doubleClick))
      { _mapCListBoxEH[&ctrl] = _PFN_CListBoxEH(ctrlEH); }
      //!
      //! Event handler setter for CSLTextBox.
      template<typename DialogPageT> inline
      void Connect(CSLTextBox &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl, bool accepted))
      { _mapCSLTextBoxEH[&ctrl] = _PFN_CSLTextBoxEH(ctrlEH); }
      //!
      //! Event handler setter for CMLTextBox.
      template<typename DialogPageT> inline
      void Connect(CMLTextBox &ctrl, void(DialogPageT::*ctrlEH)(_CControl *ctrl, bool accepted))
      { _mapCMLTextBoxEH[&ctrl] = _PFN_CMLTextBoxEH(ctrlEH); }

      // Convenience initialization functions for initializing controls
      //!
      //! Initialize control: CGLayout and its derivative classes.
      inline void InitControl(CGLayout &ctrl, const _CControl *parent, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE)
      { ctrl.InitControl(*_iVideoEnv, parent, 0, borderWidth, spacing); }
      //!
      //! Initialize control: CGLayout and its derivative classes.
      inline void InitControl(CGLayout &ctrl, const CGLayout &parent, size_t blockIndex, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE)
      { ctrl.InitControl(*_iVideoEnv, &parent, blockIndex, borderWidth, spacing); }
      //!
      //! Initialize compound: CHVLayoutCompound.
      inline void InitCompound(CHVLayoutCompound &cmpnd, const _CControl *parent, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE)
      { cmpnd.InitCompound(*_iVideoEnv, parent, 0, borderWidth, spacing); }
      //!
      //! Initialize compound: CHVLayoutCompound.
      inline void InitCompound(CHVLayoutCompound &cmpnd, const CGLayout &parent, size_t blockIndex, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE)
      { cmpnd.InitCompound(*_iVideoEnv, &parent, blockIndex, borderWidth, spacing); }
      //!
      //! Initialize compound: CHVHayoutCompound.
      inline void InitCompound(CVHLayoutCompound &cmpnd, const _CControl *parent, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE)
      { cmpnd.InitCompound(*_iVideoEnv, parent, 0, borderWidth, spacing); }
      //!
      //! Initialize compound: CHVLayoutCompound.
      inline void InitCompound(CVHLayoutCompound &cmpnd, const CGLayout &parent, size_t blockIndex, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE)
      { cmpnd.InitCompound(*_iVideoEnv, &parent, blockIndex, borderWidth, spacing); }
      //!
      //! Initialize control: CPanel.
      inline void InitControl(CPanel &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h); RegisterControl(ctrl); }
      //!
      //! Initialize control: CPanel.
      inline void InitControl(CPanel &ctrl, const CGLayout &parent, size_t blockIndex)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex)); }
      //!
      //! Initialize control: CPanelBar.
      inline void InitControl(CPanelBar &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, const string& caption)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, caption.c_str()); RegisterControl(ctrl); }
      //!
      //! Initialize control: CPanelBar.
      inline void InitControl(CPanelBar &ctrl, const CGLayout &parent, size_t blockIndex, const string& caption)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), caption); }
      //!
      //! Initialize control: CStatusBar.
      inline void InitControl(CStatusBar &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, const string& text)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, text.c_str()); RegisterControl(ctrl); }
      //!
      //! Initialize control: CStatusBar.
      inline void InitControl(CStatusBar &ctrl, const CGLayout &parent, size_t blockIndex, const string& text)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), text); }
      //!
      //! Initialize control: CHLine.
      inline void InitControl(CHLine &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, EBorderStyle lineStyle = EBS_Raised)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, lineStyle ); RegisterControl(ctrl); }
      //!
      //! Initialize control: CHLine.
      inline void InitControl(CHLine &ctrl, const CGLayout &parent, size_t blockIndex, EBorderStyle lineStyle = EBS_Raised)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYcW(parent, blockIndex), lineStyle ); }
      //!
      //! Initialize control: CVLine.
      inline void InitControl(CVLine &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t h, EBorderStyle lineStyle = EBS_Raised)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, h, lineStyle ); RegisterControl(ctrl); }
      //!
      //! Initialize control: CVLine.
      inline void InitControl(CVLine &ctrl, const CGLayout &parent, size_t blockIndex, EBorderStyle lineStyle = EBS_Raised)
      { InitControl(ctrl, LAYOUT_EXPAND_PXcYH(parent, blockIndex), lineStyle ); }
      //!
      //! Initialize control: CFrame.
      inline void InitControl(CFrame &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, const string& caption, EHAlignment captionPos = EHTA_Left, EBorderStyle borderStyle = EBS_Raised)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, caption.c_str(), captionPos, borderStyle); RegisterControl(ctrl); }
      //!
      //! Initialize control: CFrame.
      inline void InitControl(CFrame &ctrl, const CGLayout &parent, size_t blockIndex, const string& caption, EHAlignment captionPos = EHTA_Left, EBorderStyle borderStyle = EBS_Raised)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), caption, captionPos, borderStyle); }
      //!
      //! Initialize control: CLabel.
      inline void InitControl(CLabel &ctrl, const _CControl *parent, int32_t x, int32_t y, const string& caption, EHAlignment captionHOrigin = EHTA_Left, EVAlignment captionVOrigin = EVTA_Top)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, caption.c_str(), captionHOrigin, captionVOrigin); RegisterControl(ctrl); }
      //!
      //! Initialize control: CLabel.
      void InitControl(CLabel &ctrl, const CGLayout &parent, size_t blockIndex, const string& caption, EHAlignment captionHOrigin = EHTA_Left, EVAlignment captionVOrigin = EVTA_Top);
      //!
      //! Initialize control: CImageDisplay (still in preliminary experiment stage).
      inline void InitControl(CImageDisplay &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, const char *xpmDataBuffer[])
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, xpmDataBuffer); RegisterControl(ctrl); }
      //!
      //! Initialize control: CImageDisplay (still in preliminary experiment stage).
      inline void InitControl(CImageDisplay &ctrl, const CGLayout &parent, size_t blockIndex, const char *xpmDataBuffer[])
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), xpmDataBuffer); }
      //!
      //! Initialize control: CButton.
      inline void InitControl(CButton &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, const string& text, bool insideOutline = false, bool thickOutline = false)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, text.c_str(), insideOutline, thickOutline); RegisterControl(ctrl); }
      //!
      //! Initialize control: CButton.
      inline void InitControl(CButton &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, const char *xpmDataBuffer[], bool insideOutline = false, bool thickOutline = false)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, xpmDataBuffer, insideOutline, thickOutline); RegisterControl(ctrl); }
      //!
      //! Initialize control: CButton.
      inline void InitControl(CButton &ctrl, const CGLayout &parent, size_t blockIndex, const string& text, bool insideOutline = false, bool thickOutline = false)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), text, insideOutline, thickOutline); }
      //!
      //! Initialize control: CButton.
      inline void InitControl(CButton &ctrl, const CGLayout &parent, size_t blockIndex, const char *xpmDataBuffer[], bool insideOutline = false, bool thickOutline = false)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), xpmDataBuffer, insideOutline, thickOutline); }
      //!
      //! Initialize control: CCheckBox.
      inline void InitControl(CCheckBox &ctrl, const _CControl *parent, int32_t x, int32_t y, const string& text, bool originatedFromRight = false)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, text.c_str(), originatedFromRight); RegisterControl(ctrl); }
      //!
      //! Initialize control: CCheckBox.
      inline void InitControl(CCheckBox &ctrl, const CGLayout &parent, size_t blockIndex, const string& text, bool originatedFromRight = false)
      { InitControl(ctrl, &parent, originatedFromRight ? parent[blockIndex].x2 : parent[blockIndex].x1, parent[blockIndex].y1, text, originatedFromRight); }
      //!
      //! Initialize control: CCheckBox.
      inline void InitControl(CCheckBox &ctrl, const CGLayout &parent, size_t blockIndex, uint32_t xOffset, const string& text, bool originatedFromRight = false)
      { InitControl(ctrl, &parent, originatedFromRight ? (parent[blockIndex].x2 - xOffset) : (parent[blockIndex].x1 + xOffset), parent[blockIndex].y1, text, originatedFromRight); }
      //!
      //! Initialize control: CRadioButton.
      inline void InitControl(CRadioButton &ctrl, const _CControl *parent, int32_t x, int32_t y, const string& text, bool originatedFromRight = false)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, text.c_str(), originatedFromRight); RegisterControl(ctrl); }
      //!
      //! Initialize control: CRadioButton.
      inline void InitControl(CRadioButton &ctrl, const CGLayout &parent, size_t blockIndex, const string& text, bool originatedFromRight = false)
      { InitControl(ctrl, &parent, originatedFromRight ? parent[blockIndex].x2 : parent[blockIndex].x1, parent[blockIndex].y1, text, originatedFromRight); }
      //!
      //! Initialize control: CRadioButton.
      inline void InitControl(CRadioButton &ctrl, const CGLayout &parent, size_t blockIndex, uint32_t xOffset, const string& text, bool originatedFromRight = false)
      { InitControl(ctrl, &parent, originatedFromRight ? (parent[blockIndex].x2 - xOffset) : (parent[blockIndex].x1 + xOffset), parent[blockIndex].y1, text, originatedFromRight); }
      //!
      //! Initialize control: CSpinButton.
      inline void InitControl(CSpinButton &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, ESBStyle spinButtonStyle = ESBS_VLeft, EHAlignment textAlignment = EHTA_Left, ESBValue valueType = ESBV_Integer, bool wrapValue = true)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, spinButtonStyle, textAlignment, valueType, wrapValue); RegisterControl(ctrl); }
      //!
      //! Initialize control: CSpinButton.
      inline void InitControl(CSpinButton &ctrl, const CGLayout &parent, size_t blockIndex, ESBStyle spinButtonStyle = ESBS_VLeft, EHAlignment textAlignment = EHTA_Left, ESBValue valueType = ESBV_Integer, bool wrapValue = true)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), spinButtonStyle, textAlignment, valueType, wrapValue); }
      //!
      //! Initialize control: CListBox.
      inline void InitControl(CListBox &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h); RegisterControl(ctrl); }
      //!
      //! Initialize control: CListBox.
      inline void InitControl(CListBox &ctrl, const CGLayout &parent, size_t blockIndex)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex)); }
      //!
      //! Initialize control: CSLTextBox.
      inline void InitControl(CSLTextBox &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, ETBType textType = ETBT_TextNormal, bool pressingEscCancel = false, PFNSLTextBoxCallback pfnUserCallback = NULL)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h, textType, pressingEscCancel, pfnUserCallback); RegisterControl(ctrl); }
      //!
      //! Initialize control: CSLTextBox.
      inline void InitControl(CSLTextBox &ctrl, const CGLayout &parent, size_t blockIndex, ETBType textType = ETBT_TextNormal, bool pressingEscCancel = false, PFNSLTextBoxCallback pfnUserCallback = NULL)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), textType, pressingEscCancel, pfnUserCallback); }
      //!
      //! Initialize control: CMLTextBox.
      inline void InitControl(CMLTextBox &ctrl, const _CControl *parent, int32_t x, int32_t y, uint32_t w, uint32_t h, bool pressingEscCancel = false, PFNMLTextBoxCallback pfnUserCallback = NULL)
      { ctrl.InitControl(*_iVideoEnv, parent, x, y, w, h,  pressingEscCancel, pfnUserCallback); RegisterControl(ctrl); }
      //!
      //! Initialize control: CMLTextBox.
      inline void InitControl(CMLTextBox &ctrl, const CGLayout &parent, size_t blockIndex, bool pressingEscCancel = false, PFNMLTextBoxCallback pfnUserCallback = NULL)
      { InitControl(ctrl, LAYOUT_EXPAND_PXYWH(parent, blockIndex), pressingEscCancel, pfnUserCallback); }

      // Drawing functions
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! redraw all the controls.
      //! \n\n
      //! \snto
      virtual void RedrawAllControls();

      // Keyboard shortcut handling functions
      //!
      //! Must be called by the class which use this CDialogPage class if it needs
      //! support for keyboard shortcut. Preferably if called from the user's overridden
      //! HandleEvent_OnKeyPress() virtual member function.
      bool HandleKeyboardShortcut(EKeySymbol keySymbol, const char *str, bool modCtrl = false, bool modAlt = false, bool modShift = false);

      // Generic event handling functions
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle key press event (if not using CDialogPageManager class).
      //! \n\n
      //! \ddncbo
      virtual void HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
      { UNUSED_PARAMS_2(keySymbol, str); }
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle key release event (if not using CDialogPageManager class).
      //! \n\n
      //! \ddncbo
      virtual void HandleEvent_OnKeyRelease(EKeySymbol keySymbol, const char *str)
      { UNUSED_PARAMS_2(keySymbol, str); }
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle mouse button press event (if not using CDialogPageManager class).
      //! \n\n
      //! \snto
      virtual void HandleEvent_OnButtonPress(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY);
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle mouse button release event (if not using CDialogPageManager class).
      //! \n\n
      //! \snto
      virtual void HandleEvent_OnButtonRelease(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY);
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle mouse pointer move event (if not using CDialogPageManager class).
      //! \n\n
      //! \snto
      virtual void HandleEvent_OnMouseMove(int32_t pointerXPos, int32_t pointerYPos, bool buttonState[5]);
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle update frame event (if not using CDialogPageManager class).
      //! \n\n
      //! \ddncbo
      virtual void HandleEvent_OnUpdateFrame(bool windowIsActive)
      { UNUSED_PARAM(windowIsActive); }
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle window activated event (if not using CDialogPageManager class).
      //! \n\n
      //! \snto
      virtual void HandleEvent_OnWindowActivated();
      //!
      //! Must be called by the class which use this CDialogPage class to
      //! handle window deactivated event (if not using CDialogPageManager class).
      //! \n\n
      //! \snto
      virtual void HandleEvent_OnWindowDeactivated();

      // Page's event handling functions
      //!
      //! Will be called by CDialogPageManager class just after registering this
      //! dialog page. All newly registered dialog pages will be in disabled state.
      //! \n\n
      //! If not using the CDialogPageManager class, this function should be called
      //! manually after all initialization of this dialog page completed and after
      //! putting this dialog page in disabled state.
      //! \n\n
      //! \ddncbo
      virtual void OnPageRegistered()
      {}
      //!
      //! Will be called by CDialogPageManager class just before deselecting (and
      //! disabling) this dialog page.
      //! \n\n
      //! If not using the CDialogPageManager class, this function should be called
      //! manually just before de-selecting (disabling) this dialog page.
      //! \n\n
      //! \ddncbo
      virtual void OnPageDeselected()
      {}
      //!
      //! Will be called by CDialogPageManager class just after selecting (and
      //! enabling) this dialog page.
      //! \n\n
      //! If not using the CDialogPageManager class, this function should be called
      //! manually just after selecting (enabling) this dialog page.
      //! \n\n
      //! \ddncbo
      virtual void OnPageSelected()
      {}

      // Controls' event handling functions
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CButton controls.
      //! \n\n
      //! \snto
      virtual void OnCommandButton(CButton &ctrl);
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CCheckBox controls.
      //! \n\n
      //! \snto
      virtual void OnCheckBox(CCheckBox &ctrl);
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CRadioButton controls.
      //! \n\n
      //! \snto
      virtual void OnRadioButton(CRadioButton &ctrl);
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CSpinButton controls.
      //! \n\n
      //! \snto
      virtual void OnSpinButton(CSpinButton &ctrl);
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CListBox controls.
      //! \n\n
      //! \snto
      virtual void OnListBox(CListBox &ctrl, bool doubleClick);
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CSLTextBox controls.
      //! \n\n
      //! \snto
      virtual void OnSLTextBox(CSLTextBox &ctrl, bool accepted);
      //!
      //! Will be called by HandleEvent_OnButtonRelease for all registered
      //! CMLTextBox controls.
      //! \n\n
      //! \snto
      virtual void OnMLTextBox(CMLTextBox &ctrl, bool accepted);

      // Make the dialog manager class as friend class
      friend class CDialogPageManager;

 protected:
      // Tooltip functions
      //!
      //! Will be called if the dialog page needs to display/set tooltip text
      //! from a particular control.
      //! \n\n
      //! \ddncbo
      virtual void DisplayTooltip(const char *tooltip)
      { UNUSED_PARAM(tooltip); }

      // Custom drawing functions
      //!
      //! Will be called by the RedrawAllControls() function after drawing
      //! all the controls.
      //! \n\n
      //! \ddncbo
      virtual void CustomDraw_OnRedrawAllControls()
      {}
      //!
      //! Will be called by the HandleEvent_OnButtonPress() function after
      //! drawing all the controls.
      //! \n\n
      //! \ddncbo
      virtual void CustomDraw_OnButtonPress(EBtnSymbol buttonSymbol)
      { UNUSED_PARAM(buttonSymbol); }
      //!
      //! Will be called by the HandleEvent_OnButtonRelease() function after
      //! drawing all the controls.
      //! \n\n
      //! \ddncbo
      virtual void CustomDraw_OnButtonRelease(EBtnSymbol buttonSymbol)
      { UNUSED_PARAM(buttonSymbol); }
      //!
      //! Will be called by the HandleEvent_OnMouseMove() function after
      //! drawing all the controls.
      //! \n\n
      //! \ddncbo
      virtual void CustomDraw_OnMouseMove(int32_t pointerXPos, int32_t pointerYPos)
      { UNUSED_PARAMS_2(pointerXPos, pointerYPos); }

      //!
      //! \brief A structure used for storing keyboard shortcut's key combination
      //!
      //! \api preliminary experiment
      struct VGE_EXPORT _TKsc {
        EKeySymbol keySymbol; //!< Key symbol.
        string     keyString; //!< Key string, only used if keySymbol is VKS_UNKNOWN.
        bool       modCtrl;   //!< Flag to indicate if ctrl-key needs to be hold.
        bool       modAlt;    //!< Flag to indicate if alt-key needs to be hold.
        bool       modShift;  //!< Flag to indicate if shift-key needs to be hold.

        //!
        //! Use the given 8-bit character as the shortcut key.
        inline _TKsc(char keyChar_)
        : keySymbol(EKeySymbol(keyChar_)), modCtrl(false), modAlt(false), modShift(false)
        {}
        //!
        //! Use the given 8-bit character and the modifier keys as the shortcut key combination.
        inline _TKsc(char keyChar_, bool modCtrl_, bool modAlt_, bool modShift_)
        : keySymbol(EKeySymbol(keyChar_)), modCtrl(modCtrl_), modAlt(modAlt_), modShift(modShift_)
        {}
        //!
        //! Use the given key symbol as the shortcut key.
        inline _TKsc(EKeySymbol keySymbol_)
        : keySymbol(keySymbol_), modCtrl(false), modAlt(false), modShift(false)
        {}
        //!
        //! Use the given key symbol and the modifier keys as the shortcut key combination.
        inline _TKsc(EKeySymbol keySymbol_, bool modCtrl_, bool modAlt_, bool modShift_)
        : keySymbol(keySymbol_), modCtrl(modCtrl_), modAlt(modAlt_), modShift(modShift_)
        {}
        //!
        //! Use the given key string (UTF-8) as the shortcut key.
        inline _TKsc(const string &keyString_)
        : keySymbol(VKS_UNKNOWN), keyString(keyString_), modCtrl(false), modAlt(false), modShift(false)
        {}
        //!
        //! Use the given key string (UTF-8) and given modifier keys as the shortcut key combination.
        inline _TKsc(const string &keyString_, bool modCtrl_, bool modAlt_, bool modShift_)
        : keySymbol(VKS_UNKNOWN), keyString(keyString_), modCtrl(modCtrl_), modAlt(modAlt_), modShift(modShift_)
        {}

        // Overloaded operators
        bool operator==(const _TKsc &val) const;
        bool operator<(const _TKsc &val) const;
      };

      // For convenience
      typedef vector<CRadioButton*>          _CRBGroup; // List of radio buttons in a particular group
      typedef vector<_CRBGroup*>             _CRBGList; // List of all radio button groups.
      typedef map<CRadioButton*, _CRBGroup*> _CRBGMap;  // Map between a radio button and its group.

      // These data can be accessed directly by derivative classes, if needed
      bool               _enabled;       //!< Enable state of the dialog page.
      CEventHandlerBase *_iEventHandler; //!< CEventHandlerBase instance used by this dialog page.
      CVideoEnv         *_iVideoEnv;     //!< CVideoEnv instance used by this dialog page.
      C3DEnv            *_i3DEnv;        //!< C3DEnv instance used by this dialog page.
      CControlList       _ctlList;       //!< List of all controls in this dialog page.
      _CRBGList          _rbgList;       //!< List of all radio button groups.
      _CRBGMap           _rbgMap;        //!< Map between a radio button and its group.

      // Keyboard shortcut's event handler map
      typedef void(CDialogPage::*_PFNKsc)();
      typedef map<_TKsc, _PFNKsc> _CKeySCutEHMap;
      _CKeySCutEHMap _mapKeySCutEH;

      // Controls' event handler maps
#ifndef _DOXYGEN_SHOULD_IGNORE_THIS_
#define _IMPL_MAP___ControlEH_(NameC)                             \
      typedef void(CDialogPage::*_PFN_C##NameC##EH)(_CControl*);  \
      typedef map<C##NameC*, _PFN_C##NameC##EH> _C##NameC##EHMap; \
      _C##NameC##EHMap _mapC##NameC##EH;
#define _IMPL_MAP___ControlEH_B_(NameC, NameB)                         \
      typedef void(CDialogPage::*_PFN_C##NameC##EH)(_CControl*, bool); \
      typedef map<C##NameC*, _PFN_C##NameC##EH> _C##NameC##EHMap;      \
      _C##NameC##EHMap _mapC##NameC##EH;
          _IMPL_MAP___ControlEH_  (Button)
          _IMPL_MAP___ControlEH_  (CheckBox)
          _IMPL_MAP___ControlEH_  (RadioButton)
          _IMPL_MAP___ControlEH_  (SpinButton)
          _IMPL_MAP___ControlEH_B_(ListBox,   doubleClick)
          _IMPL_MAP___ControlEH_B_(SLTextBox, accepted)
          _IMPL_MAP___ControlEH_B_(MLTextBox, accepted)
#undef _IMPL_MAP___ControlEH_
#undef _IMPL_MAP___ControlEH_B_
#endif
  };


  //!
  //! \brief A manager class for managing dialog pages
  //!
  //! \see CDialogApplication, for easier development of applications
  //!      that utilize VortexGE dialog classes.
  //!
  //! \api experiment
  class VGE_EXPORT CDialogPageManager : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct the CDialogPageManager instance without setting the main page.
      CDialogPageManager();
      //!
      //! Construct the CDialogPageManager instance and set the main page
      //! instance and name.
      CDialogPageManager(CDialogPage &mainPageInst, const string &mainPageName);
      //!
      //! Standard dtor.
      virtual ~CDialogPageManager() throw();

      // Management functions
      //!
      //! Register the given CDialogPage page instance using the given name.
      //!
      //! If no main page is set (registered) yet, the 1st registered page
      //! will become the main page.
      void RegisterPage(CDialogPage &pageInst, const string &pageName);
      //!
      //! Select (activate) page with the given name.
      void SelectPage(const string &pageName);

      // Information functions
      //!
      //! Returns the name of the main page.
      inline const string &MainPage() const
      { return(_mainPage); }
      //!
      //! Returns the name of the currently active page.
      inline const string &ActivePage() const
      { return(_activePage); }

      // Drawing functions
      //!
      //! Redraw the currently active page.
      virtual void RedrawPage();

      // Standard event handlers
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle key press event.
      //!
      //! \snto
      virtual void HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str);
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle key release event.
      //!
      //! \snto
      virtual void HandleEvent_OnKeyRelease(EKeySymbol keySymbol, const char *str);
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle button press event.
      //!
      //! \snto
      virtual void HandleEvent_OnButtonPress(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY);
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle button release event.
      //!
      //! \snto
      virtual void HandleEvent_OnButtonRelease(EBtnSymbol buttonSymbol, int32_t posX, int32_t posY);
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle mouse pointer move event.
      //!
      //! \snto
      virtual void HandleEvent_OnMouseMove(int32_t posX, int32_t posY, bool buttonState[5]);
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle update frame event.
      //!
      //! \snto
      virtual void HandleEvent_OnUpdateFrame(bool windowIsActive);
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle window activated event.
      //!
      //! \snto
      virtual void HandleEvent_OnWindowActivated();
      //!
      //! Must be called by the class which use this CDialogPageManager class to
      //! handle window deactivated event.
      //!
      //! \snto
      virtual void HandleEvent_OnWindowDeactivated();

    protected:
      typedef map<string, CDialogPage*> _CPageMap;

      string    _mainPage;   // Name of the main page
      string    _activePage; // Name of the currently active page
      _CPageMap _pages;      // Map of pages
  };

} // namespace vge
#endif
