/** \file VideoBase.h */
// ############################################################################################
//  VideoBase.h
//  Definition of various base types used by CVideoEnv
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___VIDEO_BASE_H_
#define _VIDEO_ENVIRONMENT___VIDEO_BASE_H_

// VortexGE header
#include "Color.h"


namespace vge {

  // Foward reference from the video system implementation details
  // ("lib/VideoEnvironment/VID_*/*")
  struct TFont;
  struct TFontSet;
  class  CVideoEnv;


  //!
  //! \brief Typedef used for specifying fill style
  //!
  //! \api beta
  typedef int32_t XFillStyle;

  const XFillStyle FILL_SOLID    = 0; //!< \brief Use solid fill.
  const XFillStyle FILL_GRADIENT = 1; //!< \brief Use gradient fill (disabled in 8-bpp display mode).
  
  const XFillStyle FILL_STIPPLED = 2; //!< \brief Use stippled fill.
                                      //!< \internal Don't ever use it.



  //!
  //! \brief Enumeration used for specifying text alignment mode
  //!
  //! \api pretty stable
  enum ETextAlignment {
    ETA_Left, ETA_Right,ETA_Center,
    ETA_CenterScreen
  };


  //! Size of the stipple pattern
  const uint32_t _STIPPLE_SIZE_ = 32;

  //! Size of the stipple pattern array
  const uint32_t _STIPPLE_ARRAY_SIZE_ = sizeof(uint32_t) * _STIPPLE_SIZE_;

  //!
  //! \brief Structure used for holding drawing style settings
  //!
  //! \api beta
  struct TStyle {
    TFont     *ptFont;                          //!< Font
    TFontSet  *ptFontSet;                       //!< Font set (for international characters)
    XColor24   xOutline;                        //!< Outline color
    XColor24   xLightBorder;                    //!< Light border color
    XColor24   xDarkBorder;                     //!< Dark border color
    XColor24   xFace;                           //!< Face color
    XColor24   xFaceLocked;                     //!< Locked face color
    XColor24   xFaceSelect;                     //!< Selection color
    XColor24   xText;                           //!< Text color
    XColor24   xHotPress;                       //!< Hot-press color
    XColor24   xIconBorder;                     //!< Icon border color
    XColor24   xIconBackground;                 //!< Icon border color
    XColor24   xIconInformation;                //!< Icon-information color
    XColor24   xIconQuestion;                   //!< Icon-question color
    XColor24   xIconWarning;                    //!< Icon-warning color
    XColor24   xIconError;                      //!< Icon-error color
    XFillStyle xFillStyle;                      //!< Fill style
    uint32_t   aun32SPat[_STIPPLE_ARRAY_SIZE_]; //!< Stipple pattern

    TStyle()
    : ptFont          (NULL), ptFontSet    (NULL), xOutline    (0),
      xLightBorder    (0),    xDarkBorder  (0),    xFace       (0), xFaceSelect    (0),
      xText           (0),    xHotPress    (0),    xIconBorder (0), xIconBackground(0),
      xIconInformation(0),    xIconQuestion(0),    xIconWarning(0), xIconError     (0),
      xFillStyle      (0)
    { memset(aun32SPat, 0, sizeof(aun32SPat)); }
  };


  //!
  //! \internal
  //! \brief Triplet uint8 RGB
  //!
  //! \api pretty stable
  struct _attr_packed TUN8RGB {
    uint8_t un8B, un8G, un8R;
  };

  // Internal inline functions implementation related to
  // Sutherland-Hodgman algorithm
  namespace xprv {

    inline void _SHCopy(TUN8RGB &tClip, const TUN8RGB &t1)
    {
      tClip.un8R = t1.un8R;
      tClip.un8G = t1.un8G;
      tClip.un8B = t1.un8B;
    }

    inline void _SHClip(TUN8RGB &tClip, const TUN8RGB &t1, const TUN8RGB &t2, float fT)
    {
      tClip.un8R = uint8_t(int32_t(t1.un8R) + int32_t((int32_t(t2.un8R) - int32_t(t1.un8R)) * fT));
      tClip.un8G = uint8_t(int32_t(t1.un8G) + int32_t((int32_t(t2.un8G) - int32_t(t1.un8G)) * fT));
      tClip.un8B = uint8_t(int32_t(t1.un8B) + int32_t((int32_t(t2.un8B) - int32_t(t1.un8B)) * fT));
    }

  }


  //!
  //! \internal
  //! \brief Structure used for holding render buffer data
  //!
  //! \api pretty stable
  struct _TRenderBuffer {
    uint32_t  un32Width;    // Weight of the buffer
    uint32_t  un32Height;   // Height of the buffer
    uint32_t  un32Size;     // Size of the buffer and the Z buffer in bytes
    uint32_t  un32Size48;   // Size of the buffer and the Z buffer in dword divided by 8
    uint32_t  un32Size4875; // un32Size48 * 3 / 4
    uint32_t  un32Size4850; // un32Size48 * 2 / 4
    uint32_t  un32Size4825; // un32Size48 * 1 / 4
    uint32_t  un32SizeSL;   // Size of the buffer and the Z buffer scanline in bytes
    int32_t  *pn32ZBuffer;  // The Z buffer
    uint8_t  *pun8Buffer;   // The render buffer
    float     fWidth2;      // Weight of the buffer / 2
    float     fHeight2;     // Height of the buffer / 2
    int32_t   n32ZTrans;    // The Z buffer translation factor
    int32_t   n32ZTransInc; // The Z buffer increment factor
    int32_t   n32ZWait;     // The Z buffer clearing delay
  };

#ifdef __USE_DITHERING_FOR_8BPP__
  // Internal inline functions implementation related to
  // 8 bpp display mode dithering
  namespace xprv {

    // Dither RGB555 to RGB332 (to be used as an index to 8 bpp palette table).
    //
    // Thank you to Mike Sharov (msharov[AT]users[DOT]sourceforge[DOT]net)
    // for optimizing this dithering code.
    //
    // Read the file "lib/VideoEnvironment/Dither.txt" for more explanation
    // about this optimization.
    inline uint8_t _DitherRGB88ToRGB332_(uint32_t pixel, uint32_t x, uint32_t y)
    {
      // Multiplexed dithering matrix
      // NOTE: The original (non multiplexed) matrix was generated using
      //       the utility application, "util/minor/bym.cc" by using 31 as
      //       the scaling factor.
      const uint32_t dkernel[8][8] =
        { { 0x808080, 0x989898, 0x868686, 0x9E9E9E, 0x818181, 0x999999, 0x878787, 0x9F9F9F },
          { 0x909090, 0x888888, 0x969696, 0x8E8E8E, 0x919191, 0x898989, 0x979797, 0x8F8F8F },
          { 0x848484, 0x9C9C9C, 0x828282, 0x9A9A9A, 0x858585, 0x9D9D9D, 0x838383, 0x9B9B9B },
          { 0x949494, 0x8C8C8C, 0x929292, 0x8A8A8A, 0x959595, 0x8D8D8D, 0x939393, 0x8B8B8B },
          { 0x818181, 0x999999, 0x878787, 0x9F9F9F, 0x808080, 0x989898, 0x868686, 0x9E9E9E },
          { 0x919191, 0x898989, 0x979797, 0x8F8F8F, 0x909090, 0x888888, 0x969696, 0x8E8E8E },
          { 0x858585, 0x9D9D9D, 0x838383, 0x9B9B9B, 0x848484, 0x9C9C9C, 0x828282, 0x9A9A9A },
          { 0x959595, 0x8D8D8D, 0x939393, 0x8B8B8B, 0x949494, 0x8C8C8C, 0x929292, 0x8A8A8A } };

      const uint32_t b = dkernel[y % 8][x % 8];

      uint32_t s = (pixel & 0xFFFFFF00) | (uint8_t(pixel) >> 1);

      uint32_t r = (s >> 5) & 0x00070707;
      uint32_t d = ~(b - (s & 0x0018181C));

      d = (d & (0x00868682 - r) & 0x00808080) >> 7;
      s = r + d;

      return(uint8_t(s | (s >> 6) | (s >> 11)));
    }

  } // namespace xprv
#endif

} // namespace vge
#endif
