/** \file Auxiliaries.h */
// ############################################################################################
//  Auxiliaries.h
//  Definition & implementation of VortexLua auxiliary classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VORTEX_LUA___AUXILIARIES_H_
#define _VORTEX_LUA___AUXILIARIES_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"


namespace vge {

  // Foward reference from "lib/VortexLua/ExecutableByteCodes.h"
  class CLuaExecutable;

  // Foward reference from "lib/VortexLua/ExecutionSpace.h"
  class CLuaExecutionSpace;


  //!
  //! \brief A class used for storing Lua variable's value
  //!
  //! \api beta
  class VGE_EXPORT CLuaValue {
    public:
      // Set the value explicitly for safety, since the enumeration
      // can be used as index in array access
      enum ELVT { ELVT_Invalid  = 0, // Special ID, should never occured on normal operation
                  ELVT_Any      = 1, // Special ID, used for enumerating variables of any type
                  ELVT_Boolean  = 2,
                  ELVT_Number   = 3,
                  ELVT_String   = 4};

      // Must be the same with the one used by Lua
      typedef double NumberT;

      //!
      //! Construct a new empty (invalid) value.
      CLuaValue();
      //!
      //! Construct a new value of the given type.
      CLuaValue(ELVT elvt);
      //!
      //! Construct a new boolean value.
      CLuaValue(bool value);
      //!
      //! Construct a new number value.
      CLuaValue(NumberT value);
      //!
      //! Construct a new string value.
      CLuaValue(const string &value);
      //!
      //! Copy ctor.
      CLuaValue(const CLuaValue &value);
      //!
      //! Standard dtor.
      ~CLuaValue() throw();

      // Assigments
      //!
      //! Assigment from boolean.
      const CLuaValue &operator = (bool value);
      //!
      //! Assigment from number.
      const CLuaValue &operator = (NumberT value);
      //!
      //! Assigment from string.
      const CLuaValue &operator = (const string &value);
      //!
      //! Assigment from another CLuaValue.
      const CLuaValue &operator = (const CLuaValue &value);

      // Information functions
      //!
      //! Check if the value is an empty (invalid) value.
      inline bool IsInvalid() const
      { return(_type == ELVT_Invalid); }
      //!
      //! Check if the value is a non empty (valid) value.
      inline bool IsValid() const
      { return(_type != ELVT_Invalid); }
      //!
      //! Check if the value is a boolean.
      inline bool IsBoolean() const
      { return(_type == ELVT_Boolean); }
      //!
      //! Check if the value is a number.
      inline bool IsNumber() const
      { return(_type == ELVT_Number); }
      //!
      //! Check if the value is a string.
      inline bool IsString() const
      { return(_type == ELVT_String); }
      //!
      //! Get the type code.
      inline ELVT GetType() const
      { return(_type); }

      // Used for accessing the stored value
      //!
      //! Get the value as a boolean, will throw if the type mismatch.
      bool ToBoolean() const;
      //!
      //! Get the value as a number, will throw if the type mismatch.
      NumberT ToNumber() const;
      //!
      //! Get the value as a string, will throw if the type mismatch.
      const string &ToString() const;

      // I/O functions
      //!
      //! Load contents from a pistream.
      void LoadFromStream(pistream &is);
      //!
      //! Save contents to an ostream.
      void SaveToStream(ostream &os) const;

    private:
      ELVT     _type; // Type of the value
      bool     _vb;   // For storing boolean value
      NumberT  _vn;   // For storing number value
      string  *_vs;   // For storing tring value

      // Friend classes
      friend class CLuaActionCallbackParams;
  };


  //!
  //! \brief A class used for passing parameters (VortexLua's function parameters)
  //!        in callback function
  //!
  //! \api beta
  class VGE_EXPORT CLuaActionCallbackParams : VCLASS_UNCOPYABLE {
    public:
      typedef CLuaValue::NumberT NumberT;

    public:
      inline CLuaActionCallbackParams() {}

      // Used for managing values
      //!
      //! Push a new boolean value.
      void PushBoolean(bool value);
      //!
      //! Push a new number value.
      void PushNumber(NumberT value);
      //!
      //! Push a new string value.
      void PushString(const string &value);
      //!
      //! Remove all stored values.
      inline void Clear()
      { _pars.clear(); }

      // Used for checking stored values
      //!
      //! Return the number of stored values.
      inline uint32_t Count() const
      { return(_pars.size()); }
      //!
      //! Check if the n-th value is a boolean.
      bool IsBoolean(uint32_t index) const;
      //!
      //! Check if the n-th value is a number.
      bool IsNumber(uint32_t index) const;
      //!
      //! Check if the n-th value is a string.
      bool IsString(uint32_t index) const;

      // Used for accessing stored values
      //!
      //! Get the n-th value as a boolean, will throw if the type mismatch or not exist.
      bool ToBoolean(uint32_t index) const;
      //!
      //! Get the n-th value as a number, will throw if the type mismatch or not exist.
      NumberT ToNumber(uint32_t index) const;
      //!
      //! Get the n-th value as a string, will throw if the type mismatch or not exist.
      const string &ToString(uint32_t index) const;

    protected:
      vector<CLuaValue> _pars; // List of parameters
  };


  //!
  //! \brief A class which holds and manages a list of action callbacks
  //!
  //! \api beta
  class VGE_EXPORT CLuaActionList : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Must be returned by any action callback function implementation
      enum EActionCallbackResult { EACR_OK, EACR_ActionNotFound, EACR_NotEnoughParameter };

      //! Typedef of the action callback function
      typedef EActionCallbackResult(*PFNLuaActionCallback)(CLuaExecutionSpace&, const CLuaActionCallbackParams&, CLuaActionCallbackParams&);

    public:
      inline CLuaActionList() {}

      // Registration functions
      //!
      //! Register a new action.
      void Add(const string &name, PFNLuaActionCallback callbackFuncPtr);
      //!
      //! Delete a registered action.
      void Remove(const string &name);
      //!
      //! Delete all registered actions.
      inline void Clear()
      { _list.clear(); }
      //!
      //! Retreive the callback function ptr.
      PFNLuaActionCallback FuncPtrByName(const string &name) const;

      // Invocation functions
      //!
      //! Invoke the action.
      EActionCallbackResult Invoke(const string &actionName, CLuaExecutionSpace &executingNode, const CLuaActionCallbackParams &actionInParams, CLuaActionCallbackParams &actionOutParams);

    protected :
      typedef pair<const string, PFNLuaActionCallback> _TActionCallbackPair;
      typedef list<_TActionCallbackPair>               _CActionCallbackList;

      _CActionCallbackList _list; // List of action callbacks
  };


  //!
  //! \brief A structure used for storing ACL table
  //!
  //! \api beta
  struct VGE_EXPORT TLuaACL {
    set<string> allowRead;      //!< List of nodes which are allowed to read
    set<string> denyRead;       //!< List of nodes which are denied from reading
    set<string> allowWrite;     //!< List of nodes which are allowed to write
    set<string> denyWrite;      //!< List of nodes which are denied from writing
    set<string> allowPenetrate; //!< List of nodes which are allowed to penetrate
    set<string> denyPenetrate;  //!< List of nodes which are denied from penetrating

    // Initialization functions
    //!
    //! Clear the tables.
    void Clear();
    //!
    //! Set the tables with the default per node ACL rules.
    void SetDefaultPerNodeACL();
    //!
    //! Set the tables with the default per variable ACL rules.
    void SetDefaultPerVariableACL();

    // ACL information functions
    //!
    //! Check if the given group is allowed to read.
    bool IsReadableBy(const string &groupName) const;
    //!
    //! Check if the given group is allowed to write.
    bool IsWriteableBy(const string &groupName) const;
    //!
    //! Check if the given group is allowed to penetrate.
    bool IsPenetrateableBy(const string &groupName) const;

    // ACL management functions
    //!
    //! Allow the given node to read.
    void AllowRead(const string &groupName);
    //!
    //! Allow the given node to write.
    void AllowWrite(const string &groupName);
    //!
    //! Allow the given node to penetrate.
    void AllowPenetrate(const string &groupName);
    //!
    //! Deny the given node from reading.
    void DenyRead(const string &groupName);
    //!
    //! Deny the given node from writing.
    void DenyWrite(const string &groupName);
    //!
    //! Deny the given node from penetrating.
    void DenyPenetrate(const string &groupName);

    // I/O functions
    //!
    //! Load the tables' contents from a pistream.
    void LoadFromStream(pistream &is);
    //!
    //! Save the tables' contents to an ostream.
    void SaveToStream(ostream &os) const;
  };


  //!
  //! \brief A structure used for storing execution spaces' variables
  //!
  //! \api beta
  struct VGE_EXPORT TLuaXVars {
    string    name;     //!< The variable's name
    CLuaValue value;    //!< The variable's value
    TLuaACL   aclTable; //!< The variable's per variable ACL rules

    //!
    //! Construct an invalid variable.
    inline TLuaXVars()
    : name(""), value()
    {}
    //!
    //! Construct a variable with the given name and type.
    inline TLuaXVars(const string &varName, CLuaValue::ELVT varType)
    : name(varName), value(varType)
    { aclTable.SetDefaultPerVariableACL(); }

    // ACL information functions
    //!
    //! Check if the given group is allowed to read.
    inline bool IsReadableBy(const string &groupName) const
    { return(aclTable.IsReadableBy(groupName)); }
    //!
    //! Check if the given group is allowed to write.
    inline bool IsWriteableBy(const string &groupName) const
    { return(aclTable.IsWriteableBy(groupName)); }
    //!
    //! Check if the given group is allowed to penetrate.
    inline bool IsPenetrateableBy(const string &groupName) const
    { return(aclTable.IsPenetrateableBy(groupName)); }

    // I/O functions
    //!
    //! Load the variable's data from a pistream.
    void LoadFromStream(pistream &is);
    //!
    //! Save the variable's data to an ostream.
    void SaveToStream(ostream &os) const;
  };

} // namespace vge
#endif
