// ############################################################################################
//  ExecutionSpace.cc
//  Implementation of CLuaExecutionSpace class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "ExecutableByteCodes.h"
#include "ExecutionSpace.h"

// 3rd party headers
extern "C" {
#include "../../3rdParty/lua/lauxlib.h"
#include "../../3rdParty/lua/lualib.h"
}


namespace vge {

//
// Constants used in all execution space state (*.vl.xspace) files
//
static const uint32_t  _XSPACE_FILE_VER_ = 100;
static const char     *_XSPACE_MAGIC_    = "VGE-XSPACE";


// VortexLua functions prototype: version checking
static int _extensionVersion(lua_State *L);
static int _extensionVersionMinimum(lua_State *L);

// Special Lua's constants
static const char LIBRARY_TABLE_NAME[] = "vge";
static const char LIBRARY_CLASS_IVAR[] = "__calling_CLuaExecutionSpace_instance_address__";

// Arrays for storing symbols to be registered to Lua
const CLuaExecutionSpace::_FFuncReg CLuaExecutionSpace::_ffr[] =
      // Version checking
  { { "extensionVersion",        _extensionVersion        },
    { "extensionVersionMinimum", _extensionVersionMinimum },
    { 0, 0 } };

const CLuaExecutionSpace::_MFuncReg CLuaExecutionSpace::_mfr[] =
      // Helper functions
  { { "errorString",           &CLuaExecutionSpace::_errorString           },
    { "getNodeName",           &CLuaExecutionSpace::_getNodeName           },
    { "getGroupName",          &CLuaExecutionSpace::_getGroupName          },
      // Per node query & ACL control
    { "aclNodeReadable",       &CLuaExecutionSpace::_aclNodeReadable       },
    { "aclNodeWriteable",      &CLuaExecutionSpace::_aclNodeWriteable      },
    { "aclNodePenetrateable",  &CLuaExecutionSpace::_aclNodePenetrateable  },
    { "aclNodeAllowRead",      &CLuaExecutionSpace::_aclNodeAllowRead      },
    { "aclNodeAllowWrite",     &CLuaExecutionSpace::_aclNodeAllowWrite     },
    { "aclNodeAllowPenetrate", &CLuaExecutionSpace::_aclNodeAllowPenetrate },
    { "aclNodeDenyRead",       &CLuaExecutionSpace::_aclNodeDenyRead       },
    { "aclNodeDenyWrite",      &CLuaExecutionSpace::_aclNodeDenyWrite      },
    { "aclNodeDenyPenetrate",  &CLuaExecutionSpace::_aclNodeDenyPenetrate  },
    { "queryNodeInGroup",      &CLuaExecutionSpace::_queryNodeInGroup      },
      // Variable creation & deletion
    { "addVar",                &CLuaExecutionSpace::_addVar                },
    { "delVar",                &CLuaExecutionSpace::_delVar                },
    { "hasVar",                &CLuaExecutionSpace::_hasVar                },
      // Per variable ACL control
    { "aclVarAllowRead",       &CLuaExecutionSpace::_aclVarAllowRead       },
    { "aclVarAllowWrite",      &CLuaExecutionSpace::_aclVarAllowWrite      },
    { "aclVarAllowPenetrate",  &CLuaExecutionSpace::_aclVarAllowPenetrate  },
    { "aclVarDenyRead",        &CLuaExecutionSpace::_aclVarDenyRead        },
    { "aclVarDenyWrite",       &CLuaExecutionSpace::_aclVarDenyWrite       },
    { "aclVarDenyPenetrate",   &CLuaExecutionSpace::_aclVarDenyPenetrate   },
      // Variable manipulation
    { "getVar",                &CLuaExecutionSpace::_getVar                },
    { "setVar",                &CLuaExecutionSpace::_setVar                },
      // Invoking action
    { "act",                   &CLuaExecutionSpace::_act                   },
    { 0, 0 } };

const CLuaExecutionSpace::_ConstReg CLuaExecutionSpace::_csr[] =
      // Variable type
  { { "typeAny",                CLuaExecutionSpace::_vt_typeAny                },
    { "typeBoolean",            CLuaExecutionSpace::_vt_typeBoolean            },
    { "typeNumber",             CLuaExecutionSpace::_vt_typeNumber             },
    { "typeString",             CLuaExecutionSpace::_vt_typeString             },
      // Error code
    { "ok",                      CLuaExecutionSpace::_ec_ok                     },
    { "errorInvalidParameter",   CLuaExecutionSpace::_ec_errorInvalidParameter   },
    { "errorNodeNotFound",       CLuaExecutionSpace::_ec_errorNodeNotFound      },
    { "errorActionNotFound",     CLuaExecutionSpace::_ec_errorActionNotFound    },
    { "errorAccessDenied",       CLuaExecutionSpace::_ec_errorAccessDenied      },
    { "errorAlreadyExist",       CLuaExecutionSpace::_ec_errorAlreadyExist      },
    { "errorNotExist",           CLuaExecutionSpace::_ec_errorNotExist          },
    { "errorNotEnoughParameter", CLuaExecutionSpace::_ec_errorNotEnoughParameter },
    { 0, 0 } };

// Special group and node names
const string CLuaExecutionSpace::_sn_NONE = "[NONE]";
const string CLuaExecutionSpace::_sn_ALL  = "[ALL]";
const string CLuaExecutionSpace::_sn_SELF = "[SELF]";
const string CLuaExecutionSpace::_sn_EXE  = "[EXE]";


//
// A class in which executable byte codes will be run
//
CLuaExecutionSpace::CLuaExecutionSpace(const char *nodeName, const char *groupName)
: _mgr(NULL), _nameID(nodeName ? nodeName : ""), _groupID(groupName ? groupName : "")
{
  // Set the default per node ACL
  _aclTable.SetDefaultPerNodeACL();

  // Create a new lua state
  _L = lua_open();
  if(!_L) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new Lua state");

  // Open the Lua standard libraries
  lua_gc(_L, LUA_GCSTOP, 0);
  luaopen_base(_L);
  luaopen_table(_L);
  luaopen_string(_L);
  luaopen_math(_L);
  luaopen_debug(_L);
  lua_gc(_L, LUA_GCRESTART, 0);

  // Create a new table for the VortexLua symbols
  lua_newtable(_L);

  // Register the current class instance address
  char cia[64];
  snprintf(cia, sizeof(cia), "0x%lX", reinterpret_cast<u_long>(this));
  lua_pushstring(_L, LIBRARY_CLASS_IVAR);
  lua_pushstring(_L, cia);
  lua_settable(_L, -3);

  // Register the free functions
  uint32_t i = 0;
  while(_ffr[i].name) {
    lua_pushstring(_L, _ffr[i].name);
    lua_pushcfunction(_L, _ffr[i].ffunc);
    lua_settable(_L, -3);
    i++;
  }

  // Register the member functions
  i = 0;
  while(_mfr[i].name) {
    lua_pushstring(_L, _mfr[i].name);
    lua_pushnumber(_L,  i);
    lua_pushcclosure(_L, _Thunk, 1);
    lua_settable(_L, -3);
    i++;
  }

  // Register the constants
  i = 0;
  while(_csr[i].name) {
    lua_pushstring(_L, _csr[i].name);
    lua_pushnumber(_L, _csr[i].value);
    lua_settable(_L, -3);
    i++;
  }

  // Make the VortexLua table name as global
  lua_setglobal(_L, LIBRARY_TABLE_NAME);
}

CLuaExecutionSpace::~CLuaExecutionSpace() throw()
{
  if(_mgr) _mgr->LXUnregister(*this);
  lua_close(_L);
}

int CLuaExecutionSpace::_Thunk(lua_State *L)
{
  // Get the member function index
  uint32_t i = uint32_t(lua_tonumber(L, lua_upvalueindex(1)));

  // Get the calling class instance address
  lua_getglobal(L, LIBRARY_TABLE_NAME);
  lua_pushstring(L, LIBRARY_CLASS_IVAR);
  lua_gettable(L, -2);
  const char    *cia = lua_tostring(L, -1);
  unsigned long addr = strtoul(cia, 0, 16);
  lua_pop(L, 1);

  // Call the appropriate member function
  return((reinterpret_cast<CLuaExecutionSpace*>(addr)->*_mfr[i].mfunc)(L));
}

CLuaExecutionSpace::NumberT CLuaExecutionSpace::_GetCodeNumber(const string &codeString)
{
  int i = 0;
  while(_csr[i].name) {
    if(_csr[i].name == codeString) return(_csr[i].value);
    i++;
  }
  return(-1.0);
}

string CLuaExecutionSpace::_GetCodeString(const int32_t codeNumber)
{
  int i = 0;
  while(_csr[i].name) {
    if(_csr[i].value == codeNumber) return(string("vge.") + _csr[i].name);
    i++;
  }
  return("<unknown error code>");
}

bool CLuaExecutionSpace::_ValidVarType(int32_t _vt_type, bool allowAny)
{
  if(allowAny && (_vt_type == _vt_typeAny)) return(true);
  if((_vt_type == _vt_typeBoolean) ||
     (_vt_type == _vt_typeNumber ) ||
     (_vt_type == _vt_typeString )) return(true);
  return(false);
}

bool CLuaExecutionSpace::_VarTypesCompatible(int32_t _vt_type_ref, int32_t _vt_type_check)
{
  if(_vt_type_ref == _vt_typeInvalid)
    return(_vt_type_check == _vt_typeInvalid);
  else if(_vt_type_ref == _vt_typeAny)
    return(_vt_type_check != _vt_typeInvalid);
  else
    return(_vt_type_ref == _vt_type_check);
}

CLuaValue::ELVT CLuaExecutionSpace::_ToELVT(int32_t _vt_type)
{
  static const CLuaValue::ELVT _et[5] = { CLuaValue::ELVT_Invalid,
                                          CLuaValue::ELVT_Any,
                                          CLuaValue::ELVT_Boolean,
                                          CLuaValue::ELVT_Number,
                                          CLuaValue::ELVT_String };
  return(_et[_vt_type - _vt_typeInvalid]);
}

int32_t CLuaExecutionSpace::_ToLuaVarType(CLuaValue::ELVT elvt)
{
  static const int32_t _vt[5] = { _vt_typeInvalid,
                                  _vt_typeAny,
                                  _vt_typeBoolean,
                                  _vt_typeNumber,
                                  _vt_typeString };
  return(_vt[elvt]);
}

bool CLuaExecutionSpace::_HasVar(const string &varName, size_t *varIndexRet , CLuaValue::ELVT *varTypeRet) const
{
  for(size_t i  = 0; i < _varList.size(); ++i) {
    if(_varList[i].name == varName) {
      const CLuaValue::ELVT elvt = _varList[i].value.GetType();

      vassert(elvt != CLuaValue::ELVT_Invalid,
              "VortexLua variable \"%s\" has an invalid type; this should never happen in normal script execution",
              _varList[i].name.c_str());
    
      if(varTypeRet)  *varTypeRet  = elvt;
      if(varIndexRet) *varIndexRet = i;
      return(true);
    }
  }
  return(false);
}

void CLuaExecutionSpace::_PushVar(lua_State *L, size_t index)
{
  if(_varList[index].value.IsBoolean())
    lua_pushboolean(L, _varList[index].value.ToBoolean());
  else if(_varList[index].value.IsNumber())
    lua_pushnumber(L, _varList[index].value.ToNumber());
  else if(_varList[index].value.IsString())
    lua_pushstring(L, _varList[index].value.ToString().c_str());
  else
    vassert(false,
            "VortexLua variable \"%s\" has a non-allowed type; this should never happen in normal script execution",
            _varList[index].name.c_str());
}

void CLuaExecutionSpace::Execute(const CLuaExecutable &exe, CLuaExecutionSpace *pair)
{
  // Checking
  if(!_mgr) VError(VERR_ExecuteScript, _NFL_, "This CLuaExecutionSpace is registered to nothing\nScript execution is prohibited");
  if(!exe._mgr) VError(VERR_ExecuteScript, _NFL_, "The given CLuaExecutable is registered to nothing\nScript execution is prohibited");
  if(!_mgr->LERegistered(exe)) VError(VERR_ExecuteScript, _NFL_, "The given CLuaExecutable is not registered in the same CLuaManager as this CLuaExecutionSpace\nScript execution is prohibited");

  // Set the paired node
  _pair = pair;

  // Load the byte codes
  if(luaL_loadbuffer(_L, exe.ByteCodes(), exe.ByteCodesSize(), exe.GetName().c_str())) {
    const string msg(lua_tostring(_L, -1));
    lua_gc(_L, LUA_GCCOLLECT, 0);
    lua_pop(_L, 1);
    VError(VERR_ExecuteScript, _NFL_, "%s", msg.c_str());
  }
  // Execute it
  if(lua_pcall(_L, 0, 0, 0)) {
    const string msg(lua_tostring(_L, -1));
    lua_gc(_L, LUA_GCCOLLECT, 0);
    lua_pop(_L, 1);
    VError(VERR_ExecuteScript, _NFL_, "%s", msg.c_str());
  }

  lua_gc(_L, LUA_GCCOLLECT, 0);
  lua_pop(_L, 1);
}

TLuaXVars *CLuaExecutionSpace::GetVarByName(const string &varName, size_t *varIndexRet)
{
  for(vector<string>::size_type i  = 0; i < _varList.size(); ++i)
    if(_varList[i].name == varName) {
      if(varIndexRet) *varIndexRet = i;
      return(&_varList[i]);
    }
  return(NULL);
}

const TLuaXVars *CLuaExecutionSpace::GetVarByName(const string &varName, size_t *varIndexRet) const
{
  for(vector<string>::size_type i  = 0; i < _varList.size(); ++i)
    if(_varList[i].name == varName) {
      if(varIndexRet) *varIndexRet = i;
      return(&_varList[i]);
    }
  return(NULL);
}

bool CLuaExecutionSpace::AddVar(const string &varName, CLuaValue::ELVT varType)
{
  // Check if variable with the same name already exist
  if(_HasVar(varName)) return(false);

  // Construct the variable
  _varList.push_back(TLuaXVars(varName, varType));
  return(true);
}

bool CLuaExecutionSpace::DelVar(const string &varName)
{
  // Check if variable is not exist
  size_t vIndex;
  if(!_HasVar(varName, &vIndex)) return(false); 

  // Delete the variable
  _varList.erase(_varList.begin() + vIndex);
  return(true);
}

void CLuaExecutionSpace::DelVar(size_t varIndex)
{
  _varList.erase(_varList.begin() + varIndex);
}

const string CLuaExecutionSpace::_GetNNameForQueryingNodeOrGroupName(lua_State *L, int32_t nArg)
{
  luaL_checktype(L, nArg, LUA_TSTRING);

  const string &name(lua_tostring(L, nArg));

  _if_unlikely((name == _sn_NONE) || (name == _sn_ALL))
    luaL_argerror(L, nArg, "[NONE] and [ALL] can't be used when querying for group/node name");

  return(name);
}

const string CLuaExecutionSpace::_GetGNameForQueryingNodeInGroup(lua_State *L, int32_t nArg)
{
  luaL_checktype(L, nArg, LUA_TSTRING);

  const string &name(lua_tostring(L, nArg));

  _if_unlikely(name == _sn_NONE)
    luaL_argerror(L, 1, "[NONE] can't be used when querying nodes in group");

  return(name);
}

const string CLuaExecutionSpace::_GetNNameForQueryingACL(lua_State *L, int32_t nArg)
{
  luaL_checktype(L, nArg, LUA_TSTRING);

  const string &name(lua_tostring(L, nArg));

  _if_unlikely((name == _sn_NONE) || (name == _sn_ALL))
    luaL_argerror(L, nArg, "[NONE] and [ALL] can't be used when querying ACL");

  return(name);
}

const string CLuaExecutionSpace::_GetGNameForSettingACL(lua_State *L, int32_t nArg)
{
  luaL_checktype(L, nArg, LUA_TSTRING);

  const string &name(lua_tostring(L, nArg));

  _if_unlikely((name == _sn_SELF) || (name == _sn_EXE))
    luaL_argerror(L, 1, "[SELF] and [EXE] can't be used when setting ACL");

  return(name);
}

const string CLuaExecutionSpace::_GetNNameForVarAccess(lua_State *L, int32_t nArg)
{
  luaL_checktype(L, nArg, LUA_TSTRING);

  const string &name(lua_tostring(L, nArg));

  _if_unlikely((name == _sn_NONE) || (name == _sn_ALL))
    luaL_argerror(L, nArg, "[NONE] and [ALL] can't be used when accessing variable");

  return(name);
}

bool CLuaExecutionSpace::HasVar(const string &varName, CLuaValue::ELVT varType) const
{
  // Check if the variable is exist
  CLuaValue::ELVT vType;
  if(!_HasVar(varName, NULL, &vType)) return(false);

  // Check its type
  return(_VarTypesCompatible(_ToLuaVarType(varType),  _ToLuaVarType(vType)));
}

void CLuaExecutionSpace::LoadFromStream(pistream &is)
{
  // Read the magic numbers
  char mnb[64];
  memset(mnb, 0, sizeof(mnb));
  is.read(mnb, strlen(_XSPACE_MAGIC_));
  if(strcmp(mnb, _XSPACE_MAGIC_))
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data");

  // Read the version number
  uint32_t ver;
  is.read(reinterpret_cast<char*>(&ver), sizeof(ver));
  if(ver != _XSPACE_FILE_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "Invalid version number");

  // Read the nameID
  uint32_t nz;
  is.read(reinterpret_cast<char*>(&nz), sizeof(nz));
  if(nz) {
    string tmp;
    tmp.resize(nz);
    is.read(&tmp[0], nz);
    if(tmp != _nameID) {
      VWarning(_NFL_, "The state data may not belongs to CLuaExecutionSpace instance \"%s\"",
               _nameID.c_str());
    }
  }

  // Read the groupID
  uint32_t gz;
  is.read(reinterpret_cast<char*>(&gz), sizeof(gz));
  if(gz) {
    string tmp;
    tmp.resize(gz);
    is.read(&tmp[0], gz);
    if(tmp != _groupID) {
      VWarning(_NFL_, "The state file has \"%s\" as the group ID but the\nCLuaExecutionSpace specify \"%s\" as the group ID",
               tmp.c_str(), _groupID.c_str());
    }
  }

  // Load the per node ACL table
  _aclTable.LoadFromStream(is);

  // Load the variables
  uint32_t vz;
  is.read(reinterpret_cast<char*>(&vz), sizeof(vz));
  if(vz) {
    _varList.resize(vz);
    for(size_t i = 0; i < vz; ++i) {
      _varList[i].LoadFromStream(is);
    }
  }
  else
    _varList.clear();
}

void CLuaExecutionSpace::SaveToStream(ostream &os) const
{
  // Write the magic numbers
  os.write(_XSPACE_MAGIC_, strlen(_XSPACE_MAGIC_));

  // Write the version number
  uint32_t ver = _XSPACE_FILE_VER_;
  os.write(reinterpret_cast<char*>(&ver), sizeof(ver));

  // Save the nameID
  uint32_t nz = _nameID.length();
  os.write(reinterpret_cast<char*>(&nz), sizeof(nz));
  if(nz) os.write(_nameID.data(), nz);

  // Save the groupID
  uint32_t gz = _groupID.length();
  os.write(reinterpret_cast<char*>(&gz), sizeof(gz));
  if(gz) os.write(_groupID.data(), gz);

  // Save the per node ACL table
  _aclTable.SaveToStream(os);

  // Save all the variables
  uint32_t vz = _varList.size();
  os.write(reinterpret_cast<char*>(&vz), sizeof(vz));
  for(size_t i = 0; i < vz; ++i) _varList[i].SaveToStream(os);
}


//
// VortexLua functions prototype: helper functions
//
int CLuaExecutionSpace::_errorString(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TNUMBER);

  // Extract the error string of the given code
  lua_pushstring(L, _GetCodeString(int32_t(lua_tonumber(L, 1))).c_str());
  return(1);
}

int CLuaExecutionSpace::_getNodeName(lua_State *L)
{
  // Get the function call's parameters
  const string &nName = _GetNNameForQueryingNodeOrGroupName(L, 1);

  // Get the node name
  if(nName == _sn_SELF)
    lua_pushstring(L, _nameID.c_str());
  else if(nName == _sn_EXE) {
    if(!_pair) lua_pushnil(L);
    else       lua_pushstring(L, _pair->_nameID.c_str());
  }
  else
    lua_pushnil(L);
  return(1);
}

int CLuaExecutionSpace::_getGroupName(lua_State *L)
{
  // Get the function call's parameters
  const string &nName = _GetNNameForQueryingNodeOrGroupName(L, 1);

  // Get the node name
  if(nName == _sn_SELF)
    lua_pushstring(L, _groupID.c_str());
  else if(nName == _sn_EXE) {
    if(!_pair) lua_pushnil(L);
    else       lua_pushstring(L, _pair->_groupID.c_str());
  }
  else
    lua_pushnil(L);
  return(1);
}


//
// VortexLua functions: per node query & ACL control
//
int CLuaExecutionSpace::_aclNodeReadable(lua_State *L)
{
  // Get the function call's parameters
  const string &nName = _GetNNameForQueryingACL(L, 1);

  // Check the ACL
  if(nName == _sn_SELF)
    lua_pushnumber(L, _ec_ok);
  else if(nName == _sn_EXE) {
    if(!_pair)
      lua_pushnumber(L, _ec_errorNodeNotFound);
    else {
      if(_pair->_IsReadableBy(_groupID)) lua_pushnumber(L, _ec_ok);
      else                               lua_pushnumber(L, _ec_errorAccessDenied);
    }
  }
  else {
    // Find the node
    CLuaExecutionSpace *xn = _mgr->LXFindByName(nName);
    if(!xn)
      lua_pushnumber(L, _ec_errorNodeNotFound);
    else {
      if(xn->_IsReadableBy(_groupID)) lua_pushnumber(L, _ec_ok);
      else                            lua_pushnumber(L, _ec_errorAccessDenied);
    }
  }
  return(1);
}

int CLuaExecutionSpace::_aclNodeWriteable(lua_State *L)
{
  // Get the function call's parameters
  const string &nName = _GetNNameForQueryingACL(L, 1);

  // Check the ACL
  if(nName == _sn_SELF)
    lua_pushnumber(L, _ec_ok);
  else if(nName == _sn_EXE) {
    if(!_pair)
      lua_pushnumber(L, _ec_errorNodeNotFound);
    else {
      if(_pair->_IsWriteableBy(_groupID)) lua_pushnumber(L, _ec_ok);
      else                                lua_pushnumber(L, _ec_errorAccessDenied);
    }
  }
  else {
    // Find the node
    CLuaExecutionSpace *xn = _mgr->LXFindByName(nName);
    if(!xn)
      lua_pushnumber(L, _ec_errorNodeNotFound);
    else {
      if(xn->_IsWriteableBy(_groupID)) lua_pushnumber(L, _ec_ok);
      else                             lua_pushnumber(L, _ec_errorAccessDenied);
    }
  }
  return(1);
}

int CLuaExecutionSpace::_aclNodePenetrateable(lua_State *L)
{
  // Get the function call's parameters
  const string &nName = _GetNNameForQueryingACL(L, 1);

  // Check the ACL
  if(nName == _sn_SELF)
    lua_pushnumber(L, _ec_ok);
  else if(nName == _sn_EXE) {
    if(!_pair)
      lua_pushnumber(L, _ec_errorNodeNotFound);
    else {
      if(_pair->_IsPenetrateableBy(_groupID)) lua_pushnumber(L, _ec_ok);
      else                                    lua_pushnumber(L, _ec_errorAccessDenied);
    }
  }
  else {
    // Find the node
    CLuaExecutionSpace *xn = _mgr->LXFindByName(nName);
    if(!xn)
      lua_pushnumber(L, _ec_errorNodeNotFound);
    else {
      if(xn->_IsPenetrateableBy(_groupID)) lua_pushnumber(L, _ec_ok);
      else                                 lua_pushnumber(L, _ec_errorAccessDenied);
    }
  }
  return(1);
}

int CLuaExecutionSpace::_aclNodeAllowRead(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForSettingACL(L, 1);

  // Set the ACL
  _aclTable.AllowRead(gName.c_str());
  return(0);
}

int CLuaExecutionSpace::_aclNodeAllowWrite(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForSettingACL(L, 1);

  // Set the ACL
  _aclTable.AllowWrite(gName.c_str());
  return(0);
}

int CLuaExecutionSpace::_aclNodeAllowPenetrate(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForSettingACL(L, 1);

  // Set the ACL
  _aclTable.AllowPenetrate(gName.c_str());
  return(0);
}

int CLuaExecutionSpace::_aclNodeDenyRead(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForSettingACL(L, 1);

  // Set the ACL
  _aclTable.DenyRead(gName.c_str());
  return(0);
}

int CLuaExecutionSpace::_aclNodeDenyWrite(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForSettingACL(L, 1);

  // Set the ACL
  _aclTable.DenyWrite(gName.c_str());
  return(0);
}

int CLuaExecutionSpace::_aclNodeDenyPenetrate(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForSettingACL(L, 1);

  // Set the ACL
  _aclTable.DenyPenetrate(gName.c_str());
  return(0);
}

int CLuaExecutionSpace::_queryNodeInGroup(lua_State *L)
{
  // Get the function call's parameters
  const string &gName = _GetGNameForQueryingNodeInGroup(L, 1);

  // For storing queried node
  vector<string> nds;

  // Query for [SELF]'s group
  if(gName == _sn_SELF) {
    if(!_groupID.empty())
      _mgr->QueryNodesInGroup(nds, _groupID);
    else
      nds.push_back(_nameID);
  }
  // Query for [EXE]'s group
  else if(gName == _sn_EXE) {
    if(_pair) {
      if(!_pair->_groupID.empty())
        _mgr->QueryNodesInGroup(nds, _pair->_groupID);
      else
        nds.push_back(_pair->_nameID);
    }
  }
  // Query for other groups
  else {
    _mgr->QueryNodesInGroup(nds, gName);
  }

  // Convert it to table
  if(nds.empty())
    lua_pushnil(L);
  else {
    lua_newtable(_L);
    // Push the queried nodes' names
    for(size_t i = 0; i < nds.size(); ++i) {
      lua_pushnumber(L, i);
      lua_pushstring(L, nds[i].c_str());
      lua_settable(L, -3);
    }
    // Push terminator
    lua_pushnumber(L, nds.size());
    lua_pushnil(L);
    lua_settable(L, -3);
  }
  return(1);
}


//
// VortexLua functions: variable creation & deletion
//
int CLuaExecutionSpace::_addVar(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 2, LUA_TSTRING);
  luaL_checktype(L, 3, LUA_TNUMBER);

  // Get the function call's parameters
  const string &nName = _GetNNameForVarAccess(L, 1);
  const char   *vName = lua_tostring(L, 2);
  const int32_t vType = int32_t(lua_tonumber(L, 3));

  // Check the variable type
  if(!_ValidVarType(vType, false)) {
    lua_pushnumber(L, _ec_errorInvalidParameter);
    return(1);
  }

  // Create variable in the current node
  if(nName == _sn_SELF) {
    // Creating variable in SELF is always granted regardless the
    // per node ACL sets in SELF
    if(!AddVar(vName, _ToELVT(vType))) {
      lua_pushnumber(L, _ec_errorAlreadyExist);
      return(1);
    }
  }
  // Create variable in other node
  else {
    CLuaExecutionSpace *xn = NULL;
    // Find the node
    if(nName == _sn_EXE) xn = _pair;
    else                 xn = _mgr->LXFindByName(nName);
    // Check if the node is not found
    if(!xn) {
      lua_pushnumber(L, _ec_errorNodeNotFound);
      return(1);
    }
    // Check the per node ACL
    if(!xn->_IsPenetrateableBy(_groupID)) {
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(1);
    }
    // Construct the variable
    if(!xn->AddVar(vName, _ToELVT(vType))) {
      lua_pushnumber(L, _ec_errorAlreadyExist);
      return(1);
    }
  }

  // Done !
  lua_pushnumber(L, _ec_ok);
  return(1);
}

int CLuaExecutionSpace::_delVar(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 2, LUA_TSTRING);

  // Get the function call's parameters
  const string &nName = _GetNNameForVarAccess(L, 1);
  const char   *vName = lua_tostring(L, 2);

  // Delete variable in the current node
  if(nName == _sn_SELF) {
    // Find the variable
    size_t idx = 0;
    if(!GetVarByName(vName, &idx)) {
      lua_pushnumber(L, _ec_errorNotExist);
      return(1);
    }
    // Deleting variable in SELF is always granted regardless the per node
    // and per variable ACLs set in SELF and in the SELF's variable
    DelVar(idx);
  }
  // Delete variable in other node
  else {
    CLuaExecutionSpace *xn = NULL;
    // Find the node
    if(nName == _sn_EXE) xn = _pair;
    else                 xn = _mgr->LXFindByName(nName);
    // Check if the node is not found
    if(!xn) {
      lua_pushnumber(L, _ec_errorNodeNotFound);
      return(1);
    }
    // Check the per node ACL
    if(!xn->_IsPenetrateableBy(_groupID)) {
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(1);
    }
    // Find the variable
    size_t idx = 0;
    if(!xn->GetVarByName(vName, &idx)) {
      lua_pushnumber(L, _ec_errorNotExist);
      return(1);
    }
    // Check the per variable ACL
    if(!xn->_varList[idx].IsPenetrateableBy(_groupID)) {
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(1);
    }
    // Delete the variable
    xn->DelVar(idx);
  }

  // Done !
  lua_pushnumber(L, _ec_ok);
  return(1);
}

int CLuaExecutionSpace::_hasVar(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 2, LUA_TSTRING);

  // Get the function call's parameters
  const string &nName = _GetNNameForVarAccess(L, 1);
  const char   *vName = lua_tostring(L, 2);

  // If the 3rd parameter (varType) is present, the stack top will be 4 (not 3)
  // because of the single upvalue given when registering using lua_pushcclosure()
  int32_t vtRef = _vt_typeAny;
  if(lua_gettop(L) > 3) {
    luaL_checktype(L, 3, LUA_TNUMBER);
    vtRef = int32_t(lua_tonumber(L, 3));
    if(vtRef!=_vt_typeAny     && 
       vtRef!=_vt_typeBoolean &&
       vtRef!=_vt_typeNumber  &&
       vtRef!=_vt_typeString)
      luaL_argerror(L, 3, "valid vge.type* expected, got invalid code number");
  }

  // Check variable in the current node
  bool ret = false;
  if(nName == _sn_SELF) {
    ret = HasVar(vName, _ToELVT(vtRef));
  }
  // Check variable in other node
  else {
    CLuaExecutionSpace *xn = NULL;
    // Find the node
    if(nName == _sn_EXE) xn = _pair;
    else                 xn = _mgr->LXFindByName(nName);
    if(xn) {
      // Check if variable is exist
      ret = xn->HasVar(vName, _ToELVT(vtRef));
    }
  }

  // Done !
  lua_pushboolean(L, ret);
  return(1);
}


//
// VortexLua functions: per variable ACL control
//
int CLuaExecutionSpace::_aclVarAllowRead(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the function call's parameters
  const char   *vName = lua_tostring(L, 1);
  const string &gName = _GetGNameForSettingACL(L, 2);

  // Set the ACL
  size_t idx = 0;
  if(!GetVarByName(vName, &idx))
    lua_pushnumber(L, _ec_errorNotExist);
  else {
    _varList[idx].aclTable.AllowRead(gName);
    lua_pushnumber(L, _ec_ok);
  }
  return(1);
}

int CLuaExecutionSpace::_aclVarAllowWrite(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the function call's parameters
  const char   *vName = lua_tostring(L, 1);
  const string &gName = _GetGNameForSettingACL(L, 2);

  // Set the ACL
  size_t idx = 0;
  if(!GetVarByName(vName, &idx))
    lua_pushnumber(L, _ec_errorNotExist);
  else {
    _varList[idx].aclTable.AllowWrite(gName);
    lua_pushnumber(L, _ec_ok);
  }
  return(1);
}

int CLuaExecutionSpace::_aclVarAllowPenetrate(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the function call's parameters
  const char   *vName = lua_tostring(L, 1);
  const string &gName = _GetGNameForSettingACL(L, 2);

  // Set the ACL
  size_t idx = 0;
  if(!GetVarByName(vName, &idx))
    lua_pushnumber(L, _ec_errorNotExist);
  else {
    _varList[idx].aclTable.AllowPenetrate(gName);
    lua_pushnumber(L, _ec_ok);
  }
  return(1);
}

int CLuaExecutionSpace::_aclVarDenyRead(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the function call's parameters
  const char   *vName = lua_tostring(L, 1);
  const string &gName = _GetGNameForSettingACL(L, 2);

  // Set the ACL
  size_t idx = 0;
  if(!GetVarByName(vName, &idx))
    lua_pushnumber(L, _ec_errorNotExist);
  else {
    _varList[idx].aclTable.DenyRead(gName);
    lua_pushnumber(L, _ec_ok);
  }
  return(1);
}

int CLuaExecutionSpace::_aclVarDenyWrite(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the function call's parameters
  const char   *vName = lua_tostring(L, 1);
  const string &gName = _GetGNameForSettingACL(L, 2);

  // Set the ACL
  size_t idx = 0;
  if(!GetVarByName(vName, &idx))
    lua_pushnumber(L, _ec_errorNotExist);
  else {
    _varList[idx].aclTable.DenyWrite(gName);
    lua_pushnumber(L, _ec_ok);
  }
  return(1);
}

int CLuaExecutionSpace::_aclVarDenyPenetrate(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the function call's parameters
  const char   *vName = lua_tostring(L, 1);
  const string &gName = _GetGNameForSettingACL(L, 2);

  // Set the ACL
  size_t idx = 0;
  if(!GetVarByName(vName, &idx))
    lua_pushnumber(L, _ec_errorNotExist);
  else {
    _varList[idx].aclTable.DenyPenetrate(gName);
    lua_pushnumber(L, _ec_ok);
  }
  return(1);
}


//
// VortexLua functions: variable manipulation
//
int CLuaExecutionSpace::_getVar(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 2, LUA_TSTRING);

  // Get the function call's parameters
  const string &nName = _GetNNameForVarAccess(L, 1);
  const char   *vName = lua_tostring(L, 2);

  // If the 3rd parameter (varType) is present, the stack top will be 4 (not 3)
  // because of the single upvalue given when registering using lua_pushcclosure()
  int32_t vRef = _vt_typeAny;
  if(lua_gettop(L) > 3) {
    luaL_checktype(L, 3, LUA_TNUMBER);
    vRef = int32_t(lua_tonumber(L, 3));
    if(vRef!=_vt_typeAny     && 
       vRef!=_vt_typeBoolean &&
       vRef!=_vt_typeNumber  &&
       vRef!=_vt_typeString)
      luaL_argerror(L, 3, "valid vge.type* expected, got invalid code number");
  }

  // Get variable in the current node
  if(nName == _sn_SELF) {
    // Find the variable
    size_t idx = 0;
    if(!GetVarByName(vName, &idx)) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorNotExist);
      return(2);
    }
    // Check the variable types
    if(!_VarTypesCompatible(vRef, _ToLuaVarType(_varList[idx].value.GetType()))) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorNotExist);
      return(2);
    }
    // Getting variable in SELF is always granted regardless the per node
    // and per variable ACLs set in SELF and in the SELF's variable
    _PushVar(L, idx);
  }
  // Get variable in other node
  else {
    CLuaExecutionSpace *xn = NULL;
    // Find the node
    if(nName == _sn_EXE) xn = _pair;
    else                 xn = _mgr->LXFindByName(nName);
    // Check if the node is not found
    if(!xn) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorNodeNotFound);
      return(2);
    }
    // Check the per node ACL
    if(!xn->_IsReadableBy(_groupID)) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(2);
    }
    // Find the variable
    size_t idx = 0;
    if(!xn->GetVarByName(vName, &idx)) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorNotExist);
      return(2);
    }
    // Check the per variable ACL
    if(!xn->_varList[idx].IsReadableBy(_groupID)) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(2);
    }
    // Check the variable types
    if(!_VarTypesCompatible(vRef, _ToLuaVarType(xn->_varList[idx].value.GetType()))) {
      lua_pushnil(L);
      lua_pushnumber(L, _ec_errorNotExist);
      return(2);
    }
    // Get the variable
    xn->_PushVar(L, idx);
  }

  // Done !
  lua_pushnumber(L, _ec_ok);
  return(2);
}

int CLuaExecutionSpace::_setVar(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 2, LUA_TSTRING);

  // Get the function call's parameters
  const string &nName = _GetNNameForVarAccess(L, 1);
  const char   *vName = lua_tostring(L, 2);
  CLuaValue     vVal;
  switch(lua_type(L, 3)) {
    case LUA_TBOOLEAN: vVal =              bool(lua_toboolean(L, 3)); break;
    case LUA_TNUMBER : vVal =    VLUA_TO_NUMBER(lua_tonumber (L, 3)); break;
    case LUA_TSTRING : vVal =            string(lua_tostring (L, 3)); break;
    default          : luaL_typerror(L, 3, "boolean/number/string expected"); return(0);
  };

  // Get variable in the current node
  if(nName == _sn_SELF) {
    // Find the variable
    size_t idx = 0;
    if(!GetVarByName(vName, &idx)) {
      lua_pushnumber(L, _ec_errorNotExist);
      return(1);
    }
    // Check the value types
    if(!_VarTypesCompatible(_ToLuaVarType(_varList[idx].value.GetType()), _ToLuaVarType(vVal.GetType()))) {
      lua_pushnumber(L, _ec_errorInvalidParameter);
      return(1);
    }
    // Setting variable in SELF is always granted regardless the per node
    // and per variable ACLs set in SELF and in the SELF's variable
    _varList[idx].value = vVal;
  }
  // Get variable in other node
  else {
    CLuaExecutionSpace *xn = NULL;
    // Find the node
    if(nName == _sn_EXE) xn = _pair;
    else                 xn = _mgr->LXFindByName(nName);
    // Check if the node is not found
    if(!xn) {
      lua_pushnumber(L, _ec_errorNodeNotFound);
      return(1);
    }
    // Check the per node ACL
    if(!xn->_IsWriteableBy(_groupID)) {
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(2);
    }
    // Find the variable
    size_t idx = 0;
    if(!xn->GetVarByName(vName, &idx)) {
      lua_pushnumber(L, _ec_errorNotExist);
      return(1);
    }
    // Check the per variable ACL
    if(!xn->_varList[idx].IsWriteableBy(_groupID)) {
      lua_pushnumber(L, _ec_errorAccessDenied);
      return(1);
    }
    if(!_VarTypesCompatible(_ToLuaVarType(xn->_varList[idx].value.GetType()), _ToLuaVarType(vVal.GetType()))) {
      lua_pushnumber(L, _ec_errorInvalidParameter);
      return(1);
    }
    // Setting variable in SELF is always granted regardless the per node
    // and per variable ACLs set in SELF and in the SELF's variable
    xn->_varList[idx].value = vVal;
  }

  // Done !
  lua_pushnumber(L, _ec_ok);
  return(1);
}


//
// VortexLua functions: invoking action
//
int CLuaExecutionSpace::_act(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TSTRING);

  // Get the action name
  const char *name = lua_tostring(L, 1);

  // Get the input parameters from the script
  CLuaActionCallbackParams inParams;
  for(ssize_t idx = 2; idx < lua_gettop(L); idx++) {
    int32_t type = lua_type(L, idx);
    switch(type) {
      case LUA_TBOOLEAN: inParams.PushBoolean(lua_toboolean(L, idx)); break;
      case LUA_TNUMBER : inParams.PushNumber (lua_tonumber (L, idx)); break;
      case LUA_TSTRING : inParams.PushString (lua_tostring (L, idx)); break;
      default          : luaL_typerror(L, idx, "boolean/number/string expected"); return(0);
    };
  }

  // Invoke the action
  CLuaActionCallbackParams              outParams;
  CLuaActionList::EActionCallbackResult acr = _alist.Invoke(name, *this, inParams, outParams);

  // Push the result code
  switch(acr) {
    case CLuaActionList::EACR_OK                : lua_pushnumber(L, _ec_ok);                      break;
    case CLuaActionList::EACR_ActionNotFound    : lua_pushnumber(L, _ec_errorActionNotFound);     break;
    case CLuaActionList::EACR_NotEnoughParameter: lua_pushnumber(L, _ec_errorNotEnoughParameter); break;
  }

  // Push the parameter count
  lua_pushnumber(L, outParams.Count());

  // Extract the output parameters and send them back to the script
  for(size_t idx = 0; idx < outParams.Count(); idx++) {
         if(outParams.IsBoolean(idx)) lua_pushboolean(L, outParams.ToBoolean(idx));
    else if(outParams.IsNumber (idx)) lua_pushnumber (L, outParams.ToNumber(idx));
    else if(outParams.IsString (idx)) lua_pushstring (L, outParams.ToString(idx).c_str());
    else VError(VERR_InvalidParameter, _NFL_, "The user action callback function returns (an) invalid parameter(s)");
  }
  return(outParams.Count() + 2);
}


//
// VortexLua functions: version checking
//
static int _extensionVersion(lua_State *L)
{
  // Return the current version number
  lua_pushnumber(L, __VORTEXLUA_MAJOR__);
  lua_pushnumber(L, __VORTEXLUA_MINOR__);
  lua_pushnumber(L, __VORTEXLUA_REVISION__);
  return(3);
}

static int _extensionVersionMinimum(lua_State *L)
{
  // Make sure that the parameters have the correct types
  luaL_checktype(L, 1, LUA_TNUMBER);
  luaL_checktype(L, 2, LUA_TNUMBER);
  luaL_checktype(L, 3, LUA_TNUMBER);

  // Check if the requested minimum version is supported
  if(__VORTEXLUA_MAJOR__ < lua_tonumber(L, 1))
    lua_pushboolean(L, false);
  else if(__VORTEXLUA_MAJOR__ > lua_tonumber(L, 1))
    lua_pushboolean(L, true);
  else {
    if(__VORTEXLUA_MINOR__ < lua_tonumber(L, 2))
      lua_pushboolean(L, false);
    else if(__VORTEXLUA_MINOR__ > lua_tonumber(L, 2))
      lua_pushboolean(L, true);
    else {
      if(__VORTEXLUA_REVISION__ < lua_tonumber(L, 3))
        lua_pushboolean(L, false);
      else
        lua_pushboolean(L, true);
    }
  }
  return(1);
}

} // namespace vge
