// convert.engine.cc
// Rename standalone symbol names in the input file based on the name table
// Copyright (C) 2004-2006 Aloysius Indrayanto
//
// This small utility is licensed under the GNU LGPL
//
// Usage: ./convert.engine in-file [> out-file]
//
#include <iostream>
#include <fstream>
#include <vector>
using namespace std;


// Used when renaming symbols
struct RenameTable {
  string search;
  string replace;

  RenameTable(const string &s, const string &r)
  : search(s), replace(r)
  {}
};

// Used when finding symbols in a line
struct SymbolToken {
  string::size_type startPos;
  string            symbol;

  SymbolToken(string::size_type p, const string &s)
  : startPos(p), symbol(s)
  {}
};


// Prototypes
void loadRenameTable(vector<RenameTable>& renameTable, const string &tableFileName);
void renameSymbols(string& line, const vector<RenameTable>& renameTable);


// Main program
int main(int argc, char *argv[])
{
  // Check parameter
  if(argc != 2) {
    cerr << endl;
    cerr << "Usage:" << endl;
    cerr << "  " << argv[0] << " in-file [> out-file]" << endl;
    cerr << endl;
    exit(1);
  }

  // Load name table
  string            tmp    = argv[0];
  string::size_type slsPos = tmp.rfind('/');
  string tableFileName(tmp, 0, slsPos);
  tableFileName += "/rename.table";

  vector<RenameTable> renameTable;
  loadRenameTable(renameTable, tableFileName);

  // Open the input file
  const char *inFile = argv[1];
  ifstream    is;

  is.open(inFile);
  if(is.fail()) {
    cerr << endl;
    cerr << "Couldn't open input file \"" << inFile << "\"" << endl;
    cerr << endl;
    exit(1);
  }

  // Read the file contents and perform conversion
  while(!is.eof()) {
    // Read one line
    string line;
    getline(is, line);
    // Rename symbols
    renameSymbols(line, renameTable);
    // Write the converted line to stdout
    cout << line << endl;
  }
}


// Load the renaming table
void loadRenameTable(vector<RenameTable>& renameTable, const string &tableFileName)
{
  // Open the name table file
  const char *inFile = tableFileName.c_str();
  ifstream    is;

  is.open(inFile);
  if(is.fail()) {
    cerr << endl;
    cerr << "Couldn't open name table file \"" << inFile << "\"" << endl;
    cerr << endl;
    exit(1);
  }

  // Read and parse its contents
  while(!is.eof()) {
    string line;
    getline(is, line);

    // Remove whitespace
    string::size_type p1 = line.find_first_not_of(' ');
    string::size_type p2 = line.find_last_not_of(' ');
    if((p1 == string::npos) || (p2 == string::npos)) continue;
    if((p1 > 0) || (p2 < (line.length()-1))) {
      string cut(line, p1, p2 - p1 + 1);
      line = cut;
    }

    // Check if it is an empty line or a comment line
    if(line.empty()) continue;
    if(line[0] == '#') continue;

    // Search for "\t"
    string::size_type p = line.find('\t');

    // Check if invalid
    if(p == string::npos) continue;
    if((p + 1) >= line.length()) continue;

    // Extract the "search" and "replace" values
    const string s(line, 0, p);
    const string r(line, p + 1, string::npos);

    // Check if the "search" and "replace" values are the same
    if(s == r) {
      cerr << endl;
      cerr << "Same replacement value \"" << s << "\" in name table file \"" << inFile << "\"" << endl;
      cerr << endl;
      exit(1);
    }

    // Put the value pairs in the table
    renameTable.push_back(RenameTable(s, r));
    //cerr << "\"" << s << "\"" << " -> " << "\"" << r << "\"" << endl;
  }
}

// Rename symbols on the give line
void renameSymbols(string& line, const vector<RenameTable>& renameTable)
{
  struct _HF {
    static inline bool isSymbolBChar(char ch)
    { return( (ch == '_') || (isalpha(ch) != 0) ); }

    static inline bool isSymbolSChar(char ch)
    { return( (ch == '_') || (isalnum(ch) != 0) ); }

    static void findSymbols(vector<SymbolToken> &tokens, const string& line)
    {
      enum EState { OutsideSymbol, InsideSymbol };

      tokens.clear();

      EState            curState = OutsideSymbol;
      string::size_type curIndex = 0;
      string::size_type curSmPos = 0;
      string            curSmTok;

      while(curIndex < line.length()) {
        char ch = line[curIndex];

        switch(curState) {
          case OutsideSymbol:
            if(isSymbolBChar(ch)) {
              curState = InsideSymbol;
              curSmPos = curIndex;
              curSmTok = ch;
            }
            break;
          case InsideSymbol:
            if(isSymbolSChar(ch))
              curSmTok += ch;
            else {
              curState = OutsideSymbol;
              tokens.push_back(SymbolToken(curSmPos, curSmTok));
            }
            break;
          default:
            break;
        }
        curIndex++;
      }

      if(curState == InsideSymbol) tokens.push_back(SymbolToken(curSmPos, curSmTok));

    }
  };

  // Find symbol tokens
  vector<SymbolToken> tokens;
  _HF::findSymbols(tokens, line);

  // Rename the found symbols based on the renaming table
  long offsetPos = 0;
  for(vector<SymbolToken>::const_iterator it = tokens.begin(); it != tokens.end(); ++it) {
    string::size_type startPos = it->startPos + offsetPos;
    const string      &symbol  = it->symbol;

    // Search within the renaming table
    for(vector<RenameTable>::const_iterator tt = renameTable.begin(); tt != renameTable.end(); ++tt) {
      const string &search  = tt->search;
      const string &replace = tt->replace;

      // If matched, replace it
      if(symbol == search) {
        const long sl = search.length();
        const long rl = replace.length();
        line.replace(startPos, sl, replace);
        offsetPos += (rl - sl);
        break;
      }
    }
  }
}
