// ############################################################################################
//  roomedit-RoomEdit-Main.cc
//  Implementation of roomedit's CRoomBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// roomedit's headers
#include "roomedit-RoomBuilder.h"

//
// Our room builder class
//
CRoomBuilder::CRoomBuilder()
: _open(false)
{
}

void CRoomBuilder::NewRoom()
{
  CloseRoom();
  _pBuilder.SetGridSize(); // Use default grid size from CPlane3DBuilder
  _open = true;
}

void CRoomBuilder::NewRoom(const TFloatXYZ &gridSize)
{
  CloseRoom();
  _pBuilder.SetGridSize(gridSize);
  _open = true;
}

void CRoomBuilder::CloseRoom()
{
  _open = false;
  _pBuilder.ClearAll();
  for(CP3DBufferMap::iterator it = _pBuffers.begin(); it != _pBuffers.end(); ++it) {
    delete it->second;
  }
  _pBuffers.clear();
}

void CRoomBuilder::NewPlane(const string &planeName, const TFloatXYZ &planeOrigin, uint32_t planeRowSize, uint32_t planeColSize, EPlaneDir planeDir)
{
  // Check if already exist
  if(_pBuffers.find(planeName) != _pBuffers.end())
    VError(VERR_Exist, _NFL_, "A plane with the same name is already exist");

  // Generate the plane
  _pBuilder.GeneratePlane(planeName, planeOrigin, planeRowSize, planeColSize, planeDir);

  // Generate CPanel3DBuffer for the plane
  vector<string> planeList;
  planeList.push_back(planeName);
  _pBuffers[planeName] = new CPanel3DBuffer;
  _pBuilder.GeneratePanel3DBuffer(*_pBuffers[planeName], planeList);
  _pBuffers[planeName]->SetColor(31, 0, 0);
}

void CRoomBuilder::RenamePlane(const string &oldPlaneName, const string &newPlaneName)
{
  // Check if not exist
  CP3DBufferMap::iterator it = _pBuffers.find(oldPlaneName);
  if(it == _pBuffers.end())
    VError(VERR_NotExist, _NFL_, "A plane with the given name is not exist");

  // Rename the plane
  _pBuilder.RenamePlane(oldPlaneName, newPlaneName);

  // Get the CPanel3DBuffer instance, delete the old map entry and put the new entry in the map
  CPanel3DBuffer *p3d = it->second;
  _pBuffers.erase(it);
  _pBuffers[newPlaneName] = p3d ;
}

void CRoomBuilder::DelPlane(const string &planeName)
{
  // Check if not exist
  CP3DBufferMap::iterator it = _pBuffers.find(planeName);
  if(it == _pBuffers.end())
    VError(VERR_NotExist, _NFL_, "A plane with the given name is not exist");

  // Delete the CPanel3DBuffer instance of the plane
  delete it->second;
  _pBuffers.erase(it);

  // Delete the plane
  _pBuilder.DeletePlane(planeName);
}

void CRoomBuilder::TranslatePlane(const string &planeName, const TFloatXYZ &factor)
{
  // Check if not exist
  CP3DBufferMap::iterator it = _pBuffers.find(planeName);
  if(it == _pBuffers.end())
    VError(VERR_NotExist, _NFL_, "A plane with the given name is not exist");
  
  // Translate the plane and recreate the plane's CPanel3DBuffer instance
  vector<string> planeList;
  planeList.push_back(planeName);
  _pBuilder.TranslateVertices(factor, planeList);
  _pBuilder.GeneratePanel3DBuffer(*_pBuffers[planeName], planeList);
  _pBuffers[planeName]->SetColor(31, 0, 0);
}

const CPlane3D &CRoomBuilder::Plane(const string &planeName)
{
  return(_pBuilder[planeName]);
}

void CRoomBuilder::RenderAll(const CMatrix3D &mat, const string &selectedObject)
{
  CP3DBufferMap::iterator  fn = _pBuffers.find(selectedObject);
  CPanel3DBuffer          *si = (fn != _pBuffers.end()) ? fn->second : NULL;

  for(CP3DBufferMap::iterator it = _pBuffers.begin(); it != _pBuffers.end(); ++it) {
    if(it->second == si) it->second->CFRender(mat);
    else                 it->second->UFRender(mat);
  }
}
